#include <dos.h>                        // FP_SEG() FP_OFF() getvect() setvect() enable() disable()
#include <conio.h>                      // getch() kbhit()
#include <stdio.h>                      // printf()
#include <stdlib.h>                     // exit()
#include <stddef.h>                     // offsetof()
#include <string.h>                     // strstr()
#include <mem.h>                        // memset()
#include <errno.h>                      // extern errno
#include <alloc.h>                      // coreleft()

#undef outportb
#undef outportw
#undef inportb
#undef inportw

#include "..\dpmi32\phys.h"
#include "..\dpmi32\intr.h"
#include "..\pci\pci.h"
#include "..\intel\timestmp.h"          // class PentiumClock
#include "..\fore\cardstmp.h"           // class CardClock
#include "..\intel\i8259a.h"            // class InterruptController
#include "..\intel\i8254-2.h"           // class HardwareTimer
#include "..\aal5\atm.h"                // struct AtmHeader
#include "..\fore\pca200e.h"
#include "..\intel\byteswap.h"          // ntohl()
#include "..\intel\fpu.h"               // struct FpuControl, get_fpu_control() set_fpu_control()
#include "..\wattcp\tcp.h"              // sock_XXX()
#include "..\integrat\buffer.h"         // class BufferBlock, class BufferQueue
#include "..\integrat\flat.h"           // align()
#include "..\integrat\dump.h"
#include "..\integrat\trace.h"          // struct tracerec
#include "..\integrat\fs2flows.h"       // flows_init() flows_per_block() flows_cleanup()



// conditional compilation switches
//
#define FORCE_PACKET_RESYNC             FALSE       // always on on FORE card
#define OLD_AALI                        FALSE
#define UART_TALKS                      TRUE
#define UART_LISTENS                    FALSE       // TRUE to type commands to ATM cards' ROM monitor (but interferes with flows_keystroke)
#define ALLOW_SONET_INTERRUPTS          FALSE
#define ALLOW_ATM_INTERRUPTS            TRUE
#define ALLOW_ATM_RX_FREE_EMPTY_IRQ     TRUE        // FALSE to only collect 2 blocks
#define ALLOW_TIMER_INTERRUPTS          FALSE       // FALSE to debug handler code
#define AFFECT_TIMER_INTERRUPTS         FALSE       // FALSE to allow BIOS and its piggybacks to get their timeslices
#define ONE_CARD_PER_TIMER_IRQ          TRUE
#define ESSENTIAL_STATUS_QUERIES        FALSE       // TRUE does 2 extra PCI slave reads (slows card)
#define UNEXPECTED_ATM_IRQ_PARANOIA     FALSE       // TRUE does 1 extra PCI slave read  (slows card)
#define DEBUG_STACK_OVERFLOW            FALSE
#define ATM_EXIT_ITERATES_TIMER_HANDLER FALSE       // TRUE to test w/o interrupts
#define ATM_EXIT_FILLS_PARTIAL_BLOCK    TRUE        // FALSE to see only what interrupt handlers gave you
#define ATM_EXIT_RELEASES_BLOCKS        TRUE
#define ATM_EXIT_DUMPS_ATM_HEADERS      FALSE
#define ATM_EXIT_DUMPS_TIMES            FALSE
#define PENDING_BLOCK_PARANOIA          TRUE
#define ATM_INIT_RESETS_SAR             TRUE
#define ATM_INIT_PROVIDES_BUFFERS       FALSE
#define ATM_INIT_SETS_PCI_TIMING        FALSE       // may contribute to lockup on 82439HX with 200 MHz CPU when TRUE
#define ATM_INIT_SETS_PCI_CONTROL       FALSE       // causes host CPU to get too few PCI bus cycles
#define ATM_INIT_SETS_CACHING           FALSE
#define ATM_SHIFT_CLEANUP_DEBUG         FALSE
#define ATM_HANDLER_BYPASSED            FALSE
#define SAVE_INTERRUPT_THIS_PTR         FALSE
#define SWAP_FIRMWARE                   FALSE
#define COPY_FIRMWARE_HEADER_TO_CARD    TRUE
#define PCI_INIT_STOPS_EARLY            FALSE
#define PCI_INIT_SETS_PCI_TIMING        FALSE       // causes lockup on SuperMicro P5STE (82439HX) with 200 MHz CPU when TRUE
#define PCI_INIT_SETS_PCI_MULTI_TIMING  TRUE        // causes cards to keep up better
#define PCI_INIT_SETS_PCI_CONTROL       FALSE       // causes lockup on ?ASUS? (82439HX) when TRUE
#define PCI_INIT_SETS_CACHING           FALSE       // causes lockup on SuperMicro P5STE when DPMIMEM=MAXMEM 65535 and caching already set to 512M in BIOS setup (CMOS)
#define CHECK_MEMORY_CONSISTENCY        TRUE        // no time wasted except on startup and shutdown
#define CHECK_MEMORY_WHILE_RUNNING      FALSE       // extra checks while >=1 card still running, affects RX free ring unavail count
#define CHECK_MEMORY_WHILE_LOOPING      FALSE       // extra checks inside main loop, big performance hit
#define DISABLE_TIMER_REENTRY           FALSE       // thought it would allow debugger to single-step interrupt handler, but it doesn't
#define LOCK_PAGES                      FALSE
#define ALLOW_STATS_CALLS               TRUE        // main call to flows_per_block()
#define ALLOW_STATS_CALLS_2             TRUE        // extra call to flows_per_block() when empty queue
#define ALLOW_STATS_CALLS_3             TRUE        // call to flows_query()
#define BUSYWAIT_STATS_CALLS            TRUE        // only if not allowed
#define BUSYWAIT_STATS_CALLS_2          FALSE       // not implemented, only if not allowed
#define BUSYWAIT_STATS_CALLS_3          FALSE       // only if not allowed
#define TIME_STATS_CALLS                TRUE        // main call to flows_per_block()
#define TIME_STATS_CALLS_2              FALSE       // not implemented, extra call to flows_per_block() when queried and queue is empty
#define TIME_STATS_CALLS_3              TRUE        // call to flows_query()
#define SIMULATE_FIBER                  FALSE       // when TRUE, please don't hook up any physical fiber
#define TEST_DIVISION                   TRUE
#define CARD_KEEPS_FIFO_STATS           TRUE
#define DO_FULLNESS_STATS               TRUE        // keep/display min/max/mean of BufferQueue filled list
#define TRACE_BUFFER_BLOCKS             FALSE       // save a BufferBlock we predict will be bad to disk



// compile-time switches which depend on others
//
#define PCI_EXIT_CALLS_BUFFER_DUMP      !DO_FOREGROUND_STATS

#if (CELLS_PER_BUFFER > 0) && !HOST_MUST_REASSEMBLE
#define CARD_HANDLES_TIMESTAMP_WRAP     TRUE
#else
#define CARD_HANDLES_TIMESTAMP_WRAP     FALSE
#endif



// manifest constants
//
//#define TIMER_FREQUENCY   65536       // should be every 55.07 milliseconds
//#define TIMER_FREQUENCY   12100       // is every 10.063 milliseconds
//#define TIMER_FREQUENCY   11900       // should be every 10.0 milliseconds
//#define TIMER_FREQUENCY   3840        // is every 3.21 milliseconds
//#define TIMER_FREQUENCY   3500        // is every 2.79 milliseconds
//#define TIMER_FREQUENCY   1279        // is every .8542 milliseconds
//#define TIMER_FREQUENCY   1280        // is every 1.066 milliseconds
//#define TIMER_FREQUENCY   1190        // should be every 1.00 milliseconds, is .858 mS
//#define TIMER_FREQUENCY   900         // is every .642 milliseconds (also)
//#define TIMER_FREQUENCY   800         // is every .642 milliseconds
//#define TIMER_FREQUENCY   700         // is every .429 milliseconds     ***
#define TIMER_FREQUENCY   500         // is every .214 milliseconds
//#define TIMER_FREQUENCY   300         // is every .208 microseconds
//#define TIMER_FREQUENCY   250         // is every .208 microseconds

#define TIMER_CALIBRATE     3           // record start timestamp here

#define ATM_CALIBRATE       5           // record interrupt timestamp here

#define SONET_RESET_DELAY   60          // 1/2 microsecond worth of 120 MHz Pentium clocks
#define ATM_RESET_DELAY     64.0        // count of PCI clocks
#define ROM_START_DELAY     15.0        // count of (floating point) seconds
#define FIRM_START_DELAY    5.0         // count of (floating point) seconds
#define FIRM_INIT_DELAY     1.0         // count of seconds
//#define TIME_CLEAR_DELAY    0.001       // count of seconds (1 mS)
#define TIME_CLEAR_DELAY  100.0         // count of seconds (100 S)
#define SHUTDOWN_DELAY     10.0         // count of seconds (10 S)
//#define BEAT_CHANGE_DELAY 100.0         // count of seconds (100 S)
#define BEAT_CHANGE_DELAY  10.0         // count of seconds (10 S)
#define BEAT_CHANGE_HOLDOFF 0.1         // count of seconds
#define UART_DELAY_INITIAL  0.1         // count of seconds
#define UART_DELAY_HEART    0.1         // count of seconds
#define UART_DELAY_FINAL    0.5         // count of seconds
#define UART_DELAY_COMMAND  0.1         // count of seconds
#define UART_DELAY_RUNNING  0.0         // count of seconds
#define UART_DELAY_FAILURE  3600.0      // count of seconds (1 hour)
#define UART_DELAY_SHUTDOWN 1.0         // count of seconds (1 S)

#define FLOW_BLOCK_TIME     400         // how long should dummy flows_per_block() take? (microseconds)
//#define FLOW_QUERY_TIME   56          // how long should dummy flows_query() take? (milliseconds)
#define FLOW_QUERY_TIME     0           // how long should dummy flows_query() take? (milliseconds)

#define DESIRED_MIN_GRANT   2

//#define DESIRED_CACHE_CONTROL    0x01   // (default w/o cache card) L1 cache enabled, L2 cache not forcing miss, NA# signal enabled, SRAM is 0K pipe burst
#define DESIRED_CACHE_CONTROL    0xB1   // (default with cache card installed) L1 cache enabled, L2 cache not forcing miss, L2 stops at 64 MB, NA# signal enabled, SRAM is 512K dual bank pipe burst
//#define DESIRED_CACHE_CONTROL    0xB0   // L1 cache disabled, L2 cache disabled, NA# signal enabled, SRAM is 512K dual bank pipelined burst
//#define DESIRED_CACHE_CONTROL    0xB3   // L1 cache enabled, L2 cache is forcing miss, NA# signal enabled, SRAM is 512K dual bank pipelined burst

//#define DESIRED_HOST_CONTROL   0xE0   // CPU inactivity timer = 8 PCI clocks
#define DESIRED_HOST_CONTROL     0x80   // (default) CPU inactivity timer = 5 PCI clocks
//#define DESIRED_HOST_CONTROL   0x40   // CPU inactivity timer = 3 PCI clocks
//#define DESIRED_HOST_CONTROL   0x20   // CPU inactivity timer = 2 PCI clocks
//#define DESIRED_HOST_CONTROL   0x00   // CPU inactivity timer = 1 PCI clocks

#define DESIRED_HOST_LATENCY        0   // works well when card latency is also 0
//#define DESIRED_HOST_LATENCY      8   // works just as badly as 64, regardless of card latency
//#define DESIRED_HOST_LATENCY     64   // AMI Triton BIOS default

#define DESIRED_MULTI_LATENCY      32   // ???works??? when CPU latency=32, card latency=64

//#define DESIRED_CARD_LATENCY        0   // only setting that allows 2 cards to keep up with IP-header-only capture
#define DESIRED_CARD_LATENCY      8   // raw capture uses 10 RX FIFO slots with buf=4, 48-byte cells
//#define DESIRED_CARD_LATENCY     16   // raw capture uses 7 RX FIFO slots with buf=4, 48-byte cells
//#define DESIRED_CARD_LATENCY     32   // raw capture uses 6 RX FIFO slots with buf=4, 48-byte cells
//#define DESIRED_CARD_LATENCY     64   // (default) raw capture uses 5 RX FIFO slots with buf=4, 48-byte cells
//#define DESIRED_CARD_LATENCY    128   // raw capture uses 5 RX FIFO slots with buf=4, 48-byte cells
//#define DESIRED_CARD_LATENCY    248   // raw capture uses 5 RX FIFO slots with buf=4, 48-byte cells

#define UART_BUFFER_SIZE         1000   // how much input do we care about?

// how much addressible memory does card require?
//
#define CARD_MEM_FOOTPRINT              sizeof(pca200e_memmap)

#define HEARTBEAT_CHECKS            5   // how many times do we check heartbeat to be sure it is increasing/decreasing?
//#define HEARTBEAT_CHECKS          0   // how many times do we check heartbeat to be sure it is increasing/decreasing?

#define MAX_FIFO_DEPTH           2047   // how many buckets for possible SONET-to-i960 FIFO depths?



// error return codes
//
#define ERROR_CARD_BASE                 2000
#define ERROR_CARD_SHIFT_TOO_SLOW       (ERROR_CARD_BASE + 0)
#define ERROR_CARD_CANNOT_OPEN_FILE     (ERROR_CARD_BASE + 1)
#define ERROR_CARD_CANNOT_READ_HDR      (ERROR_CARD_BASE + 2)
#define ERROR_CARD_CANNOT_READ_BODY     (ERROR_CARD_BASE + 2)
#define ERROR_CARD_CANNOT_SEEK_END      (ERROR_CARD_BASE + 3)
#define ERROR_CARD_CANNOT_SEEK_BEG      (ERROR_CARD_BASE + 4)
#define ERROR_CARD_CANNOT_TELL_SIZE     (ERROR_CARD_BASE + 5)
#define ERROR_CARD_BSTAT_STUCK          (ERROR_CARD_BASE + 6)
#define ERROR_CARD_CP_NOT_RUNNING       (ERROR_CARD_BASE + 7)
#define ERROR_CARD_SELF_TEST_FAILED     (ERROR_CARD_BASE + 8)
#define ERROR_CARD_SELF_TEST_UNKNOWN    (ERROR_CARD_BASE + 9)
#define ERROR_CARD_HEARTBEAT_STUCK      (ERROR_CARD_BASE + 10)
#define ERROR_CARD_CP_FAULTED           (ERROR_CARD_BASE + 11)
#define ERROR_CARD_CMD_STUCK            (ERROR_CARD_BASE + 12)
#define ERROR_CARD_CMD_INVALID          (ERROR_CARD_BASE + 13)
#define ERROR_CARD_CMD_FAILED           (ERROR_CARD_BASE + 14)
#define ERROR_CARD_PREV_CMD_PENDING     (ERROR_CARD_BASE + 15)

#define ERROR_PCI_BASE                  3000
#define ERROR_PCI_TOO_FEW_CARDS_FOUND   (ERROR_PCI_BASE + 0)



unsigned timer_interrupt_card_index = 0;
unsigned long timer_interrupt_count = 0;
const unsigned timer_irq_level = 0;
unsigned char timer_interrupt;
vfp timer_thunk;
Bit48 old_timer_handler;
PentiumClock timer_start_time;
PentiumClock timer_end_time;
InterruptController *irq = 0;
HardwareTimer *timer = 0;
class AtmCard *cards[ MAX_ATM_NICS ? MAX_ATM_NICS : 1 ];
Bit8 old_pci_latency_timer;
#if INTEL_82439HX
    Bit8 old_pci_multi_timer;
#endif
static union { PciControl old_pci_control; Bit8 old_pci_control_alias; };
static union { PciCacheControl old_pci_cache_control; Bit8 old_pci_cache_control_alias; };
#if SAVE_INTERRUPT_THIS_PTR
    class AtmCard *atm_this_pointer_during_interrupt=(AtmCard *)0x12345678;         // debug only!
#endif
#if DEBUG_STACK_OVERFLOW
    int stack_ptr_recorded = FALSE;
    Bit16 stack_seg_during_main_top = 0;
    Bit16 stack_seg_during_main_loop = 0;
    Bit16 stack_seg_during_global_timer = 0;
    Bit16 stack_seg_during_class_timer = 0;
    Bit32 stack_seg_during_atm_shift = 0;
    Bit32 stack_ptr_during_main_top = 0;
    Bit32 stack_ptr_during_main_loop = 0;
    Bit32 stack_ptr_during_global_timer = 0;
    Bit32 stack_ptr_during_class_timer = 0;
    Bit32 stack_ptr_during_atm_shift = 0;
#endif
#if LOCK_PAGES
    int atm_shift_dummy_before( void );
#endif
#if TRACE_BUFFER_BLOCKS
    Bit8 *block_trace = (Bit8 *) 0x1F6F9000;
#endif



// all the data that is per-card
//
class AtmCard
{

private:
    enum AtmShiftTiming
    {
        during_initialization,
        during_interrupt,
        during_cleanup
    };

    Bit32 phys_handle;
    Bit32 phys_address;
    Bit8 old_pci_latency_timer;
    pca200e_memmap *memmap;
    unsigned irq_level;
    unsigned char interrupt_number;
    vfp thunk;
    Bit48 old_handler;
    CardClock atm_interrupt_time;
    PentiumClock atm_enabled;
    PentiumClock before;
    PentiumClock elapsed;
    PentiumClock elapsed_tot;
    PentiumClock elapsed_min;
    PentiumClock elapsed_max;
    int atm_interrupt_error;
    int timer_interrupts_allowed;
    unsigned atm_out_fifo_full;
    unsigned atm_in_fifo_almost_full;
    unsigned long atm_interrupt_count;
    unsigned long atm_interrupt_unexpected;
    unsigned long sonet_interrupt_count;
    unsigned long timer_interrupt_count;
    union { HostControl_read atm_status; Bit32 atm_status_alias; };
    union { PciMasterControl old_pci_master_control; Bit8 old_pci_master_control_alias; };
    int atm_uart_new_line;
    int init_result;

    // note: my_block has been moved to public area, below
    BufferBlock *next_block;

    Status *atm_command_result;
    Qcard_cmd *atm_command;
    const char *atm_command_name;
    PentiumClock atm_command_start_time;

#if ATM_SHIFT_CLEANUP_DEBUG
    Bit32 bufcount_overflow_first_used;
    Bit32 bufcount_overflow_last_used;
    Bit32 bufcount_overflow_highest_used;
    Bit32 bufcount_nonoverflow_last_used;
    Bit32 bufcount_overflow_sar_rx_free;
    unsigned bufcount_overflow_prev_free;
    unsigned bufcount_overflow_this_free;
    unsigned bufcount_overflow_next_free;
    Bit32 bufcount_overflow_phys_free_base;
    PentiumClock bufcount_overflow_first_time;
    PentiumClock bufcount_overflow_last_time;
    PentiumClock bufcount_overflow_highest_time;
    PentiumClock bufcount_nonoverflow_last_time;
#endif

    // nonstatic card interrupt handler that does the real work
    //
    int handler( void );

    // separate handlers for up to 2 cards, so we can identify
    // which card caused the interrupt
    //
    static int handler0(
        Bit32 ebp,
        Bit32 edi,
        Bit32 esi,
        Bit32 ds,
        Bit32 es,
        Bit32 edx,
        Bit32 ecx,
        Bit32 ebx,
        Bit32 eax,
        Bit32 eip,
        Bit32 cs,
        Bit32 flags
    )
    {
#if ATM_HANDLER_BYPASSED
        flat_put_long( 0xFFFBE008, 0 );
        InterruptController::acknowledge( 3 );
        return 0;
#else
        return cards[ 0 ]->handler();
#endif
    }

    static int handler1(
        Bit32 ebp,
        Bit32 edi,
        Bit32 esi,
        Bit32 ds,
        Bit32 es,
        Bit32 edx,
        Bit32 ecx,
        Bit32 ebx,
        Bit32 eax,
        Bit32 eip,
        Bit32 cs,
        Bit32 flags
    )
    {
        return cards[ 1 ]->handler();
    }

    // re-route the card interrupt to our code
    //
    // note that we must set global variable cards[cardindex]=this
    // rather than relying upon our caller to do it for us,
    // because if an interrupt occurs after we enable them,
    // but before he has assigned to global, handler0() will
    // read garbage (or zero) for the object pointer, and we
    // will corrupt host memory and probably lock up!
    //
    // it is safe for caller to assign to same global,
    // since he is not really changing the value there
    //
    void interrupt_init( unsigned cardindex )
    {
        interrupt_number = irq_to_intr( irq_level );
        old_handler = getvect( interrupt_number );
        disable();

#if ALLOW_ATM_INTERRUPTS || ALLOW_SONET_INTERRUPTS
        if( cardindex == 0 )
        {
            thunk = make_thunk( handler0, &old_handler );
        }
        else if( cardindex == 1 )
        {
            thunk = make_thunk( handler1, &old_handler );
        }
        else
        {
            printf( "fatal internal error!  cannot set vector for more than 2 ATM cards\n" );
            exit( 123 );
        }
#if THROW_IS_LEGAL
        if( thunk == NULL )
            throw "error!  cannot allocate card interrupt thunk!\n";
#endif

        Bit48 pointer;
        pointer.offset = (Bit32) thunk;
        pointer.selector = _CS;
        setvect( interrupt_number, pointer );

        cards[ cardindex ] = this;        // very important!  see above
        irq->unmask( irq_level );
#else
        irq->mask( irq_level );
#endif
        enable();
    }

    void interrupt_exit( void )
    {
        disable();
#if ALLOW_ATM_INTERRUPTS || ALLOW_SONET_INTERRUPTS
        setvect( interrupt_number, old_handler );
        free( thunk );
#endif
        enable();
    }

    void sonet_init( void );

    void sonet_exit( void );

    void sonet_handler( void );

    int atm_reset_hardware( void );

    int atm_download( void );

    int atm_check_heartbeat( int want_increasing );

    int atm_await_command( double uart_delay, double command_delay, volatile Status *status_location );

    int atm_shutdown_command( void );

    int atm_init_command( void );

    int atm_init( void );

    void atm_exit( void );

    void atm_handler( void );

    int atm_shift_hw_rings( AtmShiftTiming timing );

public:
    // this field is public so main() can simulate a packet from us
    //
    // maybe we should use an accessor function instead?
    //
    Bit16 pci_handle;

    // index into people's global arrays of this card
    //
    Bit8 cardindex;

    // this field is public so it can be passed to the
    // foreground stats code when a remote query occurs,
    // so the query will reflect all data seen so far,
    // in case it will be 10 days until the block fills
    //
    // maybe we should use an accessor function instead?
    //
    BufferBlock *my_block;

    int get_init_result( void )
    {
        return init_result;
    }

    // called when BufferQueue stops being empty,
    // in case this card was in need of a buffer
    //
    void atm_foreground_replenish( void );

    // allows user to wait for command to complete, but doesn't allow
    // them to specify the location of the status variable
    //
    // I tried to do this using one public definition of this function
    // where the last argument was defaulted, but the brain-dead Borland
    // C++ 4.5 compiler said "Error capture.cpp 1159: Member
    // atm_command_result cannot be used without an object"
    // even after I made atm_command_result public!
    //
    // oh well, this is safer anyways
    //
    int atm_await_command( double uart_delay, double command_delay )
    {
        return atm_await_command( uart_delay, command_delay, atm_command_result );
    }

    int atm_clear_timestamp_command( void );

    void atm_uart_io( const char *output, char *input, unsigned input_size, double delay );

    // called when main() wants to flush flows for a
    // partially-filled block because it got queried
    //
    void set_pending_block_usage( void )
    {
        if( my_block )
        {
            Bit32 current = memmap->shared.queues.current_host_block;
            if( !current )
                my_block->used = COMPL_RING_ENTRIES;
            else
            {
                my_block->used = (current - flat_to_physical( my_block->buffers )) / BUFFER_SIZE;
#if PENDING_BLOCK_PARANOIA
                if( my_block->used > COMPL_RING_ENTRIES )
                {
                    printf( "card 0x%.4X internal error: calculated %d usage in pending block; truncated to max %d\n", pci_handle, my_block->used, COMPL_RING_ENTRIES );
                    printf( "card 0x%.4X internal error: current=0x%.8X buffers=0x%.8X phys\n", pci_handle, current, flat_to_physical(my_block->buffers) );
                    my_block->used = COMPL_RING_ENTRIES;
                }
#endif
            }
        }
    }

    // nonstatic timer interrupt handler
    //
    // DANGER!  accessing stack variables, including
    // the hidden ones used for structure return values,
    // causes exception E and locks up machine!
    //
    void timer_handler( void )
    {
#if DEBUG_STACK_OVERFLOW
        if( !stack_ptr_recorded )
        {
            //stack_ptr_recorded = TRUE;
            stack_ptr_during_class_timer = _ESP;
            stack_seg_during_class_timer = _SS;
        }
#endif

        if( !timer_interrupts_allowed )
            return;

#if DISABLE_TIMER_REENTRY
        timer_interrupts_allowed = FALSE;
#endif

#if ESSENTIAL_TIME_STAMPS
        before.set();
#endif

#if ESSENTIAL_STATUS_QUERIES
        // get the status so the foreground can tell
        // whether RX overrun occurred before last cell
        //
        // probably only useful for code debug
        //
        atm_status_alias = memmap->dword_00h;

        if( atm_status.out_fifo_full )
            atm_out_fifo_full++;

        if( atm_status.in_fifo_almost_full )
            atm_in_fifo_almost_full++;
#endif

        int result = atm_shift_hw_rings( during_interrupt );
        if( result != 0 )
            atm_interrupt_error = result;

#if ESSENTIAL_TIME_STAMPS
        elapsed.set();
        timer_interrupt_count++;
        elapsed -= before;
        elapsed_tot += elapsed;
        if( elapsed > elapsed_max )
            elapsed_max = elapsed;
        if( elapsed < elapsed_min )
            elapsed_min = elapsed;
#endif

#if DISABLE_TIMER_REENTRY
        timer_interrupts_allowed = TRUE;
#endif
    }

    // constructor
    //
    AtmCard(
        Bit16 pci_handle0,
        Bit32 new_phys_address,
        unsigned new_irq_level,
        unsigned cardindex
    )
    :
        cardindex( cardindex ),
        pci_handle( pci_handle0 ),
        phys_address( new_phys_address ),
        irq_level( new_irq_level ),
        timer_interrupts_allowed( FALSE ),
        atm_uart_new_line( TRUE ),
        atm_command( NULL )
    {
        // get addressibility for card hardware
        //
        memmap = (pca200e_memmap *) map_physical( phys_address, CARD_MEM_FOOTPRINT, &phys_handle );

#if ESSENTIAL_TIME_STAMPS
        elapsed_min.high = -1;
        elapsed_min.low  = -1;
        elapsed_max.high = 0;
        elapsed_max.low  = 0;
        elapsed_tot.high = 0;
        elapsed_tot.low  = 0;
        timer_interrupt_count = 0;
#endif

        interrupt_init( cardindex );

        sonet_init();
        printf( "card 0x%4.4X SONET initialized\n", pci_handle );

        init_result = atm_init();
        printf( "card 0x%4.4X ATM initialized with result %d\n", pci_handle, init_result );
    }

    // destructor
    //
    ~AtmCard()
    {
        // clean up ATM SAR chip
        //
        atm_exit();

        // clean up SONET framer chip
        //
        sonet_exit();

        // clean up interrupt grabbing
        //
        interrupt_exit();

        // give up addressibility for card
        //
        unmap_physical( CARD_MEM_FOOTPRINT, phys_handle );
    }
};



// deals with interrupts from any of 2 sources
// (SONET framer chip and/or ATM SAR chip) and 2 cards
// (monitoring send and receive fibers for a single link)
//
// since we poll the shared memory, we only use
// this interrupt to report severe errors
//
int AtmCard::handler( void )
{
#if ESSENTIAL_TIME_STAMPS
#endif

    // assume ATM SAR wants attention...
    //
    atm_handler();

    // tell the interrupt controllers we are done
    //
    irq->acknowledge( irq_level );

#if ESSENTIAL_TIME_STAMPS
#endif

    // tell the thunking code not to chain to the old handler
    //
    return 0;
}



// initialize the SONET framer chip
//
void AtmCard::sonet_init( void )
{
    // we haven't had any interrupts from
    // this chip on this card yet
    //
    sonet_interrupt_count = 0;
}



// deals with interrupts from the TI TNETA1500 SONET framer
//
// (alarms and MIB counters not implemented yet)
//
void AtmCard::sonet_handler( void )
{
    // set a flag to indicate that this occurred
    //
    sonet_interrupt_count++;

    // maybe retrieve the counters?
    //
}



void AtmCard::sonet_exit( void )
{
    // dump our meager statistics
    //
    printf( "card 0x%4.4X had %ld SONET framer interrupts\n", pci_handle, sonet_interrupt_count );
}



// writes any output you give me to the software UART,
// reads any input already available from the UART,
// and waits the given amount of time for more input
//
// based on conditional compile symbols, this code may
// also display input from UART on standard output and/or
// accept keystrokes to output to the card
//
// prefixes each line displayed on STDOUT so you can tell
// which card provided it, and that it came from the UART
//
// assumes input_size is nonzero when input is non-NULL
//
// will always terminate the input string
//
void AtmCard::atm_uart_io( const char *output, char *input, unsigned input_size, double delay )
{
    char c;
    Soft_UART *uart = &memmap->shared.mon960.uart;
    PentiumClock now, initial, alarm;
    double timing;
#if UART_TALKS
    char signon[ 100 ];
#endif

#if UART_TALKS
    sprintf( signon, "card 0x%4.4X UART: ", pci_handle );
#endif

    // preallocate the terminating NUL character
    //
    input_size--;

    initial.set();
    alarm.put_seconds( delay );
    alarm += initial;

    // at least once...
    //
    do
    {
        now.set();

        // if there is input ready for us...
        //
        if( CAN_GETC_i960( uart ) )
        {
            // read it into a local variable
            //
            GETC_i960( uart, c );

#if UART_TALKS
            // if nothing has been output on this screen line
            // and this character is not going to wrap back to
            // the beginning of the line...
            //
            if( atm_uart_new_line && c!='\r')
            {
                // remember we output on this line
                //
                atm_uart_new_line = FALSE;

                // insert a prefix message before the character
                //
                fputs( signon, stdout );
            }

            // show on screen
            //
            putchar( c );

            // if this was a newline character...
            //
            if( c == '\n' )

                // forget we have output any lines
                //
                atm_uart_new_line = TRUE;
#endif

            // if the caller wants input,
            // and there is room left...
            //
            if( input && input_size )
            {
                // save input character for caller
                //
                *input++ = c;

                // less space available now
                //
                input_size--;
            }

            // reset timer
            //
            initial = now;
            alarm.put_seconds( delay );
            alarm += initial;
        }

        // if there is room in output buffer...
        //
        if( CAN_PUTC_i960(uart,*output) )
        {
            // if the caller wants to output
            // and we haven't done all of that yet...
            //
            if( output && *output )
            {
                c = *output++;
            }
#if UART_LISTENS
            else if( kbhit() )
            {
                c = getch();
                if( c == 27 )           // escape
                    break;
            }
#endif
            else
                c = 0;

            if( c )
            {
                // spit out one character
                //
                PUTC_i960( uart, c );

                // reset timer
                //
                initial = now;
                alarm.put_seconds( delay );
                alarm += initial;
            }
        }
    }
    // ...as long as time hasn't run out
    //
    while( now < alarm );


    // terminate the input string
    //
    if( input )
        *input = '\0';
}



int AtmCard::atm_reset_hardware( void )
{
    // set the boot status word to cold start, so the
    // i960 ?ROM code will do the right thing after we reset it
    //
    memmap->shared.mon960.bstat = cold_start;

    union { HostControl_write hcr_w; Bit32 hcr_w_alias; };

    // save any PCI registers that will get trashed by the reset
    //
    // ???apparently none get trashed
    //

    // start a software reset
    //
    // FORE recommends 8-bit accesses for this,
    // although testing shows that 32-bit works too
    //
    hcr_w_alias = 0;
    hcr_w.board_reset = TRUE;
#if 1
    memmap->dword_00h = hcr_w_alias;
#else
    memmap->hcr_w = hcr_w;
#endif

    // wait for the NIC to finish resetting
    //
    // FORE recommends 63 PCI clocks
    //
    PentiumClock::wait( (ATM_RESET_DELAY / PCI_CLOCK_RATE) * CLOCK_RATE );

    // stop resetting the card
    //
    // FORE recommends 8-bit accesses for this,
    // although testing shows that 32-bit works too
    //
    hcr_w.board_reset = FALSE;
#if 1
    memmap->dword_00h = hcr_w_alias;
#else
    memmap->hcr_w = hcr_w;
#endif

    // restore any PCI registers that got trashed by the reset
    //
    // ???apparently none were trashed
    //

    // wait a few seconds for the boot status word
    // to make a transition
    //
    Boot_status prev_status = cold_start;
    Boot_status status;
    PentiumClock now, initial, alarm;
    double timing;

    initial.set();
    now = initial;
    alarm.put_seconds( ROM_START_DELAY );
    alarm += now;

    while( now<alarm )
    {
        status = memmap->shared.mon960.bstat;
        if( status != prev_status )
        {
            now -= initial;
            timing = now.get_seconds();

            // check that self test succeeded
            //
            if( status == self_test_ok )
                break;

            if( status == self_test_fail )
            {
                printf( "card 0x%4.4X firmware self test failed after %g seconds\n", pci_handle, timing );
                return ERROR_CARD_SELF_TEST_FAILED;
            }

            printf( "card 0x%4.4X warning: firmware transitioned to unknown status 0x%8.8lX after %g seconds\n", pci_handle, status, timing );
            prev_status = status;
        }

        now.set();
    }

    if( status == prev_status )
    {
        printf( "card 0x%4.4X ROM stayed in 0x%8.8lX state past %g seconds\n", pci_handle, status, ROM_START_DELAY );
        if( status == cold_start )
            return ERROR_CARD_BSTAT_STUCK;

        return ERROR_CARD_SELF_TEST_UNKNOWN;
    }

    // tell the user how long it took to succeed
    //
    printf( "card 0x%4.4X ROM took %g S to cold start and self test\n", pci_handle, timing );

    return 0;
}



// download i960 code image to card
//
int AtmCard::atm_download( void )
{
    FILE *handle;
    FirmwareHeader header;
    int stdio_result;
    Bit8 *start;
    Bit32 *current;
    Bit32 one_word;
    long size, count;

    // disable cr-lf to lf translation!
    //
    handle = fopen( FIRMWARE_FILE_NAME, "rb" );
    if( handle == NULL )
    {
        printf( "card 0x%4.4X error %d opening file '%s'\n", pci_handle,
            errno, FIRMWARE_FILE_NAME );
        return ERROR_CARD_CANNOT_OPEN_FILE;
    }

    // get the file size
    //
    stdio_result = fseek( handle, 0, SEEK_END );
    if( stdio_result != 0 )
    {
        printf( "card 0x%4.4X error %d seeking to end of file '%s'\n",
            pci_handle, errno, FIRMWARE_FILE_NAME );

        fclose( handle );
        return ERROR_CARD_CANNOT_SEEK_END;
    }

    size = ftell( handle );
    if( size == -1L )
    {
        printf( "card 0x%4.4X error %d getting file '%s' position\n",
            pci_handle, errno, FIRMWARE_FILE_NAME );

        fclose( handle );
        return ERROR_CARD_CANNOT_TELL_SIZE;
    }

    // do we need to skip to a certain position in the file
    // to read the header???  it seems to be at 0x44 not 0...
    //
    fseek( handle, 0, SEEK_SET );
    if( stdio_result != 0 )
    {
        printf( "card 0x%4.4X error %d seeking to beginning of file '%s'\n",
            pci_handle, errno, FIRMWARE_FILE_NAME );

        fclose( handle );
        return ERROR_CARD_CANNOT_SEEK_BEG;
    }

    // read in the file header
    //
    stdio_result = fread( &header, sizeof(FirmwareHeader), 1, handle );
    if( stdio_result != 1 )
    {
        printf( "card 0x%4.4X error %d reading file '%s' %d byte header; got %d items\n",
            pci_handle, errno, FIRMWARE_FILE_NAME, sizeof(FirmwareHeader), stdio_result );

        fclose( handle );
        return ERROR_CARD_CANNOT_READ_HDR;
    }

    // we don't need to byte swap the values
    // because host (Pentium) and card (i960)
    // are both little-endian
    //
    // but if we ever do, make sure not to swap
    // again later along with rest of firmware image
    //

    // shortcut for the load address
    //
    start = memmap->ram8 + header.start_offset;
    printf( "card 0x%4.4X firmware image will load at 0x%8.8lX and start at 0x%8.8lX\n", pci_handle, start, memmap->ram8+header.entry_address );

#if COPY_FIRMWARE_HEADER_TO_CARD
    // copy the header to card memory
    //
    memcpy( start, &header, sizeof(FirmwareHeader) );

    start += sizeof( FirmwareHeader );
#endif

    // read the rest of the file into card memory via a local
    // variable because it only supports 32-bit accesses
    //
    for(
        count = sizeof(FirmwareHeader),
        current = (Bit32 *) start;

        count < size;

        count += sizeof( Bit32 ),
        current++
    )
    {
        stdio_result = fread( &one_word, sizeof(Bit32), 1, handle );
        if( stdio_result != 1 )
        {
            printf( "card 0x%4.4X error %d reading file '%s' body word %ld; got %d items\n",
                pci_handle, errno, FIRMWARE_FILE_NAME, count, stdio_result );

            fclose( handle );
            return ERROR_CARD_CANNOT_READ_BODY;
        }

#if SWAP_FIRMWARE
        *current = ntohl( one_word );
#else
        *current = one_word;
#endif
    }

    fclose( handle );

    static char in_buffer[ UART_BUFFER_SIZE ];
    static char out_buffer[ 20 ];
    PentiumClock initial, now, alarm;
    double timing;
    Boot_status status;

    // suck up any console output from the ROM
    //
    // should be "Mon960\r\n"
    //
    atm_uart_io( NULL, in_buffer, UART_BUFFER_SIZE, UART_DELAY_INITIAL );

    // need to prod the NINDY monitor
    // so it will accept commands from us
    //
    // should give a signon message and a command prompt
    //
    sprintf( out_buffer, "\r" );
    atm_uart_io( out_buffer, in_buffer, NULL, UART_DELAY_INITIAL );

    // if I ever re-enable this, do it with a timeout!
    //
#if 0
    // wait for the command prompt from the NINDY monitor
    //
    do
    {
        atm_uart_io( NULL, in_buffer, UART_BUFFER_SIZE, UART_DELAY_INITIAL );
    }
    while( !strstr( in_buffer, "=>" ) );
#endif

    // tell (the NINDY OS (that the card booted after reset))
    // to start the code we downloaded
    //
    sprintf( out_buffer, "go %4.4X\r", header.entry_address & 0xFFFF );
    atm_uart_io( out_buffer, in_buffer, UART_BUFFER_SIZE, UART_DELAY_INITIAL );

    // move on to the next line of screen,
    // since NINDY monitor's prompt doesn't
    //
    printf( "\n" );

    // check for a complaint from the monitor
    //
    if( strstr( in_buffer, "fault" ) )
    {
        printf( "card 0x%4.4X ROM could not start firmware; last boot status was 0x%8.8lX\n", pci_handle, status );

        // give the user a chance to play with ROM monitor manually
        //
        printf( "card 0x%4.4X is letting you talk to ROM monitor; press escape when done\n", pci_handle );
        atm_uart_io( "\r", NULL, 0, UART_DELAY_FAILURE );

        return ERROR_CARD_CP_FAULTED;
    }

    // wait a few seconds for the boot status
    // to say the CP (cell processor) is running
    //
    initial.set();
    now = initial;
    alarm.put_seconds( FIRM_START_DELAY );
    alarm += now;

    while( now < alarm )
    {
        status = memmap->shared.mon960.bstat;
        if( status == cp_running )
            break;

        now.set();
    }

    if( status != cp_running )
    {
        printf( "card 0x%4.4X firmware took longer than %g seconds to run after download; last boot status was 0x%8.8lX\n", pci_handle, FIRM_START_DELAY, status );
        return ERROR_CARD_CP_NOT_RUNNING;
    }

    // tell the user how long this took
    //
    now -= initial;
    timing = now.get_seconds();
    printf( "card 0x%4.4X firmware took %g S to run\n", pci_handle, timing );

    return 0;
}



// check that the heartbeat is steadily decreasing
// or increasing
//
// assumes that wraparound will not occur now,
// since we just started the puppy, and it is a
// 32-bit value, and we hope it isn't decrementing
// very fast
//
// does not make any assumptions about how much the
// heartbeat changes each time, since that isn't documented
//
int AtmCard::atm_check_heartbeat( int want_increasing )
{
    PentiumClock initial, now, alarm;
    double timing;
    Bit32 prev_heartbeat = memmap->shared.queues.heartbeat;
    Bit32 heartbeat;
    unsigned count;
    const char *direction;

    if( want_increasing )
    {
        direction = "increase";
    }
    else
    {
        direction = "decrease";
    }

    now.set();
    for( count=0; count<HEARTBEAT_CHECKS; count++ )
    {
        initial = now;
        alarm.put_seconds( BEAT_CHANGE_DELAY );
        alarm += initial;

        while( now<alarm )
        {
            heartbeat = memmap->shared.queues.heartbeat;

            now.set();

            if( want_increasing )
            {
                if( heartbeat > prev_heartbeat )
                    break;
            }
            else
            {
                if( heartbeat < prev_heartbeat )
                    break;
            }

            // suck up any output it has to give
            //
            atm_uart_io( NULL, NULL, 0, UART_DELAY_HEART );

            // wait a bit to give the card time to process,
            // since we stop the card CPU every time we
            // do a slave PCI access to card
            //
            PentiumClock::wait( BEAT_CHANGE_HOLDOFF * CLOCK_RATE );
        }

        if( now >= alarm )
        {
            printf( "card 0x%4.4X heartbeat took longer than %g seconds to %s\n", pci_handle, BEAT_CHANGE_DELAY, direction );
            return ERROR_CARD_HEARTBEAT_STUCK;
        }

        alarm = now;
        alarm -= initial;
        timing = alarm.get_seconds();
        printf( "card 0x%4.4X iteration %d took %.4f S to %s from 0x%8.8lX to 0x%8.8lX\n", pci_handle, count, timing, direction, prev_heartbeat, heartbeat );

        prev_heartbeat = heartbeat;
    }

    return 0;
}



// wait for the results of a command
// previously issued to this card
//
int AtmCard::atm_await_command(
    double uart_delay,
    double command_delay,
    volatile Status *status_location
)
{
    PentiumClock now, alarm;
    double timing;
    Status status2;

    // suck up any output it has to give
    //
    atm_uart_io( NULL, NULL, 0, uart_delay );

    // wait for the card to act on the command
    //
    alarm.put_seconds( command_delay );
    alarm += atm_command_start_time;
    do
    {
        now.set();
        status2 = *status_location;
    } while( status2==stat_pending && now<alarm );

    // turn card UART usage back on, so
    // we can see output of next command
    //
    memmap->shared.queues.hlogger = TRUE;

    // check the results of command
    //
    if( status2 == stat_pending )
    {
        printf( "card 0x%4.4X '%s' cmd took longer than %g seconds\n", pci_handle, atm_command_name, command_delay );
        return ERROR_CARD_CMD_STUCK;
    }

    now -= atm_command_start_time;
    timing = now.get_seconds();

    // make sure card turned off pending bit,
    // didn't turn on free bit, and turned on
    // complete bit
    //
    if( (status2 & stat_pending)
    ||  (status2 & stat_free)
    || !(status2 & stat_complete) )
    {
        printf( "card 0x%4.4X '%s' cmd transitioned to invalid state %d after %g seconds\n", pci_handle, atm_command_name, status2, timing );
        return ERROR_CARD_CMD_INVALID;
    }

    // make sure command did not fail
    //
    if( status2 & stat_error )
    {
        printf( "card 0x%4.4X '%s' cmd failed after %g seconds (last heartbeat was 0x%8.8lX)\n", pci_handle, atm_command_name, timing, memmap->shared.queues.heartbeat );
        return ERROR_CARD_CMD_FAILED;
    }

    printf( "card 0x%4.4X '%s' command succeeded in %g seconds\n", pci_handle, atm_command_name, timing );

    return 0;
}



// issue card's "clear timestamp" command, but do not check results
// so other cards in the same machine can be checked quickly
//
int AtmCard::atm_clear_timestamp_command( void )
{
    // check that no command is pending,
    // since we only have one status address
    //
    Status status2 = *atm_command_result;
    if( (status2 & stat_pending) || !(status2 & stat_complete) )
    {
        printf( "card 0x%4.4X cannot clear timestamp; a previous '%s' command is still pending\n", pci_handle, atm_command_name );
        return ERROR_CARD_PREV_CMD_PENDING;
    }

    // remember the name of this command for later error messages
    //
    atm_command_name = "clear timestamp";

    // remember that card has not acted on our command yet
    //
    *atm_command_result = stat_pending;

    // turn off card UART usage, so this card doesn't wait
    // to print before doing the command
    //
    memmap->shared.queues.hlogger = FALSE;

    // remember when we started command
    //
    atm_command_start_time.set();

    // issue the command!
    //
    atm_command->cmd.op = op_clear_timestamp;

    return 0;
}



// issue card's "shutdown" command and wait for results
//
int AtmCard::atm_shutdown_command( void )
{
    PentiumClock now, alarm;
    double timing;
    Status status2;
    u_int heartbeat;

    // check that no command is pending,
    // since we only have one status address
    //
    status2 = *atm_command_result;
    if( (status2 & stat_pending) || !(status2 & stat_complete) )
    {
        printf( "card 0x%4.4X cannot shutdown; a previous '%s' command is still pending\n", pci_handle, atm_command_name );
        return ERROR_CARD_PREV_CMD_PENDING;
    }

    // remember the name of this command for later error messages
    //
    atm_command_name = "shutdown";

    // mark this command as pending, in case the caller tries
    // to do some other command afterwards, which would be quite
    // invalid, since we will cause the card code to EXIT!
    //
    *atm_command_result = stat_pending;

    // remember when we started command
    //
    atm_command_start_time.set();

    // issue the command!
    //
    atm_command->cmd.op = op_shutdown;

    // wait for the card to act on the command
    //
    alarm.put_seconds( SHUTDOWN_DELAY );
    alarm += atm_command_start_time;
    do
    {
        // suck up any output it has to give, so that
        // (1) user can see it, and so (2) code on card
        // does not hang trying to print and hence never
        // exit properly
        //
        atm_uart_io( NULL, NULL, 0, 0 );

        now.set();
        heartbeat = memmap->shared.queues.heartbeat;
    }
    while( heartbeat!=SAFE_SHUTDOWN_BY_USER && now<alarm );

    // did we timeout?
    //
    if( heartbeat != SAFE_SHUTDOWN_BY_USER )
    {
        printf( "card 0x%4.4X '%s' cmd took longer than %g seconds; last heartbeat was 0x%8.8X\n", pci_handle, atm_command_name, SHUTDOWN_DELAY, heartbeat );
        return ERROR_CARD_CMD_STUCK;
    }

    // suck up any final output from firmware and MON960
    //
    atm_uart_io( NULL, NULL, 0, UART_DELAY_SHUTDOWN );

    now -= atm_command_start_time;
    timing = now.get_seconds();

    // prefix this output with a newline because we MON960
    // command prompt leaves us in the middle of a line
    //
    printf( "\ncard 0x%4.4X '%s' command succeeded in %g seconds\n", pci_handle, atm_command_name, timing );

    return 0;
}



// fill in parms, issue card's initialize command, check results
//
int AtmCard::atm_init_command( void )
{
    Init_block *init = &memmap->shared.queues.init;

    // read the heap size (in bytes) from the card
    //
    // someday maybe we will calculate the sizes
    // that we want to use from the heap size
    //
    printf( "card 0x%4.4X heap base=0x%8.8lX size=0x%8.8lX\n",
        pci_handle,
        memmap->shared.queues.heap_base,
        memmap->shared.queues.heap_size
    );

#if OLD_AALI
    // all our buffers will be in the small queue for scheme 1,
    // and they will be big enough to hold the first N cells
    // of the pcaket, which we what transferred to the host
    //
    // there probably isn't memory on the card for it to know
    // about as much of host memory at once as we would like
    // (1 MB / 1 48-byte cell = 21824 buffers would be nice)
    // so we may have to lower pool_size
    //
    init->small_b1.queue_length = 2;
    init->small_b1.buffer_size = 64;
    init->small_b1.pool_size = 60 * 2;
    init->small_b1.supply_blksize = 60;

    // the large queue for scheme 1 will be empty, so the
    // remainder of the packet will be discarded
    //
    // but we cannot set the supply block size below 4
    // and we cannot set the buffer size below ???
    //
    // I am just hoping I don't have to give it any supply blocks...
    //
    init->large_b1.queue_length = 0;
    init->large_b1.buffer_size = 0;
    init->large_b1.pool_size = 0;
    init->large_b1.supply_blksize = 0;

    // both the large and small queues for buffer scheme 2
    // will be empty all the time because we will never
    // specify that buffer scheme when activating a VP/VC
    //
    init->small_b2.queue_length = 0;
    init->small_b2.buffer_size = 0;
    init->small_b2.pool_size = 0;
    init->small_b2.supply_blksize = 0;

    init->large_b2.queue_length = 0;
    init->large_b2.buffer_size = 0;
    init->large_b2.pool_size = 0;
    init->large_b2.supply_blksize = 0;

    // we will not be using any connectionless VPI/VCI combos
    //
    init->conless_vpvc = 0;

    // we won't be transmitting so no need to extend TPD
    //
    init->tpd_extension = 0;

    // the default RPD (RX packet descriptor) holds
    // 3 RBD's (RX buffer descriptors) and since we
    // only expect one buffer to be filled per packet,
    // no extension of the RPD is appropriate
    //
    init->rpd_extension = 0;

    // allow a 2 BufferBlock's to be pending host processing
    //
    init->rqueue_len = 60 * 2;

    // we won't be doing transmit, so transmit-done
    // queue can be zero-length
    //
    init->tqueue_len = 0;

    // lets try something modest for now...
    //
    // allow 1024 receive state table entries
    //
    init->num_connect = 1024;

    // idunno, probably just for connectionless VP/VC
    // which we do not use
    //
    init->receive_threshold = 0;
#else
    // tell the card to use 20 bits of VPI/VCI for indexing
    // when looking up AAL5 receive state
    //
    // should require 128 K bytes on card
    //
    init->index_bits_avail = 20;

    // 12 least significant of the index bits should come
    // from the VCI, leaving 8 most significant to come
    // from the least significant bits of VPI
    //
    // so we get 4096 VC's in each of 256 VP's
    //
    init->bits_from_vci = 12;

    // most significant bits of VPI and VCI which are not
    // used to index into RX state table should be zero
    //
    init->high_vpi_vci = 0;

    // tell the card about size of our buffer blocks
    //
    // location has already been filled in by atm_shift_hw_rings()
    //
    init->host_block_size = COMPL_RING_ENTRIES * BUFFER_SIZE;
#endif  // OLD_AALI

    // we aren't going to buffer up more than one command
    // to the card, so we might as well tell the card to
    // only allocate one command queue entry, which lets
    // us not bother cycling through the entries in it,
    // so we will always use entry 0
    //
    init->cqueue_len = 1;

    // allow the card to printf() its complaints to software UART
    // (firmware images that came on floppies from FORE do not
    // seem to have such messages compiled in)
    //
    memmap->shared.queues.hlogger = TRUE;

    // remember the name of this command for later error messages
    //
    atm_command_name = "initialize";

    // remember that card has not acted on our command yet
    //
    init->status = stat_pending;

    // remember when we started command
    //
    atm_command_start_time.set();

    // issue the command!
    //
    init->op = op_initialize;

    return atm_await_command( UART_DELAY_FINAL, FIRM_INIT_DELAY, &init->status );
}



// initialize the ATM SAR chip
//
int AtmCard::atm_init( void )
{
    Bit32 pointer;
    unsigned count;
    int result;
    PentiumClock initial, now, alarm;
    double timing;

#if ATM_SHIFT_CLEANUP_DEBUG
    // atm_shift_hw_rings() has never seen buffer count overflow
    //
    bufcount_overflow_first_used = 0;
    bufcount_overflow_last_used = 0;
    bufcount_overflow_highest_used = 0;
    bufcount_nonoverflow_last_used = 0;
    bufcount_overflow_sar_rx_free = -1;
    bufcount_overflow_prev_free = -1;
    bufcount_overflow_this_free = -1;
    bufcount_overflow_next_free = -1;
    bufcount_overflow_phys_free_base = -1;
#endif

    // we haven't had any interrupts from
    // this chip on this card yet
    //
    atm_interrupt_count = 0;
    atm_interrupt_unexpected = 0;

    // we haven't had any errors from this card yet
    //
    atm_interrupt_error = 0;

    // PBI has never run out of in FIFO or out FIFO entries
    //
    atm_out_fifo_full = 0;
    atm_in_fifo_almost_full = 0;

    // show current values of some PCI configuration registers
    //
    Bit8 value;

    value = pci_get_min_grant( pci_handle );
    printf( "card 0x%4.4X minimum-grant register was %d (%g uS)\n", pci_handle, value, value*0.25 );

    value = pci_get_max_latency( pci_handle );
    printf( "card 0x%4.4X maximum-latency register was %d (%g uS)\n", pci_handle, value, value*0.25 );

    value = pci_get_max_slave_latency( pci_handle );
    printf( "card 0x%4.4X maximum slave latency register was %d cycles\n", pci_handle, value );

    old_pci_latency_timer = pci_get_master_latency_timer( pci_handle );
    printf( "card 0x%4.4X master-latency-timer register was %d (%g uS)\n", pci_handle, old_pci_latency_timer, old_pci_latency_timer*0.25 );

    value = pci_get_cache_line_size( pci_handle );
    printf( "card 0x%4.4X cache line size was %d words\n", pci_handle, value );

    value = pci_get_threshold( pci_handle );
    printf( "card 0x%4.4X large burst or continuous request threshold was %d 32-bit FIFO words\n", pci_handle, value );

    old_pci_master_control_alias = pci_get_master_control( pci_handle );
    printf( "card 0x%4.4X PCI master control register was 0x%2.2X (cache line reads %s,\n"
        " write-and-invalidate %s, write-and-invalidate needs >%d lines,\n"
        " latency timer %s, continuous request mode %s,\n"
        " large PCI bursts %s forced, slave RAM accesses %s byteswapped)\n",
        pci_handle,
        old_pci_master_control_alias,
        old_pci_master_control.disable_cache_line_reads   ? "disabled" : "enabled",
        old_pci_master_control.disable_write_and_invals   ? "disabled" : "enabled",
        old_pci_master_control.write_inval_needs_2_lines  ? 2          : 1,
        old_pci_master_control.ignore_latency_timer       ? "ignored"  : "respected",
        old_pci_master_control.enable_cont_request_mode   ? "enabled"  : "disabled",
        old_pci_master_control.force_large_pci_bus_bursts ? "are"      : "are not",
        old_pci_master_control.byteswap_slave_ram_access  ? "are"      : "are not"
    );

#if ATM_INIT_SETS_PCI_CONTROL
    // try to change PCI master control
    // to desired value, and check that it worked
    //
    union { PciMasterControl new_pci_master_control; Bit8 new_pci_master_control_alias; };
    new_pci_master_control_alias = old_pci_master_control_alias;
    new_pci_master_control.enable_cont_request_mode = TRUE;
    pci_set_master_control( pci_handle, new_pci_master_control_alias );
    value = pci_get_master_control( pci_handle );
    if( value != new_pci_master_control_alias )
        printf( "card 0x%4.4X unable to set master control register to 0x%.2X; is now 0x%.2X\n", pci_handle, new_pci_master_control, value );
    else
        printf( "card 0x%4.4X enabled continuous request mode in master control register\n", pci_handle );
#endif

#if ATM_INIT_SETS_PCI_TIMING
    // try to change PCI minimum-grant and maximum-latency
    // to desired values, and check that it worked
    //
    pci_set_min_grant( pci_handle, DESIRED_MIN_GRANT );
    value = pci_get_min_grant( pci_handle );
    if( value != DESIRED_MIN_GRANT )
        printf( "card 0x%4.4X unable to set minimum-grant register to %d; is now %d (%g uS)\n", pci_handle, DESIRED_MIN_GRANT, value, value*0.25 );
    else
        printf( "card 0x%4.4X changed minimum-grant register to %d (%g uS)\n", pci_handle, DESIRED_MIN_GRANT, value*0.25 );

    pci_set_master_latency_timer( pci_handle, DESIRED_CARD_LATENCY );
    value = pci_get_master_latency_timer( pci_handle );
    if( value != DESIRED_CARD_LATENCY )
        printf( "card 0x%4.4X unable to set master-latency-timer register to %d; is now %d (%g uS)\n", pci_handle, DESIRED_CARD_LATENCY, value, value*0.25 );
    else
        printf( "card 0x%4.4X changed master-latency-timer register to %d (%g uS)\n", pci_handle, DESIRED_CARD_LATENCY, value*0.25 );
#endif

    // debug
    printf( "card 0x%4.4X initial boot status at log 0x%8.8lX is 0x%8.8lX\n",
        pci_handle,
        &memmap->shared.mon960.bstat,
        memmap->shared.mon960.bstat
    );
    printf( "card 0x%4.4X HCR at log 0x%8.8lX has value 0x%2.2X byte, 0x%8.8lX word\n",
        pci_handle,
        &memmap->dword_00h,
        *(Bit8 *)&memmap->dword_00h,
        memmap->dword_00h
    );

#if ATM_INIT_RESETS_SAR
    // force a hardware reset
    //
    result = atm_reset_hardware();
    if( result )
        return result;

    // download firmware to card and start it
    //
    result = atm_download();
    if( result )
        return result;

    // tell the user where we think heartbeat is
    //
    printf( "card 0x%4.4X heartbeat is at logical address 0x%8.8lX\n", pci_handle, &memmap->shared.queues.heartbeat );

    // make sure that the heartbeat is decreasing
    //
    result = atm_check_heartbeat( FALSE );
    if( result )
        return result;

    my_block = 0;
    next_block = 0;
#if ATM_INIT_PROVIDES_BUFFERS
    result = atm_shift_hw_rings( during_initialization );
    if( result )
        return result;
#endif

    // remember when we enabled receive
    //
    atm_enabled.set();

    // tell the card what sizes to make things,
    // and start it reading from all its queues
    //
    result = atm_init_command();
    if( result )
        return result;
#endif      // ATM_INIT_RESETS_SAR

    // make sure that the heartbeat is increasing
    //
    result = atm_check_heartbeat( TRUE );
    if( result )
        return result;

    // enable per-block interrupts
    //
    // when we get this, card will already have started using
    // the second block we gave it earlier, so handler has plenty
    // of time to give it another block, so allocation can occur
    // at the beginning of interrupt handler
    //
    union { HostInterruptMask himr; Bit32 himr_alias; };
    himr_alias = 0;
#if ALLOW_ATM_RX_FREE_EMPTY_IRQ
    himr.is_masked = FALSE;
#else
    himr.is_masked = TRUE;
#endif
    memmap->dword_01h = himr_alias;

    // this old method of masking interrupts has been superceded
    // by the Host Interrupt Mask register (HIMR) at the 1MB line
    // in memory space, but we will do both in case of stupid card
    // code
    //
#if ALLOW_ATM_RX_FREE_EMPTY_IRQ
    memmap->shared.queues.imask = 1;        // 1=enabled
#else
    memmap->shared.queues.imask = 0;        // 0=disabled
#endif

    // card does not need to wait, the next
    // time it wants to interrupt us
    //
    memmap->shared.queues.istat = 0;

    // allow timer interrupts
    //
    timer_interrupts_allowed = TRUE;

    // allocate the command result in locked memory
    //
    atm_command_result = (Status *) flat_alloc( 4 );

    // last command we issued succeeded
    //
    *atm_command_result = stat_complete;

    // in case somebody is stupid enough to try printing
    // the previous command name, we better initialize it
    //
    atm_command_name = "none ever issued";

    // remember where the command queue is on the card
    //
    atm_command = (Qcard_cmd *) (memmap->ram8 + (Bit32)memmap->shared.queues.command_queue);

    // tell the card where the (only) command status is
    //
    atm_command->stataddr = flat_to_physical( atm_command_result );

    return 0;
}



void AtmCard::atm_exit( void )
{
    int result;
    unsigned count;
    double timing;
    Bit8 value;
    Bit32 buffer_count, block_count;
    CardClock first, last;

#if ATM_EXIT_ITERATES_TIMER_HANDLER
    // call timer_handler() many times, as if we were the
    // interrupt handler
    //
    // warning: would cause totally inaccurate information
    // to be reported if BufferBlock didn't contain a field
    // to say how many buffers were actually filled!!!
    //
    for( count=0; count<1; count++ )
    {
        timer_handler();
    }
#endif

    // ignore timer interrupts from now on
    //
    timer_interrupts_allowed = FALSE;

#if CHECK_MEMORY_WHILE_RUNNING
    // not too nice to check here since other cards
    // may still be running
    //
    all_buffers->check();
#endif

    // maybe suck the (SONET framer) statistics out of the card???
    //

#if ATM_EXIT_FILLS_PARTIAL_BLOCK
    // count any cells left over in pending block, which were not
    // quite enough to cause an pending block full interrupt
    //
    // must do this before zeroing next_host_block pointer on card!
    //
    set_pending_block_usage();
#endif

    // disable all future card interrupts
    //
    union { HostInterruptMask himr; Bit32 himr_alias; };
    himr_alias = 0;
    himr.is_masked = TRUE;
    memmap->dword_01h = himr_alias;

    // stop the card from writing to host memory
    //
    if( atm_command )
    {
        result = atm_shutdown_command();
        atm_command->stataddr = 0;
    }

#if ATM_INIT_SETS_PCI_CONTROL
    // reset PCI master control for this card
    //
    pci_set_master_control( pci_handle, old_pci_master_control_alias );
    value = pci_get_master_control( pci_handle );
    if( value != old_pci_master_control_alias )
        printf( "card 0x%4.4X unable to restore master control register to 0x%.2X; is now 0x%.2Xn", pci_handle, old_pci_master_control_alias, value );
    else
        printf( "card 0x%4.4X restored master control register to 0x%.2X\n", pci_handle, old_pci_master_control_alias );
#endif

#if ATM_INIT_SETS_PCI_TIMING
    // reset the master latency timer for this card
    //
    pci_set_master_latency_timer( pci_handle, old_pci_latency_timer );
    value = pci_get_master_latency_timer( pci_handle );
    if( value != old_pci_latency_timer )
        printf( "card 0x%4.4X unable to restore master-latency-timer register to %d; is now %d (%g uS)\n", pci_handle, old_pci_latency_timer, value, value*0.25 );
    else
        printf( "card 0x%4.4X restored master-latency-timer register to %d (%g uS)\n", pci_handle, old_pci_latency_timer, value*0.25 );
#endif





    // dump our meager statistics
    //

    if( atm_out_fifo_full )
        printf( "card 0x%4.4X PBI out FIFO was full %d times during run\n", pci_handle, atm_out_fifo_full );
    else
        printf( "card 0x%4.4X PBI out FIFO was available all during run\n", pci_handle );

    if( atm_in_fifo_almost_full )
        printf( "card 0x%4.4X PBI in  FIFO was almost full %d times during run\n", pci_handle, atm_in_fifo_almost_full );
    else
        printf( "card 0x%4.4X PBI in  FIFO was available all during run\n", pci_handle );

    printf( "card 0x%4.4X ATM last interrupt status was 0x%8.8lX\n", pci_handle, atm_status_alias );
    atm_status_alias = memmap->dword_00h;
    printf( "card 0x%4.4X ATM final status 0x%8.8lX\n", pci_handle, atm_status_alias );

    union { PciStatus status; Bit8 status_alias; };
    status_alias = pci_get_status( pci_handle );
    printf( "card 0x%4.4X PCI final status 0x%4.4X\n", pci_handle, status_alias );

    // dump the ready-to-fill buffer block
    //
    printf( "card 0x%4.4X next buffer block is 0x%p\n", pci_handle, next_block );

    // dump the currently-being-filled buffer block
    //
    printf( "card 0x%4.4X was filling buffer block 0x%p\n", pci_handle, my_block );

#if ATM_EXIT_FILLS_PARTIAL_BLOCK
#if ATM_EXIT_DUMPS_ATM_HEADERS
    // dump the ATM headers in the completion ring
    //
    AtmHeaderCoerce *atm = (AtmHeaderCoerce *) (my_block->buffers + 8);
    for( count=0; count<my_block->used; count++ )
    {
        // if the SAR ever filled this entry in...
        //
        // we could check timestamp for nonzero
        // but that would require flows_per_block
        // to have zeroed it earlier, which is too
        // much overhead
        //
        if( TRUE )
        {
            printf( "card 0x%4.4X ATM header %d GFC=%1.1X CLP=%d PTI210=%d%d%d VCI=%d VPI=%d\n",
                pci_handle,
                count,
                atm->bits.GFC,
                atm->bits.CLP,
                atm->bits.PTI2, atm->bits.PTI1, atm->bits.PTI0,
                atm->bits.VCIL + (atm->bits.VCIH<<12),
                atm->bits.VPI
            );
        }

        atm = (AtmHeaderCoerce *) (BUFFER_SIZE + (Bit8 *)atm);
    }
#endif

    // if there is a non-empty block left over...
    //
    if( my_block && my_block->used>0 )
    {
        // place my_block on global filled list
        // even though it is not quite all filled
        //
        result = all_buffers->fill( my_block );
        if( !result )
            my_block = NULL;
    }
#endif

#if ATM_EXIT_RELEASES_BLOCKS
    if( my_block )
    {
        all_buffers->free( my_block );
        my_block = NULL;
    }

    if( next_block )
    {
        all_buffers->free( next_block );
        next_block = NULL;
    }
#endif

    printf( "card 0x%4.4X had %ld SAR interrupts, %ld of which were unexpected\n",
        pci_handle, atm_interrupt_count, atm_interrupt_unexpected );
//    if( atm_interrupt_count > 0 )
    {
        printf( "card 0x%4.4X last error during interrupt was %d\n", pci_handle, atm_interrupt_error );

    #if ESSENTIAL_TIME_STAMPS
        double elapsed_min2, elapsed_max2, elapsed_avg2;
        printf( "card 0x%4.4X had %ld timer interrupts\n", pci_handle, timer_interrupt_count );
        if( timer_interrupt_count )
        {
            elapsed_min2 = ((double)elapsed_min * 1000) / CLOCK_RATE;
            elapsed_max2 = ((double)elapsed_max * 1000) / CLOCK_RATE;
            elapsed_avg2 = (((double)elapsed_tot / timer_interrupt_count) * 1000) / CLOCK_RATE;
            printf( "card 0x%4.4X timer_handler() took min %.4f max %.4f avg %.4f mS\n", pci_handle, elapsed_min2, elapsed_max2, elapsed_avg2 );
        }
    #else
        printf( "card 0x%4.4X sorry, AtmCard::handler() timing is unavailable\n", pci_handle );
    #endif
    }

    // dump the host buffer queue states
    //
    buffer_count = all_buffers->amount_filled( pci_handle, block_count );
    printf( "card 0x%4.4X has %ld/%ld filled buffers/blocks in host queue\n", pci_handle, buffer_count, block_count );

#if CARD_KEEPS_FIFO_STATS
    int seen = 0;
    for( count=0; count<=MAX_PHY_FIFO_DEPTH; count++ )
        if( memmap->shared.queues.fifo_depth_counts[ count ] > 0 )
        {
            printf( "card 0x%.4X counted FIFO depth %d %d times\n", pci_handle, count, memmap->shared.queues.fifo_depth_counts[count] );
            seen++;
        }
    printf( "card 0x%.4X used %d FIFO depths\n", pci_handle, seen );
#endif

    BufferBlock *calibration_block = all_buffers->read( pci_handle, ATM_CALIBRATE - 1 );
    if( !calibration_block )
        printf( "card 0x%4.4X did not fill %d buffer blocks, so calibration cannot be done\n", pci_handle, ATM_CALIBRATE );
    else
    {
        last = calibration_block->last_timestamp();
        first = calibration_block->first_timestamp();
        last.strip_bitfields();
        first.strip_bitfields();
        timing = CardClock( last - first ).get_seconds();
        printf( "card 0x%4.4X took %g S total between filling the first and %dth buffer in %dth block\n", pci_handle, timing, calibration_block->used, ATM_CALIBRATE );

        if( calibration_block->used != 0 )
        {
            timing /= calibration_block->used;
            printf( "card 0x%4.4X took %g S to fill each buffer in that one block\n", pci_handle, timing );
            timing = 1 / timing;
            printf( "card 0x%4.4X saw %g packets per S in block %d\n", pci_handle, timing, ATM_CALIBRATE );
        }
#if ATM_CALIBRATE>=2
        timing = CardClock( calibration_block->last_timestamp()
               - all_buffers->read( pci_handle, (unsigned)0 )->first_timestamp() ).get_seconds();
        printf( "card 0x%4.4X took %g S total between filling the first buffer of first block and %dth buffers of %dth block\n", pci_handle, timing, calibration_block->used, ATM_CALIBRATE );
        timing /= ((ATM_CALIBRATE - 1) * (double)COMPL_RING_ENTRIES) + calibration_block->used;
        printf( "card 0x%4.4X took %g S to fill each buffer\n", pci_handle, timing );
        timing = 1 / timing;
        printf( "card 0x%4.4X saw %g packets per S in first %d blocks\n", pci_handle, timing, ATM_CALIBRATE );
#endif
    }






    // count fifo depths across all blocks
    //
    Bit16 fifo_depth;
    Bit32 fifo_depth_counts[ MAX_FIFO_DEPTH + 1 ];
    for( count=0; count<=MAX_FIFO_DEPTH; count++ )
        fifo_depth_counts[ count ] = 0;

    Bit32 sum_used = 0;

    CardClock prev_when, this_when, next_when;
    CardClock delta_when, max_delta, min_delta, sum_delta;
    min_delta.low  = BIT32_MAX;
    min_delta.high = BIT32_MAX;
    max_delta.low  = 0;
    max_delta.high = 0;
    sum_delta.low  = 0;
    sum_delta.high = 0;

    for( block_count=0; block_count<SW_RINGS; block_count++ )
    {
        BufferBlock *calibration_block = all_buffers->read( pci_handle, block_count );
        if( !calibration_block )
            break;

        sum_used += calibration_block->used;

        tracerec *data = (tracerec *) calibration_block->buffers;

        if( block_count == 0 )
        {
            prev_when = data[0].when;
            prev_when.strip_bitfields();
            count = 1;
        }
        else
        {
            count = 0;
        }
        this_when = data[count].when;
        fifo_depth = this_when.strip_bitfields();

        for( ; count<calibration_block->used; count++ )
        {
#if !CARD_HANDLES_TIMESTAMP_WRAP
            if( this_when.low < prev_when.low )
            {
                // assumes hardware timestamp is 16 bits
                // and the field it is in is >16 bits
                //
                // note that this does not fix the absolute
                // timestamp of current record, just its value
                // relative to previous record
                //
                prev_when.low -= 65536;
                prev_when.high -= 1;
            }
#endif

            delta_when = CardClock( this_when - prev_when );

#if ATM_EXIT_DUMPS_TIMES
            if( fifo_depth > 1 )
            {
                if( kbhit() )
                    break;

                printf( "card 0x%.4X noticed FIFO depth %d at timestamp 0x%X%.8X delta %d\n",
                    pci_handle,
                    fifo_depth,
                    this_when.high, this_when.low,
                    delta_when.low
                );
            }
#endif
            if( fifo_depth > MAX_FIFO_DEPTH )
                fifo_depth = MAX_FIFO_DEPTH;
            fifo_depth_counts[ fifo_depth ]++;
            next_when = data[count+1].when;
            fifo_depth = next_when.strip_bitfields();

#if CARD_HANDLES_TIMESTAMP_WRAP
            if( this_when < prev_when )
            {
                printf( "card 0x%.4X ignoring out-of-order timestamp 0x%X%.8X in record %d of block %d;"
                    "prev/next timestamps were 0x%X%.8X/0x%X%.8X\n",
                    pci_handle,
                    this_when.high, this_when.low,
                    count, block_count,
                    prev_when.high, prev_when.low,
                    next_when.high, next_when.low
                );
         //       continue;
            }
#endif

            sum_delta += delta_when;
            if( delta_when > max_delta )
                max_delta = delta_when;
            if( delta_when < min_delta )
                min_delta = delta_when;
            prev_when = this_when;
            this_when = next_when;
        }
#if ATM_EXIT_DUMPS_TIMES
        if( kbhit() )
        {
            printf( "card 0x%.4X stopped examining %dth block because of keypress\n", pci_handle, block_count );
            getch();
        }
#endif

    }   // end for all blocks

    if( sum_used > 0 )
    {
        timing = sum_delta.get_seconds() / sum_used;
        printf( "card 0x%4.4X timestamp delta in all blocks was %g/%g/%g S min/max/avg\n",
            pci_handle,
            min_delta.get_seconds(), max_delta.get_seconds(), timing
        );

        for( count=0; count<=MAX_FIFO_DEPTH; count++ )
            if( fifo_depth_counts[ count ] > 0 )
                printf( "card 0x%.4X host counted FIFO depth %d %d times across all blocks\n", pci_handle, count, fifo_depth_counts[count] );
    }
    else
    {
        printf( "card 0x%.4X captured no data, so we cannot calculate"
            " timestamp deltas or FIFO depth counts\n", pci_handle );
    }






    // dump some of the queue values in Cp_queue???
    //

#if SAVE_INTERRUPT_THIS_PTR
    // dump the "this" pointer seen by the interrupt handler
    //
    // if it doesn't match up with the value printed by pci_init()
    // we have trouble in river city!
    //
    printf( "'this' pointer last seen by interrupt handler was 0x%p\n", atm_this_pointer_during_interrupt );
    printf( "card 0x%4.4X correct 'this' pointer is 0x%p\n", pci_handle, this );
#endif

#if ATM_SHIFT_CLEANUP_DEBUG
    if( bufcount_nonoverflow_last_used != 0 )
    {
        timing = (double) (bufcount_nonoverflow_last_time - atm_enabled) / (CLOCK_RATE / 1000.0);
        printf( "card 0x%4.4X buffer count last nonoverflow to %d %g mS after enabling RX\n", pci_handle, bufcount_nonoverflow_last_used, timing );
    }

    if( bufcount_overflow_first_used != 0 )
    {
        timing = (double) (bufcount_overflow_first_time - atm_enabled) / (CLOCK_RATE / 1000.0);
        printf( "card 0x%4.4X buffer count first overflowed to %d %g mS after enabling RX\n", pci_handle, bufcount_overflow_first_used, timing );
    }

    if( bufcount_overflow_last_used != 0 )
    {
        timing = (double) (bufcount_overflow_last_time - atm_enabled) / (CLOCK_RATE / 1000.0);
        printf( "card 0x%4.4X buffer count last overflowed to %d %g mS after enabling RX\n", pci_handle, bufcount_overflow_last_used, timing );
    }

    if( bufcount_overflow_highest_used != 0 )
    {
        timing = (double) (bufcount_overflow_highest_time - atm_enabled) / (CLOCK_RATE / 1000.0);
        printf( "card 0x%4.4X buffer count highest overflowed to %d %g mS after enabling RX\n", pci_handle, bufcount_overflow_highest_used, timing );
        printf( "card 0x%4.4X buffer count overflowed when SAR RX free ptr was 0x%8.8lX\n", pci_handle, bufcount_overflow_sar_rx_free );
        printf( "card 0x%4.4X buffer count overflowed when prev_free was %d\n", pci_handle, bufcount_overflow_prev_free );
        printf( "card 0x%4.4X buffer count overflowed when this_free was %d\n", pci_handle, bufcount_overflow_this_free );
        printf( "card 0x%4.4X buffer count overflowed when next_free was %d\n", pci_handle, bufcount_overflow_next_free );
        printf( "card 0x%4.4X buffer count overflowed when my_free_base was phys 0x%8.8lX\n", pci_handle, bufcount_overflow_phys_free_base );
    }
#endif

}



void AtmCard::atm_handler( void )
{
    int result;

    // remember CPU time when clock interrupted us as early
    // in the handler as possible, so we have the best shot
    // at card clock calibration
    //
    // remember we cannot access stack variables at interrupt
    // time because the compiler assumes stack segment register
    // is same as data segment, and it isn't
    //
    before.set();

#if SAVE_INTERRUPT_THIS_PTR
    atm_this_pointer_during_interrupt = this;
#endif

#if ESSENTIAL_STATUS_QUERIES
    // get the status so the foreground can tell
    // whether RX overrun occurred before last cell
    //
    // probably only useful for code debug
    //
    // definitely slows down card by stopping its i960
    // CPU to do the slave PCI read
    //
    atm_status_alias = memmap->dword_00h;

    if( atm_status.out_fifo_full )
        atm_out_fifo_full++;

    if( atm_status.in_fifo_almost_full )
        atm_in_fifo_almost_full++;
#endif

#if UNEXPECTED_ATM_IRQ_PARANOIA
    // if the card needs next block filled in...
    //
    // probably only useful for code debug
    //
    // definitely slows down card by stopping its i960
    // CPU to do the slave PCI read
    //
    if( !memmap->shared.queues.next_host_block )
    {
#endif
        result = atm_shift_hw_rings( during_interrupt );
        if( result && !atm_interrupt_error )
            atm_interrupt_error = result;

        // if the card is working on a block now...
        //
        if( my_block )
        {
            // get the timestamp of the packet that caused
            // the card to switch blocks, which is now the
            // first in the new block
            //
            // remember that on a slow link, the packet before
            // this one might have arrived 10 minutes ago!
            //
            atm_interrupt_time = * (CardClock *) my_block->buffers;
            atm_interrupt_time.strip_bitfields();

            // compensate for card clock rate drift
            //
            atm_interrupt_time.recalibrate_clock_rate( before );
        }
#if UNEXPECTED_ATM_IRQ_PARANOIA
    }
    else
    {
        atm_interrupt_unexpected++;
    }
#endif

    // acknowledge the interrupt
    //
    // machine resets if we don't do this!
    //
    union { HostControl_write hcr_w; Bit32 hcr_w_alias; };
    hcr_w_alias = 0;
    hcr_w.clear_host_int = TRUE;
    memmap->dword_00h = hcr_w_alias;

    // count how many times this occurred
    //
    atm_interrupt_count++;
}



// called by foreground code when it notices that
// BufferQueue::allocate() has failed and it has
// just freed up a buffer
//
// note that both cards are called even when only
// one buffer has been freed up
//
//
//
// also called by foreground code to enable receive
// for the first time if we didn't give the card buffers
// before issuing the initialization command
//
void AtmCard::atm_foreground_replenish( void )
{
    int result;

    // if this card will interrupt us again...
    //
    if( memmap->shared.queues.next_host_block )

        // no need for any extra effort on our part
        //
        return;

    // try to get a BufferBlock
    //
    // when the card sees the next cell, it will grab
    // the block and interrupt us to give it another one
    //
    // note that this will cause another buffer allocation
    // if there isn't really one available, but that is OK
    // since the caller needs the flag set again in order
    // to know to call us again
    //
    result = atm_shift_hw_rings( during_interrupt );
    if( result && !atm_interrupt_error )
        atm_interrupt_error = result;
}



// manages 2 BufferBlock's each of which has 17K buffers;
// both have been given to the card, and in the steady-state
// my_block is being filled in, and next_block is on deck
//
// however, the edge condition, where the card has completed
// filling my_block and started next_block, is when we get
// a host interrupt and this code gets called
//
//
//
// this code does not deal with buffer exhaustion, so if the
// host ever runs out of BufferBlock's, the card will too!
//
// it is also unsafe to call this code when the card has NOT
// swapped its pointers, since
//
int AtmCard::atm_shift_hw_rings( AtmShiftTiming timing )
{
    int result, next_result;

    if( timing == during_initialization )
    {
        result = all_buffers->allocate( &my_block );
        if( result )
            return result;

        my_block->cardnum = pci_handle;
        my_block->cardindex = cardindex;

        memmap->shared.queues.current_host_block = flat_to_physical( my_block->buffers );
    }
    else
    {
        // if the card has any buffers...
        //
        if( my_block )
        {
            // assume all of the buffers were filled
            //
            my_block->used = COMPL_RING_ENTRIES;

            // place my_block on global filled list
            //
            result = all_buffers->fill( my_block );
            if( result != 0 )
                return result;
        }

        // shift the block pointers
        //
        my_block = next_block;
    }

    // allocate another next_block
    //
    next_result = all_buffers->allocate( &next_block );
    if( !next_result )
    {
        // remember which card was expected to fill in the block
        //
        next_block->cardnum = pci_handle;
        next_block->cardindex = cardindex;

        // tell the card about it
        //
        memmap->shared.queues.next_host_block = flat_to_physical( next_block->buffers );
    }

    return next_result;
}



// find the FORE ATM NIC's on the PCI bus
// and initialize them
//
int pci_init( void )
{
    unsigned count;
    Bit8 value;
    Bit16 handle;
    Bit32 address;
    unsigned irq_level;
    int result;

    // could new of all_buffers be here???
    //

#if SAVE_INTERRUPT_THIS_PTR
    printf( "array of pointers to AtmCard is at 0x%p\n", cards );
#endif

#if CHECK_MEMORY_CONSISTENCY
    result = all_buffers->check();
    if( result )
        return result;
#endif

    // check whether PCI BIOS is present
    //
    if( !pci_is_bios_present() )
    {
        printf( "Sorry, PCI BIOS is not present.\n" );
        exit( 1 );
    }

    handle = 0;
    old_pci_latency_timer = pci_get_master_latency_timer( handle );
    printf( "CPU master-latency-timer register was %d (%g uS)\n", old_pci_latency_timer, old_pci_latency_timer*0.25 );

#if INTEL_82439HX
    old_pci_multi_timer = pci_get_multi_transaction_timer( handle );
    printf( "CPU multi-transaction-timer register was %d (%g uS)\n", old_pci_multi_timer, old_pci_multi_timer*0.25 );
#endif

    old_pci_control_alias = pci_get_control( handle );
#if INTEL_82437FX
    printf( "CPU PCI control register was 0x%2.2X (bus concurrency %s,"
        " PCI streaming %s, CPU-to-PCI write bursting %s, peer concurrency %s,"
        " CPU inactivity timer %d)\n",
        old_pci_control_alias,
        old_pci_control.bus_concurrency_disable           ? "disabled" : "enabled",
        old_pci_control.pci_streaming_disable             ? "disabled" : "enabled",
        old_pci_control.cpu_to_pci_write_bursting_disable ? "disabled" : "enabled",
        old_pci_control.peer_concurrency_enable           ? "enabled"  : "disabled",
        old_pci_control.cpu_inactivity_timer + 1
    );
#else
#if INTEL_82439HX
    printf( "CPU PCI control register was 0x%2.2X (global TXC %s,"
        " SERR# output is %s, peer concurrency is %s, NA# %s,"
        " CPU shutdown cycle causes %s, ECC mode is %s, DRAM has %s)\n",
        old_pci_control_alias,
        old_pci_control.global_txc_enable                 ? "enabled"  : "disabled",
        old_pci_control.serr_pound_output_type            ? "actively driven" : "open drain for PCI",
        old_pci_control.peer_concurrency_enable           ? "enabled"  : "disabled",
        old_pci_control.dual_processor_na_pound_enable    ? "uses policies for dual-procesor PCI 2.1" : "uses 82437FX policies",
        old_pci_control.shutdown_to_port_92               ? "write 01h to port 92" : "shutdown cycle on PCI bus",
        old_pci_control.ecc_test_enable                   ? "test" : "normal",
        old_pci_control.dram_ecc_or_parity_select         ? "ECC" : "parity"
    );
#endif
#endif

    old_pci_cache_control_alias = pci_get_cache_control( handle );
    printf( "CPU cache control register was 0x%2.2X (first level cache %s,"
        " secondary cache miss/invalidate %sforced,"
#if INTEL_82439HX
        " L2 cache up to %s,"
#endif
        " NA# signal %s, SRAM type %d, secondary cache size %dK)\n",
        old_pci_cache_control_alias,
        old_pci_cache_control.first_level_cache_enable ? "enabled" : "disabled",
        old_pci_cache_control.secondary_cache_force_miss_or_invalidate ? "is " : "not ",
#if INTEL_82439HX
        old_pci_cache_control.extended_cachability_enable ? "512 MB" : "64 MB",
#endif
        old_pci_cache_control.na_pound_signal_disable ? "disabled" : "enabled",
        old_pci_cache_control.static_ram_type,
        old_pci_cache_control.secondary_cache_size * 256
    );

#if PCI_INIT_SETS_PCI_MULTI_TIMING
    pci_set_multi_transaction_timer( handle, DESIRED_MULTI_LATENCY );
    value = pci_get_multi_transaction_timer( handle );
    if( value != DESIRED_MULTI_LATENCY )
        printf( "unable to set CPU multi-transaction-timer register to %d; is now %d (%g uS)\n", DESIRED_MULTI_LATENCY, value, value*0.25 );
    else
        printf( "changed CPU multi-transaction-timer register to %d (%g uS)\n", DESIRED_MULTI_LATENCY, value*0.25 );
#endif

#if PCI_INIT_SETS_PCI_TIMING
    pci_set_master_latency_timer( handle, DESIRED_HOST_LATENCY );
    value = pci_get_master_latency_timer( handle );
    if( value != DESIRED_HOST_LATENCY )
        printf( "unable to set CPU master-latency-timer register to %d; is now %d (%g uS)\n", DESIRED_HOST_LATENCY, value, value*0.25 );
    else
        printf( "changed CPU master-latency-timer register to %d (%g uS)\n", DESIRED_HOST_LATENCY, value*0.25 );
#endif

#if PCI_INIT_SETS_PCI_CONTROL
    pci_set_control( handle, DESIRED_HOST_CONTROL );
    value = pci_get_control( handle );
    if( value != DESIRED_HOST_CONTROL )
        printf( "unable to set CPU control register to 0x%2.2X; is now 0x%2.2X\n", DESIRED_HOST_CONTROL, value );
    else
        printf( "changed CPU control register to 0x%2.2X\n", value );
#endif

#if PCI_INIT_SETS_CACHING
    pci_set_cache_control( handle, DESIRED_CACHE_CONTROL );
    value = pci_get_cache_control( handle );
    if( value != DESIRED_CACHE_CONTROL )
        printf( "unable to set CPU cache control register to 0x%2.2X; is now 0x%2.2X\n", DESIRED_CACHE_CONTROL, value );
    else
        printf( "changed CPU cache control register to 0x%2.2X\n", value );
#endif

    printf( "searching for FORE PCA200E ATM network interface cards\n"
        "using vendor ID 0x%4.4X and device ID 0x%4.4X\n",
        FORE_VENDOR_ID, FORE_PCA200E_DEVICE_ID
    );

    // for all the ATM cards that are on the PCI bus...
    //
    for( count=0; count<1000; count++ )
    {
        handle = pci_get_device_handle( FORE_PCA200E_DEVICE_ID,
            FORE_VENDOR_ID, count );
        if( RETCODE_NO_MORE_PCI_DEVICES == handle )
            break;

        printf( "handle 0x%4.4X found\n", handle );

        address = pci_get_base_address( handle );
        irq_level = pci_get_irq_level( handle );

        printf( "device has physical memory address 0x%8.8lX and IRQ level %d\n", address, irq_level );

        if( count>=ATM_NIC_OFFSET && count<ATM_NIC_OFFSET+MAX_ATM_NICS )
        {
            cards[ count - ATM_NIC_OFFSET ] = new AtmCard( handle, address, irq_level, count-ATM_NIC_OFFSET );
            result = cards[ count - ATM_NIC_OFFSET ]->get_init_result();
            if( result )
                return result;

#if SAVE_INTERRUPT_THIS_PTR
            printf( "instantiated AtmCard class at address 0x%p\n", cards[count-ATM_NIC_OFFSET] );
#endif
#if PCI_INIT_STOPS_EARLY
            count = ATM_NIC_OFFSET + MAX_ATM_NICS;
            break;
#endif
        }
    }

    if( count - ATM_NIC_OFFSET > MAX_ATM_NICS )
        printf( "warning: %d devices were found, but only %d will be used\n", count, MAX_ATM_NICS );
    else if( count < MAX_ATM_NICS )
    {
        printf( "error: only %d of expected %d devices were found!\n", count, MAX_ATM_NICS );
        return ERROR_PCI_TOO_FEW_CARDS_FOUND;
    }
    else
        printf( "success!  exactly %d FORE PCA200E ATM NIC's were found\n", MAX_ATM_NICS );

    printf( "PCI configuration discovery done\n\n" );

    return 0;
}



// shut down whatever cards we were using
//
int pci_exit( void )
{
    int count;
    Bit32 block_count;
    Bit32 buffer_count;
    int result;
    AtmCard *this_card;

    for( count=MAX_ATM_NICS-1; count>=0; count-- )
    {
        // we must set global pointer to nil before we
        // deallocate the card, because otherwise the next
        // timer interrupt will try to use the object after
        // it has been deallocated and possibly reused,
        // which is unsafe for MANY reasons
        //
        this_card = cards[ count ];
        cards[ count ] = NULL;
        delete this_card;

        // separate messages for one card from the other
        //
        printf( "\n" );
    }

    union { PciStatus status; Bit8 status_alias; };
    status_alias = pci_get_status( 0 );
    printf( "host PCI final status 0x%4.4X\n", status_alias );

#if CHECK_MEMORY_CONSISTENCY
    // don't want to exit when this fails because the
    // diagnostics that follow may still be useful
    //
    all_buffers->check();
#endif

    printf( "BufferQueue::allocate() failed %d times\n", all_buffers->allocate_failure_count );

    // dump the host buffer queue states
    //
    printf( "global host buffer queue has %ld free buffers\n", all_buffers->amount_free() );
    buffer_count = all_buffers->amount_filled( block_count );
    printf( "global host buffer queue has %ld/%ld filled buffers/blocks\n", buffer_count, block_count );

#if PCI_EXIT_CALLS_BUFFER_DUMP
    // dump the data in the global queue to disk
    //
    char filename[ 13 ];

    sprintf( filename, "D:\\IP.DMP" );
    printf( "pci_exit() saving IP packets with headers in file '%s'\n", filename );
    result = all_buffers->dump( filename );
    if( result != 0 )
    {
        printf( "error %d dumping queue\n", result );
    }
#endif

#if PCI_INIT_SETS_PCI_MULTI_TIMING || PCI_INIT_SETS_PCI_TIMING || PCI_INIT_SETS_PCI_CONTROL || PCI_INIT_SETS_CACHING
    Bit16 handle = 0;
    Bit8 value;
#endif

#if PCI_INIT_SETS_PCI_MULTI_TIMING
    // reset the global multi transaction timer
    //
    pci_set_multi_transaction_timer( handle, old_pci_multi_timer );
    value = pci_get_multi_transaction_timer( handle );
    if( value != old_pci_multi_timer )
        printf( "unable to restore CPU multi-transaction-timer register to %d; is now %d (%g uS)\n", old_pci_multi_timer, value, value*0.25 );
    else
        printf( "restored CPU multi-transaction-timer register to %d (%g uS)\n", old_pci_multi_timer, value*0.25 );
#endif

#if PCI_INIT_SETS_PCI_CONTROL
    // reset the control register for host
    //
    pci_set_control( handle, old_pci_control_alias );
    value = pci_get_control( handle );
    if( value != old_pci_control_alias )
        printf( "unable to restore CPU control register to 0x%2.2X; is now 0x%2.2X\n", old_pci_control_alias, value );
    else
        printf( "restored CPU control register to 0x%2.2X\n", old_pci_control_alias );
#endif

#if PCI_INIT_SETS_PCI_TIMING
    // reset the master latency timer for host
    //
    pci_set_master_latency_timer( handle, old_pci_latency_timer );
    value = pci_get_master_latency_timer( handle );
    if( value != old_pci_latency_timer )
        printf( "unable to restore CPU master-latency-timer register to %d; is now %d (%g uS)\n", old_pci_latency_timer, value, value*0.25 );
    else
        printf( "restored CPU master-latency-timer register to %d (%g uS)\n", old_pci_latency_timer, value*0.25 );

    // reset the control register for host
    //
    pci_set_control( handle, old_pci_control_alias );
    value = pci_get_control( handle );
    if( value != old_pci_control_alias )
        printf( "unable to restore CPU control register to 0x%2.2X; is now 0x%2.2X\n", old_pci_control_alias, value );
    else
        printf( "restored CPU control register to 0x%2.2X\n", old_pci_control_alias );
#endif

#if PCI_INIT_SETS_CACHING
    // reset the cache control register for host
    //
    // note: it is very important not to restore the caching
    // settings until we are done messing with high memory
    // (i.e. dumping it to disk) because we might lock up
    // the machine otherwise!
    //
    pci_set_cache_control( handle, old_pci_cache_control_alias );
    value = pci_get_cache_control( handle );
    if( value != old_pci_cache_control_alias )
        printf( "unable to restore CPU cache control register to 0x%2.2X; is now 0x%2.2X\n", old_pci_cache_control_alias, value );
    else
        printf( "restored CPU cache control register to 0x%2.2X\n", old_pci_cache_control_alias );
#endif

    // delete of all_buffers should be here
    //

    return result;
}



// turn off compiler warning about unused parameters
// since I know they won't be used and I must have
// them because calling code cannot be changed
//
// P.S. I have no clue why I don't have to do this
// for class static functions AtmCard::handler1()
// and AtmCard::handler2()
//
#pragma warn -par

// deals with interrupts from hardware timer chip
// on Pentium PC motherboard
//
// should occur at least often enough to poll the SAR
// so the receive buffers do not overflow
//
int timer_handler(
    Bit32 ebp,
    Bit32 edi,
    Bit32 esi,
    Bit32 ds,
    Bit32 es,
    Bit32 edx,
    Bit32 ecx,
    Bit32 ebx,
    Bit32 eax,
    Bit32 eip,
    Bit32 cs,
    Bit32 flags
)
{
    // count the number of interrupts for debugging
    //
    timer_interrupt_count++;

    // I need help calibrating the pentium timestamps,
    // so record 2 timestamps, one interrupt apart
    //
    if( timer_interrupt_count == TIMER_CALIBRATE )
        timer_start_time.set();
    else if( timer_interrupt_count == TIMER_CALIBRATE + 1 )
        timer_end_time.set();

#if DEBUG_STACK_OVERFLOW
    if( !stack_ptr_recorded )
    {
        stack_ptr_during_global_timer = _ESP;
        stack_seg_during_global_timer = _SS;
    }
#endif

    // any polling of SAR chips should occur here
    //
#if ONE_CARD_PER_TIMER_IRQ
    // note that we only call one card per interrupt
    //
    if( cards[ timer_interrupt_card_index ] )
        cards[ timer_interrupt_card_index ]->timer_handler();

    timer_interrupt_card_index++;
    if( timer_interrupt_card_index >= MAX_ATM_NICS )
        timer_interrupt_card_index = 0;
#else
    if( cards[0] )
        cards[0]->timer_handler();

#if MAX_ATM_NICS >= 2
    if( cards[1] )
        cards[1]->timer_handler();
#endif
#endif

    // tell the interrupt controllers we are done
    //
    irq->acknowledge( timer_irq_level );

    // tell the thunking code not to chain to the old handler
    //
    return 0;
}

// turn compiler warning about unused parameters
// back to whatever it was before I messed with it
//
#pragma warn .par




void timer_init( void )
{
    // calculate the software interrupt numbers
    // that the CPU will see for the IRQ levels
    //
    timer_interrupt = irq_to_intr( timer_irq_level );

    // disable maskable interrupts
    //
    disable();

    // save old IRQ masks
    //
    irq = new InterruptController;

    // save the old interrupt vectors
    //
    old_timer_handler = getvect( timer_interrupt );

#if AFFECT_TIMER_INTERRUPTS
    // create a thunk so our C++ function can be called at interrupt time
    //
    timer_thunk = make_thunk( timer_handler, &old_timer_handler );
#if THROW_IS_LEGAL
    if( timer_thunk == NULL )
        throw "error!  cannot allocate timer interrupt thunk!\n";
#endif

    // point the vector at our handler
    //
    Bit48 pointer = { (Bit32)timer_thunk, _CS };
    setvect( timer_interrupt, pointer );

    // allow desired IRQ level through to the processor
    //
#if ALLOW_TIMER_INTERRUPTS
    irq->unmask( timer_irq_level );
#else
    irq->mask( timer_irq_level );
#endif

    // initialize hardware timer 0
    //
    timer = new HardwareTimer;

    // set the value that hardware timer 0
    // will decrement until interrupt
    //
    timer->set( TIMER_FREQUENCY );
#endif

    // enable maskable interrupts
    //
    enable();
}



// restore interrupt handling to the way it was when
// we started up
//
void timer_exit( void )
{
    // disable maskable interrupts
    //
    disable();

#if AFFECT_TIMER_INTERRUPTS
    // reset the clock frequency to the one we think BIOS likes
    //
    timer->set( 65536 );

    // stop using hardware timer 0
    //
    delete timer;

    // point the vectors at the old handlers
    //
    // must be before we delete irq or we will
    // see an extraneous timer interrupt
    //
    // I know, I know, interrupts are supposed
    // to be disabled, but I'm not lying!
    //
    setvect( timer_interrupt, old_timer_handler );

    // free up the 32-bit assembly thunking code
    //
    free( timer_thunk );
#endif

    // restore old IRQ masking by destroying InterruptController
    //
    delete irq;

    // enable maskable interrupts
    //
    enable();

    // print out what meager statistics we
    // have for the hardware timer chip
    //
    printf( "hardware timer 0 caused %ld interrupts\n", timer_interrupt_count );
    double elapsed = timer_end_time - timer_start_time;
    printf( "%g pentium clocks elapsed between hardware timer 0 interrupts %d and %d\n",
        elapsed, TIMER_CALIBRATE, TIMER_CALIBRATE+1 );
}



#if DO_FOREGROUND_STATS
// do stats analysis on the partially-received blocks
//
void analyze_partial_blocks( void )
{
    unsigned count;

    for( count=0; count<MAX_ATM_NICS; count++ )

        // beware of NULL block pointer in AtmCard
        // which indicates that block allocation
        // failed last time around
        //
        if( cards[count]->my_block )
        {
            cards[count]->set_pending_block_usage();
            flows_per_block( cards[count]->my_block );
        }
}
#endif



int main( int argc, char *argv[] )
{
    int keystroke;
    int result;
    int read_result = 0;
    Bit32 count;
#if DO_FOREGROUND_STATS
    BufferBlock *block;
    int alloc_has_failed;
    unsigned block_count = 0;
    Bit64 buffer_count = (Bit32) 0;
    PentiumClock before, after;
    PentiumClock startup_time;
    PentiumClock active_interval, active_alarm;
    PentiumClock query_interval, query_alarm;
    double elapsed2;
    char query_msg[ 2000 ] = "";
    int query_msg_len = 0;
#if TIME_STATS_CALLS
    PentiumClock elapsed, elapsed_min, elapsed_max, elapsed_tot;
    double elapsed_min2, elapsed_max2, elapsed_avg2;
    elapsed_min.high = -1;
    elapsed_min.low  = -1;
    elapsed_max.high = 0;
    elapsed_max.low  = 0;
    elapsed_tot.high = 0;
    elapsed_tot.low  = 0;
#endif
#if DO_FULLNESS_STATS
    Bit32 max_filled, min_filled, sum_filled, count_filled;
#endif
#if !ALLOW_STATS_CALLS && BUSYWAIT_STATS_CALLS
    Bit32 busy_iterations = 0;
#endif
#endif
    PentiumClock now;
    PentiumClock keystroke_interval, keystroke_alarm;
    FpuControl float_control;
    union { FpuStatus float_status; Bit16 float_status_alias; };
    Bit8 *data;



    float_control = get_fpu_control();

    // force floating point division by zero to generate
    // negative or positive infinity, depending on the
    // sign of the numerator, instead of interrupt 0x10
    //
    // this is the same way DEC Alpha with GNU C behaves
    //
    float_control.zero_divide_mask = TRUE;

    // force conversion of a floating point infinity to
    // integer to succeed, instead of ?interrupt 0x10
    //
    // this is the behavior of DELL and Aberdeen PC's
    // since they have a null handler for that exception
    //
    float_control.invalid_operation_mask = TRUE;

    set_fpu_control( float_control );

#if TEST_DIVISION
    // try dividing a 32 bit unsigned int by float zero
    //
    Bit32 divtest_numerator = 144765;
    double divtest_denominator = 0.0;
    double divtest_result = divtest_numerator / divtest_denominator;
    printf( "Bit32 %u / double %f = double %f\n",
        divtest_numerator,
        divtest_denominator,
        divtest_result
    );

    // now try converting the result to an integer
    //
    Bit32 divtest_result_integer = divtest_result;
    printf( "double %f converts to Bit32 %u\n", divtest_result, divtest_result_integer );

    // now try dividing float zero by float zero
    //
    double divtest_numerator_float = 0.0;
    divtest_result = divtest_numerator_float / divtest_denominator;
    printf( "double %f / double %f = double %f\n",
        divtest_numerator_float,
        divtest_denominator,
        divtest_result
    );

    // now try converting the new result to an integer
    //
    divtest_result_integer = divtest_result;
    printf( "double %f converts to Bit32 %u\n", divtest_result, divtest_result_integer );

#if 0       // don't bother, it always fails
    // now try dividing a 32 bit unsigned int by an integer zero
    //
    Bit32 divtest_denominator_integer = 0;
    divtest_result_integer = divtest_numerator / divtest_denominator_integer;
    printf( "Bit32 %u / Bit32 %u = Bit32 %u\n",
        divtest_numerator,
        divtest_denominator_integer,
        divtest_result_integer
    );
#endif

    // now clear out all the floating point exceptions
    // so main() doesn't complain about them
    //
    clear_fpu_status();
#endif



    // initialize access to our own area of flat memory
    // which we know how to convert to physical addresses
    //
    // note that DOS 6.22 HIMEM.SYS cannot access more than
    // 64 MB of physical memory, although WIN 95 HIMEM.SYS
    // can do at least 128 MB
    //
    // note that DPMI32VM.OVL included with Borland PowerPack
    // 1.0 for DOS truncates the value of DPMIMEM=MAXMEM 70000
    // to 16 bits, and hence will not try to allocate more than
    // 64 MB minus 1 KB of memory, although you can patch file
    // offset 35E0h from 2B 1E 66 06 to 90 90 90 90 so it will
    // allocate until memory is exhausted (ignoring environment
    // variables)
    //
    flat_init();

#if LOCK_PAGES
    // try to lock the whole program
    //
    Bit16 lock_result = lock_linear( flat_to_linear(&atm_shift_dummy_before), 0x3000 );
    printf( "got error 0x%4.4X locking pages of 0x%8.8lX AtmCard::atm_shift_hw_rings()\n", lock_result, atm_shift_dummy_before );
#endif

#if DEBUG_STACK_OVERFLOW
    // say what's up with the stack
    //
    printf( "stack is at log/lin 0x%4.4X:%8.8lX 0x%8.8lX\n",
        _SS, _ESP, flat_to_linear((void *)_ESP) );
    printf( "stack segment has base 0x%8.8lX and limit 0x%8.8lX\n",
        get_base(_SS), get_limit(_SS) );

    if( !stack_ptr_recorded )
    {
        stack_ptr_during_main_top = _ESP;
        stack_seg_during_main_top = _SS;
    }
#endif



    // init TCP/IP library
    //
    // we must do this before any other physical memory
    // function calls because those must be undone in the
    // same order that they were done, and sock_init()
    // installs an atexit() handler which means we are not
    // in control of when sock_exit() gets called
    //
    sock_init();



    // keyboard will still be checked when foreground stats
    // are not being done, so default its alarm to once a second
    //
    keystroke_interval.put_seconds( 1.0 );

    // just get the time once for the many interval calculations
    //
    now.set();

    // in case flows module doesn't fill in the variables, also
    // default query and active feed intervals to never
    //
#if DO_FOREGROUND_STATS
    query_interval.high = -1;
    query_interval.low = -1;
    query_interval -= now;
    active_interval.high = -1;
    active_interval.low = -1;
    active_interval -= now;

    // tell the stats code about the command-line arguments
    // and give it a chance to alter the timeouts
    //
    fprintf( stderr, "%u bytes of RAM initially available to malloc()\n", coreleft() );
    fprintf( stderr, "%u bytes of RAM initially available to flat_alloc()\n", flat_remaining() );
    result = flows_init( argc, argv, &keystroke_interval, &query_interval, &active_interval );
    fprintf( stderr, "%u bytes of RAM are still available to malloc()\n", coreleft() );
    fprintf( stderr, "%u bytes of RAM are still available to flat_alloc()\n", flat_remaining() );
    if( result )
        return result;

    // now set up initial values for all the alarms
    //
    query_alarm = now + query_interval;
    active_alarm = now + active_interval;
#endif
    keystroke_alarm = now + keystroke_interval;

    // allocate all remaining flat memory to card buffer pool
    //
    // we must do this AFTER statistics initialize, so they
    // get first crack at flat memory, in case they are using
    // it instead of malloc()
    //
    all_buffers = new BufferQueue;
    if( all_buffers == NULL )
        goto cleanup;

    // say where the global buffers are
    //
    printf( "global control blocks are at 0x%p\n", &all_buffers );

    data = all_buffers->get_start()->buffers;
    printf( "global data buffers are at   log/lin/phy"
        " 0x%4.4X:%8.8lX 0x%8.8lX 0x%8.8lX\n", _DS, data,
        flat_to_linear(data), flat_to_physical(data) );

    printf( "SW_RINGS=%d\n", SW_RINGS );

    printf( "COMPL_RING_ENTRIES=%d\n", COMPL_RING_ENTRIES );

#if DO_FOREGROUND_STATS && DO_FULLNESS_STATS
    // initialize BufferQueue fullness statistics
    //
    max_filled = 0;
    min_filled = BIT32_MAX;
    sum_filled = 0;
    count_filled = 0;
#endif

    // initialize the h/w timer and interrupt controller chips
    //
    timer_init();

    // get the addresses and interrupts of
    // the PCI ATM cards
    //
    result = pci_init();
    if( result )
    {
        printf( "error %d from pci_init()\n", result );
        goto cleanup;
    }

    // synchronize the clocks on the cards by issuing the
    // "clear your timestamp" command to them all in rapid
    // sequence, then checking results after
    //
    // assumes all card pointers in array have been filled
    // in, which is guaranteed by pci_init()
    //
    // assumes functions called do their own error messages
    //
    CardClock::sync_time.set();
    for( count=0; count<MAX_ATM_NICS; count++ )
    {
        result = cards[ count ]->atm_clear_timestamp_command();
        if( result )
            goto cleanup;
    }
    for( count=0; count<MAX_ATM_NICS; count++ )
    {
        result = cards[ count ]->atm_await_command( UART_DELAY_COMMAND, TIME_CLEAR_DELAY );
        if( result )
            goto cleanup;
    }

#if !ATM_INIT_PROVIDES_BUFFERS
    // give one BufferBlock to each card,
    // to allow them to finally start receive
    //
    // we avoided it until now so init would work
    // even if the link is very busy
    //
    for( count=0; count<MAX_ATM_NICS; count++ )
    {
        cards[ count ]->atm_foreground_replenish();
    }
#endif

    // note: we cannot flush the level 1 cache because
    // the Borland PowerPack for DOS 32-bit protected
    // mode environment puts us in ring 3, so executing
    // the WBINVD instruction would cause a protection
    // violation
    //
    // we cannot fill the code cache either, because the
    // Pentium processor maintains separate data and code
    // caches, so reads of the code segment would only
    // affect the data cache
    //

    printf( "\n" );


#if DO_FOREGROUND_STATS && SIMULATE_FIBER
    // create a dummy block that has all invalid buffers
    //
    result = all_buffers->allocate( &block );
    if( result )
    {
        printf( "fatal error: main() has error %d allocating a dummy block\n", result );
        return result;
    }

    block->cardindex = 0;
    block->cardnum = cards[ 0 ]->pci_handle;
    block->used = COMPL_RING_ENTRIES;
    {
        PentiumClock now;
        now.set();
        CardClock now2 = now;
        block->timestamp() = now2;
    }

    for( count=0; count<COMPL_RING_ENTRIES; count++ )
    {
        // fudge a nonzero IP fragment offset offset
        // so flows_per_block() will ignore the data
        //
        *(block->buffers + count*BUFFER_SIZE + 12 + 11) = 1;
    }
#endif

#if DO_FOREGROUND_STATS
    // remember when we started collecting stats
    //
    startup_time.set();
#endif

    // wait for the user to unplug fiber(s)
    // to stop SAR from filling its memory
    // while we empty it
    //
    printf( "\nplease reconnect fibers\npress escape when you want to stop capturing\n" );
    while( TRUE )
    {
#if DEBUG_STACK_OVERFLOW
        stack_ptr_during_main_loop = _ESP;
        stack_seg_during_main_loop = _SS;
#endif

#if CHECK_MEMORY_WHILE_LOOPING
        result = all_buffers->check();
        //if( result )
        //    goto cleanup;
#endif

        // service the soft-UARTs, in case the card has any
        // diagnostic output for us
        //
        for( count=0; count<MAX_ATM_NICS; count++ )
            if( cards[count] )
                cards[count]->atm_uart_io( NULL, NULL, 0, UART_DELAY_RUNNING );

        // if it is time to check for a keystroke...
        //
        now.set();
        if( now >= keystroke_alarm )
        {
            keystroke_alarm = now + keystroke_interval;

            // process the whole keyboard buffer
            //
            while( kbhit() )
            {
                keystroke = getch();
                if( keystroke == 27 )
                    break;

#if DO_FOREGROUND_STATS
                flows_keystroke( keystroke );
#endif
            }

            if( keystroke == 27 )
                break;
        }

#if !ALLOW_TIMER_INTERRUPTS
#if 0
        // poll the RX free and RX compl rings from foreground
        //
        unsigned count2, count;
        for( count2=0; count2<100; count2++ )
            for( count=0; count<MAX_ATM_NICS; count++ )
                if( cards[count] )
                    cards[count]->timer_handler();
#endif
#endif

#if !DO_FOREGROUND_STATS
        // if all the cards have run out of memory...
        //
        if( all_buffers->allocate_failure_count >= MAX_ATM_NICS )

            // stop the program, since nobody
            // will ever free up memory
            //
            break;
#else
#if SIMULATE_FIBER
        block->queried = 0;
#else
#if DO_FULLNESS_STATS
        // update BufferQueue fullness statistics
        //
        count_filled++;
        all_buffers->amount_filled( count );
        sum_filled += count;
        if( max_filled < count )
            max_filled = count;
        if( min_filled > count )
            min_filled = count;
#endif

        // try to get a buffer block
        //
        read_result = all_buffers->read( &block );

        // if we got one...
        //
        if( !read_result )
#endif  // !SIMULATE_FIBER
        {
            block_count++;
            buffer_count += block->used;

#if TRACE_BUFFER_BLOCKS
            printf( "got BufferBlock pointing at 0x%.8X/0x%.8X log/phys\n",
                block->buffers, flat_to_physical(block->buffers) );
            if( 0 && block->buffers == block_trace )
            {
                printf( "saving block to disk..." );
                all_buffers->dump_open( "d:\\badblock.dmp" );
                all_buffers->dump_one( block );
                all_buffers->dump_close();
                printf( " done\n" );
            }
#endif

#if TIME_STATS_CALLS
            before.set();
#endif

#if ALLOW_STATS_CALLS
            // do statistical analysis on it
            //
            flows_per_block( block );
#else
#if BUSYWAIT_STATS_CALLS
            // busy-wait to simulate foreground CPU usage
            //
            busy_iterations += PentiumClock::wait( (CLOCK_RATE / 1000000) * FLOW_BLOCK_TIME );
#endif
#endif  // ALLOW_STATS_CALLS

#if TIME_STATS_CALLS
            after.set();
#if 0
            elapsed = after - before;
#else
            elapsed = after;
            elapsed -= before;
#endif
            elapsed_tot += elapsed;
            if( elapsed > elapsed_max )
                elapsed_max = elapsed;
            if( elapsed < elapsed_min )
                elapsed_min = elapsed;
#endif  // TIME_STATS_CALLS

#if !SIMULATE_FIBER
            // must do this before freeing
            //
            alloc_has_failed = all_buffers->has_allocate_failed_lately();

            // recycle the buffer
            //
            result = all_buffers->free( block );
            if( result )
            {
                printf( "error %d freeing buffer block 0x%8.8lX\n", result, block );
                break;
            }

            // if buffer allocation failed lately...
            //
            if( alloc_has_failed )

                // for all cards...
                //
                for( count=0; count<MAX_ATM_NICS; count++ )

                    // since cards that fail allocation do not get
                    // further interrupts, give him a shot at this
                    // block
                    //
                    cards[count]->atm_foreground_replenish();
#endif  // !SIMULATE_FIBER
        }

        // if it is time to dump active flows...
        //
        now.set();
        if( now >= active_alarm )
        {
            active_alarm = now + active_interval;

            analyze_partial_blocks();

            /* klt - 10/31/96
             * rifle thru the active list
             * and send it all and delete the
             * queue - this is single-threaded, right?
            printf("DUMPING ACTIVE LIST in main()...\n");
             */
            flows_send_active();
        }

        /* if it is time to check for a query...
         */
        now.set();
        if( now > query_alarm )
        {
            query_alarm = now + query_interval;

            if( flows_is_query_wanted() )
            {
#if ALLOW_STATS_CALLS_2
                // if we are all caught up (the filled list was
                // empty the last time we tried to get a block)...
                //
                if( read_result == ERROR_QUEUE_NOT_FILLED )
                    analyze_partial_blocks();
#endif

#if DO_FULLNESS_STATS
                // display and clear BufferQueue fullness stats
                //
                sprintf( query_msg+query_msg_len, "BufferQueue fullness %u min, %u max, %f avg over %u iterations\n",
                    min_filled, max_filled, (double)sum_filled/count_filled, count_filled );
                max_filled = 0;
                min_filled = BIT32_MAX;
                sum_filled = 0;
                count_filled = 0;
#endif

#if TIME_STATS_CALLS_3
                before.set();
#endif

#if ALLOW_STATS_CALLS_3
                // do summary analysis
                //
                result = flows_query( query_msg );
                if( result )
                {
                    printf( "error %d querying stats\n", result );
                    break;
                }

#else
#if BUSYWAIT_STATS_CALLS_3
                // busy-wait to simulate foreground CPU usage
                //
                PentiumClock::wait( (CLOCK_RATE / 1000) * FLOW_QUERY_TIME );
#endif

                extern unsigned query_count;
                query_count++;
    //          if( query_count % 100 )
    //              printf( " %d", query_count );
    //          else
    //              printf( "capture not doing query %d block %d\n", query_count, block_count );
#endif  // !ALLOW_STATS_CALLS

#if TIME_STATS_CALLS_3
                after.set();
                elapsed2 = PentiumClock( after - before ).get_seconds() * 1000;
                sprintf( query_msg, "previous flows_query() elapsed time was %g milliseconds\n%n", elapsed2, &query_msg_len );
                printf( query_msg );
#else
                query_msg[0] = '\0';
                query_msg_len = 0;
#endif
            }   // end if query wanted
        }   // end if query alarm expired
#endif  // DO_FOREGROUND_STATS

    }   // end while forever

#if DO_FOREGROUND_STATS
    // calculate how long were were collecting stats
    //
    after.set();
    after -= startup_time;
    printf( "%.2f seconds runtime\n", after.get_seconds() );

    printf( "%.0f buffers and %d blocks seen\n", (double)buffer_count, block_count );
#endif


#if DO_FOREGROUND_STATS && (TIME_STATS_CALLS || BUSYWAIT_STATS_CALLS)
    if( block_count )
    {
#if !ALLOW_STATS_CALLS && BUSYWAIT_STATS_CALLS
        printf( "%.4g avg iterations busywaiting %d uS instead of flows_per_block()\n", (double)busy_iterations/block_count, FLOW_BLOCK_TIME );
#endif
#if TIME_STATS_CALLS
        elapsed_min2 = ((double)elapsed_min * 1000) / CLOCK_RATE;
        elapsed_max2 = ((double)elapsed_max * 1000) / CLOCK_RATE;
        elapsed_avg2 = (((double)elapsed_tot / block_count) * 1000) / CLOCK_RATE;
        printf( "flows_per_block() took min %.4f max %.4f avg %.4f mS\n", elapsed_min2, elapsed_max2, elapsed_avg2 );
        printf( "%.2f seconds (%.2f%%) spent in flows_per_block() during run\n", elapsed_tot.get_seconds(), 100.0*(double)elapsed_tot/(double)after );
#endif
    }
    else
        printf( "flows_per_block() was never called\n" );
#endif  // DO_FOREGROUND_STATS && etc.



    printf( "final estimate of card clock rate: %.2f per second\n", CARD_CLOCK_RATE );

#if DEBUG_STACK_OVERFLOW
    printf( "\nstack pointer at top of main() was 0x%4.4X:%8.8lX\n", stack_seg_during_main_top, stack_ptr_during_main_top );
    printf( "stack pointer inside last loop of main() was 0x%4.4X:%8.8lX\n", stack_seg_during_main_loop, stack_ptr_during_main_loop );
    printf( "stack pointer inside first timer_handler() was 0x%4.4X:%8.8lX\n", stack_seg_during_global_timer, stack_ptr_during_global_timer );
    printf( "stack pointer inside first AtmCard::timer_handler() was 0x%4.4X:%8.8lX\n", stack_seg_during_class_timer, stack_ptr_during_class_timer );
    printf( "stack pointer inside first AtmCard::atm_shift_hw_rings() was 0x%4.4X:%8.8lX\n\n", stack_seg_during_atm_shift, stack_ptr_during_atm_shift );
#endif

#if CHECK_MEMORY_WHILE_RUNNING
    all_buffers->check();
#endif

#if DO_FOREGROUND_STATS
    // clean up stats code and report
    //
    result = flows_cleanup();
#endif

cleanup:
    // shut down the cards
    //
    pci_exit();

    // restore the original interrupt routing
    //
    timer_exit();

    // ???maybe force the CMOS clock to be read
    // into the BIOS s/w based real time clock
    //

    return result;
}


