/* conditional compile switches all go here
 */
#define DEBUG_ACTIVE_FEED       FALSE



/*
 * routines for doubly-linked lists
 */

#include <stdio.h>
#include <assert.h>
#include "queue.h"

struct queue Q;

/* klt - 10/31/96 - new Active list queue */
struct queue A;

/*
 * insque
 *  insert an already-allocated queue_entry at the
 *  end of the queue
 */

void insque( struct hashrec *qe )
{
    assert(qe != NULL);

    Q.num_entries++;

    /* empty queue */
    if (Q.tail == NULL) {
        assert(Q.head == NULL);
        Q.tail = qe;
        Q.tail->hr_qlink.next = NULL;
        Q.tail->hr_qlink.prev = NULL;
        Q.head = Q.tail;
        return;
    }

    Q.tail->hr_qlink.next = qe;
    qe->hr_qlink.next = NULL;
    qe->hr_qlink.prev = Q.tail;
    Q.tail = qe;
}

/*
 * unlink an entry from the queue, don't delete it
 */

void delque( struct hashrec *qe )
{
    assert(qe != NULL && Q.head != NULL);

    Q.num_entries--;

    /* 1-node */
    if (Q.head == Q.tail) {
        Q.head = Q.tail = NULL;
        return;
    }

    if (qe->hr_qlink.prev != NULL)
        qe->hr_qlink.prev->hr_qlink.next = qe->hr_qlink.next;
    else
        Q.head = qe->hr_qlink.next;

    if (qe->hr_qlink.next != NULL)
        qe->hr_qlink.next->hr_qlink.prev = qe->hr_qlink.prev;
    else
        Q.tail = qe->hr_qlink.prev;
}

/*
 * end queue: move an existing queue entry to the end
 * <this happens when the access time has just been touched>
 */

void endque( struct hashrec *qe )
{
    delque(qe);
    insque(qe);
}

/*
 * dumpq()
 *  diagnostic
 */

void dumpq()
{
    register struct hashrec *qe = Q.head;
    int i = 0;

    fprintf(stderr, "# queue entries: %d\n", Q.num_entries);
    while (qe != NULL) {
        fprintf(stderr, "entry[%d]@0x%x\n", i++, qe);
        dumphashrec( qe );
        qe = qe->hr_qlink.next;
    }
}

/*
 * klt - 10/31/96
 * update_active_queue
 *
 */

/* 1st implement as an insqueue as above
 * once that works mod it so it looks for
 * the flow on the queue 1st and if it
 * finds it, update the info instead of
 * adding another queue element for same flow
 */

/* should add a queue param to joel's calls to eliminate
 * this redundancy
 */

void update_active_queue( struct hashrec *qe )
{
#if DEBUG_ACTIVE_FEED
    fprintf(stderr,"in update_active_queue\n");
#endif
    assert(qe != NULL);

    /* if already on list, do nothing
     */
    if( qe->hr_alink.next || qe->hr_alink.prev ) {
#if DEBUG_ACTIVE_FEED
        fprintf(stderr,"entry is already on active queue - do nothing\n");
#endif
        return;
    }

    A.num_entries++;

    /* empty queue */
    if (A.tail == NULL) {
#if DEBUG_ACTIVE_FEED
        fprintf(stderr,"update_active_queue: empty queue\n");
#endif
        assert(A.head == NULL);
        A.tail = qe;
        A.tail->hr_alink.next = NULL;
        A.tail->hr_alink.prev = NULL;
        A.head = A.tail;
        return;
    }

    A.tail->hr_alink.next = qe;
    qe->hr_alink.next = NULL;
    qe->hr_alink.prev = A.tail;
    A.tail = qe;
}

void del_active_queue( struct hashrec *qe )
{
#if DEBUG_ACTIVE_FEED
    printf("Deleting a flow from active queue...\n");
#endif
    assert(qe != NULL);

    A.num_entries--;

    /* 1-node */
    if (A.head == A.tail)
    {
        A.head = A.tail = NULL;
#if DEBUG_ACTIVE_FEED
        printf("A queue is now empty\n");
#endif
        return;
    }

    if (qe->hr_alink.prev != NULL)
    {
        qe->hr_alink.prev->hr_alink.next = qe->hr_alink.next;
        qe->hr_alink.prev = NULL;       // remember we're off the list now
    }
    else
        A.head = qe->hr_alink.next;

    if (qe->hr_alink.next != NULL)
    {
        qe->hr_alink.next->hr_alink.prev = qe->hr_alink.prev;
        qe->hr_alink.next = NULL;       // remember we're off the list now
    }
    else
        A.tail = qe->hr_alink.prev;
}

