/*
 * definitions for queue routines
 */

#ifndef _QUEUE_H
#define _QUEUE_H



#if __unix__
#   include <netinet/in.h>              // struct in_addr
#else
#   include "..\wattcp\tcp.h"           // struct in_addr
#endif
#include "..\fore\cardstmp.h"           // class CardClock
#include "subnet.h"



struct queue_entry {
    struct hashrec *prev;
    struct hashrec *next;
};

struct queue {
    struct hashrec *head;
    struct hashrec *tail;
    int num_entries;
};

void insque( struct hashrec *qe );
void delque( struct hashrec *qe );
void endque( struct hashrec *qe );
void dumpq();

/* klt - 10/31/96 */
void update_active_queue( struct hashrec *qe );
void del_active_queue( struct hashrec *qe );

extern struct queue Q, A;



/*
 * flowdata is what makes a flow unique
 */

struct flowdata {
    struct in_addr  fd_src;     /* SRC IP Address */
    struct in_addr  fd_dst;     /* DST IP Address */
#if __alpha
    u_long      fd_len : 16;    /* Pkt length (from IP) */
    u_long      fd_proto : 8;   /* Protocol ID (IP) */
    u_long      fd_tcpflags : 8;/* TCP flags (if TCP) */
    u_long      fd_sport : 16;  /* Src. Port (TCP or UDP) */
    u_long      fd_dport : 16;  /* Dest. Port (TCP or UDP) */
#else
    u_short     fd_len;         /* Pkt length (from IP) */
    u_char      fd_proto;       /* Protocol ID (IP) */
    u_char      fd_tcpflags;    /* TCP flags (if TCP) */
    u_short     fd_sport;       /* Src. Port (TCP or UDP) */
    u_short     fd_dport;       /* Dest. Port (TCP or UDP) */
#endif
};



/* flowstats is what we know about all packets in a flow, so far
 */

struct flowstats {
    struct flowdata fs_fd;
    CardClock fs_start;             /* time flow established */
    CardClock fs_access;            /* flow last accessed */
    u_int fs_packets;               /* tot # packets on this flow */
    Bit64 fs_bytes;                 /* tot # bytes on this flow */
    u_int fs_index;                 /* bucket number */
    u_char fs_start_if;             // joel debug
    u_char fs_access_if;            // joel debug
    u_short padding;                // SGI & Sun require even multiple of 4 bytes
};



/*
 * hashrecord contains:
 *      flow last access time
 *      flow data
 *      forward/backward links for 2x-link hash chains
 */

struct hashrec {
    struct queue_entry hr_qlink;    /* earlier access times first */
    struct queue_entry hr_alink;    /* klt - 10/31/96 - active feed, least recently tagged first (not tagged again if touched again before sending) */
    struct queue_entry hr_clink;    /* hash table collision resolution, earlier start times first */
    struct flowstats hr_fs;

#   define hr_fd        hr_fs.fs_fd
#   define hr_start     hr_fs.fs_start
#   define hr_access    hr_fs.fs_access
#   define hr_packets   hr_fs.fs_packets
#   define hr_bytes     hr_fs.fs_bytes
#   define hr_index     hr_fs.fs_index
#   define hr_start_if  hr_fs.fs_start_if
#   define hr_access_if hr_fs.fs_access_if
};



struct porthashrec {
    u_int padding : 8; /* SGI & Sun require, PC wants even multiple of 4 bytes */
    u_int proto : 8;
    u_int sport : 16;  /* Src. Port (TCP or UDP) */
    u_int dport : 16;  /* Dest. Port (TCP or UDP) */
    u_int flows, packets;
    Bit64 bytes;
    double duration;
    double p2sum, b2sum, d2sum; /* sum sq's to get var */
    double pbsum, bdsum, pdsum; /* sum prod's to get covar */
    struct porthashrec *next;
};



struct nethashrec {
    u_short src;                /* AS number */
    u_short dst;                /* AS number */
    u_int flows, packets;
    Bit64 bytes;
    double duration;
    double p2sum, b2sum, d2sum; /* sum sq's to get var */
    double pbsum, bdsum, pdsum; /* sum prod's to get covar */
    struct nethashrec *next;
};



struct hosthashrec {
    u_int ip;
    u_char proto;
    u_char padding;
    u_short port;
    u_int flows, packets;
    Bit64 bytes;
    double duration;
    double p2sum, b2sum, d2sum; /* sum sq's to get var */
    double pbsum, bdsum, pdsum; /* sum prod's to get covar */
    struct hosthashrec *next;
};





/* this one is actually in FS2FLOWS.C
 * but that source file doesn't have a
 * header yet, so put prototype here
 */
void dumphashrec( struct hashrec *hp );
#endif
