// WARNING: this code should not be executed on new traces,
// since the BufferQueue::dump() member function should
// be saving to disk in network byte order in the first place
//



#include <stdio.h>                      // fopen()
#include <errno.h>                      // errno
#include <conio.h>                      // kbhit()
#include "trace.h"
#include "..\intel\byteswap.h"          // htons() htonl()



// conditional compilation switches
//
#define DO_THE_SWAP             TRUE



// global variables
//
struct traceblock buffer;



int main( int argc, char *argv[] )
{
    const char *program = argv[ 0 ];
    const char *infile  = argv[ 1 ];
    const char *outfile = argv[ 2 ];
    FILE *in_handle;
    FILE *out_handle;
    int result;
    Bit32 used;
    Bit32 count;
    Bit32 temp;
    unsigned long block_count = 0;

    printf( "each record is %d bytes\n", sizeof(struct tracerec) );
    printf( "ICMP alone accounts for %d bytes\n", sizeof(struct icmp) );
    printf( "%d records in a block\n\n", COMPL_RING_ENTRIES );

    if( argc != 3 )
    {
        printf( "byteswaps an OC3MON trace file to network byte order\n\n" );
        printf( "usage: %s infile outfile\n", program );
        return 1;
    }

    in_handle = fopen( infile, "rb" );
    if( in_handle == NULL )
    {
        printf( "error %d opening input file\n", errno );
        return 2;
    }

    out_handle = fopen( outfile, "wb" );
    if( out_handle == NULL )
    {
        printf( "error %d opening output file\n", errno );
        return 3;
    }

    while( !feof( in_handle ) && !ferror( in_handle ) && !kbhit() )
    {
        result = fread( &buffer, sizeof(buffer), 1, in_handle );
        if( result != 1 )
        {
            printf( "error %d reading input file\n", errno );
            break;
        }

#if DO_THE_SWAP
#if !HOST_IS_BIG_ENDIAN
        // must remember usage count before swapping
        // when we swap on the Intel machine
        //
        used = buffer.used;
#endif

        // swap the usage count and card number
        //
        buffer.used = htonl( buffer.used );
        buffer.cardnum = htons( buffer.cardnum );

#if HOST_IS_BIG_ENDIAN
        // must remember usage count after swapping
        // when we swap on the bigendian machine
        //
        used = buffer.used;
#endif

        // check for out-of-bound record count
        //
        if( used > COMPL_RING_ENTRIES )
        {
            printf( "error: usage count of block %u is %u, but max is %u!\n", block_count, used, COMPL_RING_ENTRIES );
            break;
        }

        // for each record in the buffer...
        //
        for( count=0; count<used; count++ )
        {
            // swap the bytes and words of the timestamp
            //
            temp = htonl( buffer.data[count].when.low );
            buffer.data[count].when.low  = htonl( buffer.data[count].when.high );
            buffer.data[count].when.high = temp;

            // swap the bytes of the ATM header
            //
            buffer.data[count].atm.whole = htonl( buffer.data[count].atm.whole );

            // the rest of the data is IP and TCP headers,
            // and was always in network byte order!
            //
        }
#endif  // DO_THE_SWAP

        result = fwrite( &buffer, sizeof(buffer), 1, out_handle );
        if( result != 1 )
        {
            printf( "error %d writing output file\n", errno );
            break;
        }

        block_count++;
        printf( "%lu ", block_count*sizeof(struct traceblock) );
    }
    printf( "\n" );

    if( kbhit() )
    {
        getch();
        fprintf( stderr, "halted by keypress!\n" );
    }

    fclose( out_handle );
    fclose( in_handle );

    return 0;
}

