#include "..\types\types.h"

#ifdef __cplusplus
extern "C" {
#endif

#if __DPMI32__
#   define far
#   define _cdecl
#endif

int far _cdecl pci_is_bios_present( void );

#define PCI_CLOCK_RATE               33E6       // 33 MHz for 166 MHz CPU, 30 MHz for 120 MHz CPU
#define INTEL_82437FX                FALSE
#define INTEL_82439HX                TRUE

#define INTEL_VENDOR_ID              0x8086
#define FORE_VENDOR_ID               0x1127
#define TI_VENDOR_ID                 0x104C
#define TRANCELL_VENDOR_ID           0x11D3
#define NUMBER9_VENDOR_ID            0x5333
#define APPTEL_VENDOR_ID             0x6A61

#define INTEL_T1_HOST2PCI_DEVICE_ID  0x122D
#define INTEL_T1_PCI2ISA_DEVICE_ID   0x122E
#define INTEL_T1_EIDE_DEVICE_ID      0x1230
#define FORE_PCA200A_DEVICE_ID       0x0210
#define FORE_PCA200E_DEVICE_ID       0x0300
#define TI_TNETA_1561_DEVICE_ID      0xA100
#define TRANCELL_PCI_155_DEVICE_ID   0x0002
#define NUMBER9_MOTION771_DEVICE_ID  0x88F0
#define APPTEL_PCI_622_DEVICE_ID     0x0622

#define RETCODE_NO_MORE_PCI_DEVICES  0xFFFF

// structure of device ID (a.k.a handle)
//      bits 15:8       bus number
//            7:3       device number
//            2:0       function number (subdevice)
//
#define PCI_BUSSES_PER_MACHINE       16         // 256 works on DELL, but causes aliases on ASUS
#define PCI_DEVICES_PER_BUS          32
#define PCI_FUNCTIONS_PER_DEVICE     8

Bit16 far _cdecl pci_get_device_handle(
    Bit16 device_id,
    Bit16 vendor_id,
    Bit16 index
);



// functions used by macros below
//

Bit32 far _cdecl pci_read_config_32( Bit16 device_id, Bit16 offset );
Bit16 far _cdecl pci_read_config_16( Bit16 device_id, Bit16 offset );
Bit8 far _cdecl pci_read_config_8( Bit16 device_id, Bit16 offset );
void far _cdecl pci_write_config_32( Bit16 device_id, Bit16 offset, Bit32 value );
void far _cdecl pci_write_config_16( Bit16 device_id, Bit16 offset, Bit16 value );
void far _cdecl pci_write_config_8( Bit16 device_id, Bit16 offset, Bit8 value );



// specific uses of PCI configuration space that work on all PCI devices
//

// mask off the control bits
//
// bit 0 = 0       use memory space
// bit 0 = 1       use I/O space (bits 1:3 redefined)
//
// bits 2:1 = 00   locate anywhere in 32-bit address space
// bits 2:1 = 01   locate below 1MB (for Intel real-mode access)
// bits 2:1 = 10   locate anywhere in 64-bit address space
// bits 2:1 = 11   reserved
//
// bit 3 = 0       not prefetchable
// bit 3 = 1       prefetchable
//
#define pci_get_base_address(h) (0xFFFFFFF0 & pci_read_config_32(h,0x10))
#define pci_get_base_address_n(h,i) (0xFFFFFFF0 & pci_read_config_32(h,0x10 + 4*(i)))

#define pci_get_irq_level(h) pci_read_config_8(h,0x3C)

#define pci_get_status(h) pci_read_config_16(h,0x06)

#define pci_get_command(h) pci_read_config_16(h,0x04)
#define pci_set_command(h,v) pci_write_config_16(h,0x04,v)

// TI wants to be an even multiple of 8; rounds down
//
#define pci_get_master_latency_timer(h) pci_read_config_8(h,0x0D)
#define pci_set_master_latency_timer(h,v) pci_write_config_8(h,0x0D,v)

// is read-only, so set does not work
//
// measured in .25 uS
//
#define pci_get_max_latency(h) pci_read_config_8(h,0x3F)
#define pci_set_max_latency(h,v) pci_write_config_8(h,0x3F,v)

// is read-only, so set doesn't work
//
// measured in .25 uS
//
#define pci_get_min_grant(h) pci_read_config_8(h,0x3E)
#define pci_set_min_grant(h,v) pci_write_config_8(h,0x3E,v)

#define pci_get_cache_line_size(h) pci_read_config_8(h,0x0C)
#define pci_set_cache_line_size(h,v) pci_write_config_8(h,0x0C,v)




// stuff that is specific to motherboard chipsets
//

// how many PCI wait-states will TVX add during the burst portion
// of a PCI master read or write cycle targeted to the TVX (I think
// that means host memory)
//
// only defined for Intel 82437VX
//
// 000 = 2 cycles
// 001 = 4 cycles
// 010 = 6 cycles
// 011 = 8 cycles (default, matches PCI 2.0)
// 1xx = reserved
//
#define pci_get_trdy_timer(h) pci_read_config_8(h,0x69)
#define pci_set_trdy_timer(h,v) pci_write_config_8(h,0x69,v)

// how long will a PCI master be allowed to retain ownership
// of the bus (from initial assertion of grant) (minimum)
//
// allows a PCI device to do multiple transactions by keeping
// its request asserted
//
// only defined for Intel 82437VX and 82439HX
//
// measured in multiples of 4 PCI clock cycles
// (bottom 2 bits are forced to zeroes)
//
#define pci_get_multi_transaction_timer(h) pci_read_config_8(h,0x69)
#define pci_set_multi_transaction_timer(h,v) pci_write_config_8(h,0x69,v)

// only applies to Intel 82437FX and 82439HX PCI/host bridge chips
//
// note that bit mapping is different on the 2 chips (see below)
//
#define pci_get_control(h) pci_read_config_8(h,0x50)
#define pci_set_control(h,v) pci_write_config_8(h,0x50,v)

// only applies to Intel 82437FX and 82439HX PCI/host bridge chips
//
#define pci_get_cache_control(h) pci_read_config_8(h,0x52)
#define pci_set_cache_control(h,v) pci_write_config_8(h,0x52,v)

// only applies to Intel 82371FB and 82371SB PCI/ISA accelerator (PIIX and PIIX3) chips
//
#define pci_get_smi_control(h) pci_read_config_8(h,0xA0)
#define pci_set_smi_control(h,v) pci_write_config_8(h,0xA0,v)

// only applies to Intel 82371FB and 82371SB PCI/ISA accelerator (PIIX and PIIX3) chips
//
#define pci_get_smi_enable(h) pci_read_config_16(h,0xA2)
#define pci_set_smi_enable(h,v) pci_write_config_16(h,0xA2,v)

// only applies to Intel 82371FB and 82371SB PCI/ISA accelerator (PIIX and PIIX3) chips
//
#define pci_get_smi_event_enable(h) pci_read_config_32(h,0xA4)
#define pci_set_smi_event_enable(h,v) pci_write_config_32(h,0xA4,v)

// only applies to Intel 82371FB and 82371SB PCI/ISA accelerator (PIIX and PIIX3) chips
//
// has actual value minus 1
//
#define pci_get_smi_fastoff_timer(h) pci_read_config_8(h,0xA8)
#define pci_set_smi_fastoff_timer(h,v) pci_write_config_8(h,0xA8,v)

// only applies to Intel 82371FB and 82371SB PCI/ISA accelerator (PIIX and PIIX3) chips
//
#define pci_get_smi_request(h) pci_read_config_16(h,0xAA)
#define pci_set_smi_request(h,v) pci_write_config_16(h,0xAA,v)

// only applies to Intel 82437FX and 82439HX PCI/host bridge chips
//
#define pci_get_smi_ram_control(h) pci_read_config_8(h,0x72)
#define pci_set_smi_ram_control(h,v) pci_write_config_8(h,0x72,v)



// stuff that is specific to certain vendor's cards
//

// only applies to FORE PCA200E
//
// measured in cycles
//
#define pci_get_max_slave_latency(h) pci_read_config_8(h,0x40)
#define pci_set_max_slave_latency(h,v) pci_write_config_8(h,0x40,v)

// nonstandard, probably only applies to FORE PCA200E
//
#define pci_get_master_control(h) pci_read_config_8(h,0x41)

// nonstandard, probably only applies to FORE PCA200E
//
#define pci_set_master_control(h,v) pci_write_config_8(h,0x41,v)

// nonstandard, probably only applies to FORE PCA200E
//
// measured in 32-bit words
//
#define pci_get_threshold(h) pci_read_config_8(h,0x42)

// nonstandard, probably only applies to FORE PCA200E
//
#define pci_set_threshold(h,v) pci_write_config_8(h,0x42,v)

#ifdef __cplusplus
}
#endif



// assumes Intel little-endian bit ordering,
// and structure packeing
//
struct PciCommand
{
    Bit16 input_output_space_enable : 1;
    Bit16 memory_space_enable       : 1;
    Bit16 bus_master_enable         : 1;
    Bit16 special_cycle_mon_enable  : 1;
    Bit16 mem_write_inval_enable    : 1;
    Bit16 vga_palette_snoop_enable  : 1;
    Bit16 parity_error_response     : 1;
    Bit16 stepping_enable           : 1;
    Bit16 system_error_response     : 1;
    Bit16 fast_back_to_back_enable  : 1;
    Bit16                           : 6;
};



struct PciStatus
{
    Bit16                           : 5;
    Bit16 is_66_mhz_capable         : 1;
    Bit16 user_def_feature_support  : 1;
    Bit16 fast_back_to_back_capable : 1;
    Bit16 signaled_parity_error     : 1;
    Bit16 devsel_timing             : 2;        // 00=fast 01=medium 10=slow 11=reserved
    Bit16 signaled_target_abort     : 1;
    Bit16 received_target_abort     : 1;
    Bit16 received_master_abort     : 1;
    Bit16 signaled_system_error     : 1;
    Bit16 detected_parity_error     : 1;
};


#if INTEL_82437FX
// as far as I know, only defined on Intel 82437FX
// Triton System Controller
//
// Intel recommends using 3 or 5 PCI clocks for the
// CPU inactivity timer
//
struct PciControl
{
    Bit8 bus_concurrency_disable    : 1;
    Bit8 pci_streaming_disable      : 1;
    Bit8 cpu_to_pci_write_bursting_disable : 1;
    Bit8 peer_concurrency_enable    : 1;
    Bit8                            : 1;
    Bit8 cpu_inactivity_timer       : 3;        // actual value minus one, measured in PCI clocks
};
#else
#if INTEL_82439HX
// as far as I know, only defined on Intel 82439HX
// Triton System Controller
//
struct PciControl
{
    Bit8 global_txc_enable          : 1;
    Bit8                            : 1;        // reserved
    Bit8 serr_pound_output_type     : 1;        // 0=open drain compatible with PCI, 1=normal output driven high when negated
    Bit8 peer_concurrency_enable    : 1;
    Bit8 dual_processor_na_pound_enable : 1;    // 0=use 82437FX policies, 1=use policies for dual-procesor PCI 2.1
    Bit8 shutdown_to_port_92        : 1;        // 0=forward host shutdown cycle to PCI bus, 1=write 01h to port 92
    Bit8 ecc_test_enable            : 1;        // 0=normal mode, 1=test mode
    Bit8 dram_ecc_or_parity_select  : 1;        // 0=parity, 1=ECC (also adjusts DRAM timings)
};
#endif
#endif



#if INTEL_82437FX || INTEL_82439HX
// as far as I know, only defined on Intel 82437FX
// and 82439HX Triton System Controller
//
struct PciCacheControl
{
    Bit8 first_level_cache_enable   : 1;
    Bit8 secondary_cache_force_miss_or_invalidate : 1;
#if INTEL_82439HX
    Bit8 extended_cachability_enable : 1;       // 0=64MB 1=512MB (enables TIO[10:8] lines)
#else
    Bit8                            : 1;        // reserved (for guess who?)
#endif
    Bit8 na_pound_signal_disable    : 1;        // must be setup before either L1 or L2 cache is enabled... but what does NA# signal do?
    Bit8 static_ram_type            : 2;        // 00=pipe burst, 01=burst, 10=async, 11=512K dual bank pipe burst
    Bit8 secondary_cache_size       : 2;        // 00=not populated, 01=256K, 10=512K, 11=reserved
};

typedef enum PciCacheControlSramType
{
    pipelined_burst = 0,
    burst,                                      // reserved on 82439HX
    asynchronous,                               // reserved on 82439HX
    dual_512k_pipe_burst
};

typedef enum PciCacheControlSecondaryCacheSize
{
    not_populated = 0,
    has_256k,
    has_512k,
    reserved
};
#endif



// as far as I know, only defined on Intel 82371FB
// Triton PCI ISA accelerator (PIIX)
//
struct PciSmiControl
{
    Bit8 smi_gate                   : 1;
    Bit8 stpclk_signal_enable       : 1;
    Bit8 stpclk_scaling_enable      : 1;
    Bit8 fastoff_timer_granularity  : 2;        // 00=minutes, 01=disabled, 10=PCICLK, 11=mS
    Bit8                            : 3;        // reserved
};

typedef enum PciSmiControlFastoffGranularity
{
    one_minute = 0,                 // actually 1.0 min @33 MHz PCICLK, 1.1 min @30 MHz, 1.32 min @25 MHz
    disabled,
    one_pci_clock,
    one_millisecond                 // actually 1.0 mS  @33 MHz PCICLK, 1.1 mS  @30 MHz, 1.32 mS  @25 MHz
};



// as far as I know, only defined on Intel 82371FB
// Triton PCI ISA accelerator (PIIX)
//
struct PciSmiEnable
{
    Bit16 irq_1_smi_enable          : 1;        // keyboard
    Bit16 irq_3_smi_enable          : 1;        // COM1/COM3 or mouse
    Bit16 irq_4_smi_enable          : 1;        // COM2/COM4 or mouse
    Bit16 irq_8_smi_enable          : 1;        // realtime clock alarm
    Bit16 irq_12_smi_enable         : 1;        // PS/2 mouse
    Bit16 fastoff_timer_smi_enable  : 1;        // when it decrements to zero
    Bit16 extsmi_smi_enable         : 1;
    Bit16 apmc_write_enable         : 1;        // program's write to APM command register
    Bit16                           : 8;        // reserved
};



// as far as I know, only defined on Intel 82371FB
// Triton PCI ISA accelerator (PIIX)
//
// system events keep the system from powering down
// reloading the fast-off timer with its initial value
//
// break events can wake up a powered-down system by
// negating STPCLK#
//
struct PciSmiEventEnable
{
    Bit32 fastoff_irq_0_enable      : 1;        // system and break events
    Bit32 fastoff_irq_1_enable      : 1;        // system and break events
    Bit32                           : 1;
    Bit32 fastoff_irq_3_enable      : 1;        // system and break
    Bit32 fastoff_irq_4_enable      : 1;        // system and break
    Bit32 fastoff_irq_5_enable      : 1;        // system and break
    Bit32 fastoff_irq_6_enable      : 1;        // system and break
    Bit32 fastoff_irq_7_enable      : 1;        // system and break
    Bit32 fastoff_irq_8_enable      : 1;        // system and break
    Bit32 fastoff_irq_9_enable      : 1;        // system and break
    Bit32 fastoff_irq_10_enable     : 1;        // system and break
    Bit32 fastoff_irq_11_enable     : 1;        // system and break
    Bit32 fastoff_irq_12_enable     : 1;        // system and break
    Bit32 fastoff_irq_13_enable     : 1;        // system and break
    Bit32 fastoff_irq_14_enable     : 1;        // system and break
    Bit32 fastoff_irq_15_enable     : 1;        // system and break
    Bit32                           : 13;
    Bit32 fastoff_nmi_enable        : 1;        // system and break
    Bit32 intr_enable               : 1;        // break only
    Bit32 fastoff_smi_enable        : 1;        // system and break
};



// as far as I know, only defined on Intel 82371FB
// Triton PCI ISA accelerator (PIIX)
//
// note that these are all edge sensitive, so if an
// SMI cause is still occurring when CPU writes zero
// to the corresponding bit in this register, the bit
// will not be set again by PIIX unless it goes away
// and comes back
//
struct PciSmiRequest
{
    Bit16 irq_1_caused_smi          : 1;
    Bit16 irq_3_caused_smi          : 1;
    Bit16 irq_4_caused_smi          : 1;
    Bit16 irq_8_caused_smi          : 1;
    Bit16 irq_12_caused_smi         : 1;
    Bit16 fastoff_timer_expired     : 1;        // but it reloaded and kept counting afterwards
    Bit16 extsmi_caused_smi         : 1;
    Bit16 apmc_write_caused_smi     : 1;
    Bit16                           : 8;
};



// as far as I know, only defined on Intel 82437FX
// Triton System Controller (TSC)
//
// note: software must ensure that smm_space_open
// and smm_space_closed are not both TRUE at the
// same time
//
struct PciSmiRamControl
{
    Bit8 smm_space_base_segment     : 3;        // must be 010 = A0000h to BFFFFh
    Bit8 sm_ram_enable              : 1;        // 128K bytes accessible when ADS# and SMIACT# asserted
    Bit8 smm_space_locked           : 1;        // smm_space_open and smm_space_closed become R/O; must reset to clear this bit
    Bit8 smm_space_closed           : 1;        // data passes through, code refs see SM RAM if SMIACT# asserted
    Bit8 smm_space_open             : 1;        // code & data see SM RAM regardles of SMIACT#
    Bit8                            : 1;        // reserved
};



// as far as I know, only defined on FORE PCA200E OC3c ATM NIC
//
struct PciMasterControl
{
    Bit8 disable_cache_line_reads   : 1;
    Bit8 disable_write_and_invals   : 1;
    Bit8 write_inval_needs_2_lines  : 1;
    Bit8 ignore_latency_timer       : 1;
    Bit8 enable_cont_request_mode   : 1;        // don't deassert REQ# if in and out BIFO's have threshold fullness
    Bit8 force_large_pci_bus_bursts : 1;        // (?input only) wait to assert REQ# until threshold input FIFO words available
    Bit8 byteswap_slave_ram_access  : 1;
    Bit8                            : 1;
};



// someday expand this interface so that the full
// 64 bytes of configuration registers defined in
// chapter 17 of PCI System Architecture, 3rd ed.
// including bit fields where necessary
//
