/* 1155, Fri 13 Sep 96

   SNMPIMPL.H:  Global declarations for snmpauth.c and anmpagnt.c
            NeTraMet version, based on CMU SNMPv2 (see below).

   Copyright (C) 1992-1998 by Nevil Brownlee,
   ITSS Technology Development, The University of Auckland */

/* $Log: snmpimpl.h,v $
 *
 * Revision 1.4  1998/06/10 21:04:57  rtfm
 * *** empty log message ***
 *
 * Revision 1.3  1998/06/03 04:42:38  rtfm
 * Fix unmatched slash/star comments
 *
 * Revision 1.2  1998/04/30 04:57:13  rtfm
 * Bug fixes, change copyright date
 *
 * Revision 1.1  1998/04/30 04:07:43  rtfm
 * Initial revision
 */

/***********************************************************
    Copyright 1988, 1989 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/


#ifdef __cplusplus
extern "C" {
#endif

/*
 * Error codes:
 *
 * These must not clash with SNMP error codes (all positive).
 */
#define PARSE_ERROR -1
#define BUILD_ERROR -2

#define COMMUNITY_MAX_LEN  64
#define MAX_NAME_LEN       64  /* Nbr of subid's in an objid (was 128!) */
          /* Needs to be this long for 'selector' strings within OIDs */
          /* >>> Should be >= AGENT_OID_LEN in agent/snmp_vars.h <<< */

#ifndef TRUE
#define TRUE    1
#endif
#ifndef FALSE
#define FALSE   0
#endif

struct packet_info {
   int  version;
   u_char   pdutype;

   /* Community based authentication */
   u_char   community[COMMUNITY_MAX_LEN + 1];
   int      community_len;
   int      community_id;

   u_char   *packet_end;
   };

#define READ        1
#define WRITE       0

#define RESERVE1    0
#define RESERVE2    1
#define COMMIT      2
#define ACTION      3
#define FREE        4

/*
 * ##############################################################
 * IMPORTANT NOTE:
 * ##############################################################
 *
 * The format of the acl word in these entries has changed.  It is still
 * 2 bits per community, offset from the right by the index of the community.
 * The leftmost two bits denotes read access, and the rightmost denotes
 * write access.
 * The change is that the rightmost two bits are now reserved for the object's
 * max-access.  This is the minimum of what makes "protocol sense" for the
 * object and whether set support was implemented for that object.
 * These two bits will not map to any community.  The first community
 * entry will map to the 3rd and 4th bits.
 */

#define RONLY    0xAAAA /* Read access for everyone */
/* #define RWRITE    0xAABB   Add write access for community private */
#define RWRITE   0xAAEB /* Add write access for 'private' (wrong bit set above!) */
#define NOACCESS 0x0000 /* No access for anybody */

#define INTEGER     ASN_INTEGER
#define STRING      ASN_OCTET_STR
#define OBJID       ASN_OBJECT_ID
#define NULLOBJ     ASN_NULL
#define BITSTRING   ASN_BIT_STR

/* defined types (from the SMI, RFC 1157) */
#define IPADDRESS   (ASN_APPLICATION | 0)
#define COUNTER     (ASN_APPLICATION | 1)
#define GAUGE       (ASN_APPLICATION | 2)
#define TIMETICKS   (ASN_APPLICATION | 3)
#define OPAQUE      (ASN_APPLICATION | 4)

/* defined types (from the SMI, RFC 1902) */
#define NSAP        (ASN_APPLICATION | 5)
#define COUNTER64   (ASN_APPLICATION | 6)
#define UINTEGER    (ASN_APPLICATION | 7)

/* Textual Conventions (from  RFC 1903) */
#define RS_ACTIVE         1  /* RowStatus values */
#define RS_NOTINSERVICE   2
#define RS_NOTREADY       3
#define RS_CREATEANDGO    4
#define RS_CREATEANDWAIT  5
#define RS_DESTROY        6


struct trapVar {
   oid     *varName;
   int     varNameLen;
   u_char  varType;
   int     varLen;
   u_char  *varVal;
   struct trapVar *next;
   };

#ifdef DEBUG
#define ERROR(string)   \
   log_msg(LOG_DEBUG, 0, "%s(%d): %s\n",__FILE__, __LINE__,string)
      /* In ..\netramet\met_vars.c */
#else
#define ERROR(string)
#endif

#ifdef DOS
extern int scr_lrow;       /* In ..\netramet\pktsnap.h */
void scpos(int x, int y);  /* In ..\netrmet\chart.cpp */
#endif

/* From snmp.c*/
extern u_char   sid[];  /* Size SID_MAX_LEN */

int snmp_agent_parse(u_char *data, int length,  /* In snmpagnt.c */
   u_char *out_data, int *out_length, u_long sourceip);

u_char *snmp_auth_parse(u_char *data, int *length,  /* In snmpauth.c */
   u_char *sid, int *slen, int *version);
u_char *snmp_auth_build(u_char *data, int *length,
   u_char *sid, int *slen, int *version, int messagelen);

typedef int (*write_fn_type)(  /* In meter/met_vars.c */
      int action, u_char *var_val, u_char var_val_type, int var_val_len,
      u_char far *statP, oid *name, int name_len);

u_char far *getStatPtr(
   oid *name, int *namelen, u_char *type, int *len,
   u_short *acl, int exact, write_fn_type *write_method,
   struct packet_info *pi, int *noSuchObject);

#ifdef __cplusplus
}           // end extern "C"
#endif
