/* DEBUG flag may be set for my internal playing */

/*
#define DEBUG
*/

#define J_DEBUG_STARTDATA_PARANOIA
#define noJ_DEBUG_TCP_SEND_STARTDATA
#define noJ_DEBUG_READ_WINDOW_UPDATE
#define noJ_DEBUG_RX_WINDOW_UPDATE
#define noJ_DEBUG_FAST_RETRANSMIT
#define noJ_DEBUG_TX             /* just dump TX seqnum and acknum */
#define noJ_DEBUG_RX             /* just dump RX seqnum and acknum */
#define noN_DEBUG_TX           /* Nevil's debug/tracing stuff */
#define noN_DEBUG_RX           /* Nevil's debug/tracing stuff */
#define noN_DEBUG_UPDOWN       /* Nevil's debug/tracing stuff */
#define noN_DEBUG_TX_RX_EXTRA  /* Nevil's debug/tracing stuff */

// compile-time switches
//
#define ALLOW_FAST_RETRANSMIT   TRUE

/*
 *  PCTCP - the true worker of Waterloo TCP
 *        - contains all opens, closes, major read/write routines and
 *          basic IP handler for incomming packets
 *        - NOTE: much of the TCP/UDP/IP layering is done at the data structure
 *          level, not in separate routines or tasks
 *
 */



#include "..\wattcp\copyrigh.h"
#include <time.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include <mem.h>
#include <dos.h>
#include <values.h>

#include "..\wattcp\wattcp.h"
#include "..\wattcp\elib.h"
#include "..\types\types.h"
#include "..\dpmi32\phys.h"

static void udp_handler(in_Header *ip);
static udp_write(udp_Socket *s, byte *datap, int len);
static int udp_read(udp_Socket *s, byte *datap, int maxlen);
static void tcp_Retransmitter(void);


#define TCP_LOCAL 0x4000

/* statics */
static int tcp_ProcessData(tcp_Socket *s, tcp_Header *tp, int len);

static char *mono;
static char *colour;
static Bit32 mono_handle, colour_handle;

static initialized = 0;
static void (*system_yield)() = NULL;
extern int multihomes;
extern word _pktipofs;
void (*_dbugxmit)(sock_type *s, in_Header *ih, tcp_Header *ph, char *line) = NULL;
void (*_dbugrecv)(sock_type *s, in_Header *ih, tcp_Header *ph, char *line) = NULL;
void (*wattcpd)(void) = NULL;

char *_hostname = "012345678901234567890123456789012345678901234567890";

word _mss = ETH_MSS;

char *_wattcp = WATTCP_C;

static int tcp_handler(in_Header *ip);
static void udp_handler(in_Header *ip);
extern void icmp_handler(in_Header *ip );

static void tcp_unthread(tcp_Socket *ds);
static void tcp_abort(tcp_Socket *s);
void tcp_sendsoon(tcp_Socket *s, int now);
static void tcp_send(tcp_Socket *s, char *from);
static void tcp_rst( in_Header *his_ip, tcp_Header *oldtcpp);
static void udp_close(udp_Socket *ds);
static void tcp_close(tcp_Socket *s);


/*
 * sock_yield - enable user defined yield function
 */
void sock_yield( tcp_Socket *s, void (*fn)())
{
    if ( s )
        s->usr_yield = fn;
    else
        system_yield = fn;
}

/*
 * sock_mode - set binary or ascii - affects sock_gets, sock_dataready
 *           - set udp checksums
 */
word sock_mode( sock_type *s, word mode )
{
     return( s->tcp.sock_mode = (s->tcp.sock_mode & 0xfffc) | mode);
}

/*
 * sock_ttl - set time to live
 *            for all packets sent in the future with this socket
 *
 * returns TTL of packets sent in the past
 */
byte sock_ttl( sock_type *s, byte ttl )
{
    byte old_ttl;

    old_ttl = s->tcp.ttl;
    s->tcp.ttl = ttl;
    return old_ttl;
}

/*
 * sock_tunnel - set IP address to tunnel through (in host byte order)
 *               for all packets sent in the future with this socket
 *
 *               0.0.0.0 disables tunneling
 *
 * returns IP address that past packets were tunneled through
 */
longword sock_tunnel( sock_type *s, longword tunneladdr )
{
    longword old_tunneladdr;

    old_tunneladdr = s->tcp.tunneladdr;
    s->tcp.tunneladdr = tunneladdr;
    return old_tunneladdr;
}

/*
 * ip user level timer stuff
 *   void ip_timer_init( void *s, int delayseconds )
 *   int  ip_timer_expired( void *s )
 *      - 0 if not expired
 */
#define MAXTICKS 0x1800b0L  // BIOS ticks per day

void ip_timer_init( udp_Socket *s , int delayseconds )
{
    if (delayseconds)
        s->usertimer = set_timeout( delayseconds );
    else
        s->usertimer = 0;
}

int ip_timer_expired( udp_Socket *s )
{
    if (! s->usertimer) /* cannot expire */
        return( 0 );
    return( chk_timeout( s->usertimer));
}

/* only used by lpr.c and lpq.c to generate a random
 * 8 bit value, so just get some bits from set_ttimeout
 */
longword MsecClock()
{
    return( set_ttimeout(0) );
}

static long make_timeout( word timeout )
{
    if ( timeout ) return( set_timeout( timeout ));
    return( 0 );
}

/*
 * check_timeout - test agains timeout clock - account for overflow
 */
static int check_timeout( unsigned long timeout )
{
    if (timeout) return( chk_timeout( timeout ));
    return( 0 );
}

/*
 * Local IP address
 */
longword my_ip_addr = 0L;       /* for external references */
longword sin_mask = 0xfffffe00L;
longword sin_gate = 0x0;


/*
 * IP identification numbers
 */

static int ip_id = 0;                   /* packet number */
static int next_tcp_port = 1024;        /* auto incremented */
static int next_udp_port = 1024;
static tcp_Socket *tcp_allsocs = NULL;
static udp_Socket *udp_allsocs = NULL;

/* Timer definitions */
#define RETRAN_STRAT_TIME  1     /* in ticks - how often do we check retransmitter tables*/
#define tcp_LONGTIMEOUT 31       /* timeout for opens */
#define tcp_TIMEOUT 13           /* timeout during a connection */

static int send_now;

#define MAXVJSA     6*18  // ticks     Nevil, 13 Jan 97
#define MAXVJSD     2*18  // ticks     (used to be in wattcp.h)
#define MAXRTO     10*18
#define MINRTO         2  // ticks     Nevil, 9 Jan 97
#ifdef __DPMI32__
#define MAXPKTSEND   150  // Max pkts sent per tcp_write() call  (Joel)
#else
#define MAXPKTSEND     8  // Max pkts sent per tcp_write() call  (Nevil)
#endif

   // My pkt driver mangled packets if I sent too many
   // of them back to back!            Nevil, 9 Jan 97
   //
   // since we put packets on a global low-level queue when
   // the packet driver tells us they have arrived, they are
   // not dequeued by TCP until application calls tcp_tick(),
   // are we going to find that we lose data even though
   // advertized window is large, because global buffers
   // got full?    if so, maybe we should decrease MAXPKTSEND
   // so TCP doesn't think there is congestion?   nah, prior
   // restraint is less efficient than runtime detection...
   //                                   Joel, 17 Jan 97

word debug_on = 0;

/*
 * look for bugs
 */
tcp_checkfor( tcp_Socket *t )
{
    tcp_Socket *p;

    for ( p = tcp_allsocs ; p ; p = p->next )
        if ( p == t ) return( 1 );
    return( 0 );
}

/*
 * Shut down the card and all services
 */
void tcp_shutdown()
{
    while (tcp_allsocs)
        tcp_abort( tcp_allsocs );
    _eth_release();

#ifdef __DPMI32__
    /* unmap physical memory for the display
     *
     * WARNING: must do this in reverse of the order that
     * we mapped originally, or we will be making the wrong
     * set of pages invalid when we decrease DS limit
     */
    unmap_physical( 4000, mono_handle );
    unmap_physical( 4000, colour_handle );
#endif

    initialized = 0;
}

/*
 * tcp_Init - Initialize the tcp implementation
 *          - may be called more than once without hurting
 */
void tcp_init()
{
    extern int _arp_last_gateway;
    extern int _last_nameserver;

    if (!initialized) {
#ifdef __DPMI32__
        /* map physical memory to flat memory for the display */
        colour = map_physical( 0xB8000, 4000, &colour_handle );
        mono = map_physical( 0xB0000, 4000, &mono_handle );
#else
        colour = (char far *)0xB8000000;
        mono = (char far *)0xB0000000;
#endif
        /* initialize ethernet interface */
        initialized = 1;
        _eth_init();

        /* reset the various tables */
        _arp_last_gateway = 0;  /* reset the gateway table */
        _last_nameserver = 0;   /* reset the nameserver table */
        _last_cookie = 0;       /* eat all remaining crumbs */
        *_hostname = 0;         /* reset the host's name */

        _eth_free( 0 );

        next_udp_port = next_tcp_port = 1024 + (set_ttimeout(0) & 0x1ff);
// !!   next_udp_port = next_tcp_port = 1024 + (get_realclock() & 0x1ff);
    }
}

// turn off compiler warning about unused parameters s and size
// since I know they won't be used and I must have
// them because calling code cannot be changed
//
#pragma warn -par

/*
 * Checks for bugs when compiling in large model C compiler
 *
 * Borland C uses a 4K stack by default.  In all memory models the
 * stack grows down toward the heap.
 *
 * If you accidentally place tcp_Socket onto the stack (like by making
 * it an automatic variable), then you will have already used up that
 * whole 4K and then some!
 *
 * In large model, this will mess up the data space in a major way
 * because the stack starts at SS:_stklen, or SS:1000, so you will
 * wrap the SP pointer back around to FFFE and start writing over
 * the far heap.  Yuck.
 *
 * In small model it usually doesn't kill your application because
 * you would have to be down to your last 4K of memory and this is
 * not as common.
 *
 * The solutions: declare your sockets as static, or put them on the
 * heap, or bump up your stack size by using the global special variable:
 *
 * unsigned _stklen = 16536;    // set stack to 16 k
 */

static void largecheck( void *s, int size )
{
}

// turn compiler warning about unused parameters
// back to whatever it was before I messed with it
//
#pragma warn .par

/*
 * findfreeport - return unused local port
 *              - oldport = 0:normal port, 1:special port (513-1023)
 *              - we need not be this picky, but it doesn't hurt
 */
static word findfreeport( word oldport )
{
    word temp;
    tcp_Socket *s;

    if (( oldport > 0 ) && (oldport < 0xffff))
        return( oldport );

    if ( oldport == 0 ) oldport = 1025;
    else oldport = 513;

    for ( temp = oldport ; temp < oldport + 510 ; ++temp ) {
        if (( s = (tcp_Socket*)udp_allsocs) != NULL ) {
            while ( s->next && (s->myport != temp))
                s = (tcp_Socket*)s->next;
            if ( s->myport == temp ) continue;
        }
        if ( (s = tcp_allsocs ) != NULL ) {
            while ( s->next && (s->myport != temp ))
                s = s->next;
            if ( s->myport == temp ) continue;
        }
        break;
    }
    return( temp );
}

/* socket, localport, destaddress */
int udp_open(udp_Socket *s, word lport, longword ina, word port, dataprocref rdataHandler)
{
    udp_close( s );
    largecheck( s, sizeof( udp_Socket ));
    memset( s, 0, sizeof( udp_Socket ));
    s->rdata = s->rddata;
    s->maxrdatalen = udp_MaxBufSize;
    s->ip_type = IPPROTO_UDP;
    s->ttl = in_DefaultTTL;

    lport = findfreeport(lport);

    s->myport = lport;
    s->myaddr = my_ip_addr;
    s->tunneladdr = 0;

    /* check for broadcast */
    if ( (long)(ina) == -1 || !ina )
        memset( s->hisethaddr, 0xff, sizeof( eth_address ));
    else if ( ! _arp_resolve(ina, (eth_address *) &s->hisethaddr[0], 0) )
        return( 0 );

    s->hisaddr = ina;
    s->hisport = port;
    s->rdataHandler = rdataHandler;
    s->usr_yield = system_yield;
    s->safetysig = SAFETYUDP;
    s->next = udp_allsocs;
    udp_allsocs = s;
    return( 1 );
}

/*
 * Actively open a TCP connection to a particular destination.
 *      - 0 on error
 */
int tcp_open(tcp_Socket *s, word lport, longword ina, word port, dataprocref rdataHandler)
{
    largecheck( s, sizeof( tcp_Socket ));   /* stack space warnings */
    tcp_unthread(s);                        /* just in case not totally closed */

    memset( s, 0, sizeof( tcp_Socket));
    s->rdata = s->rddata;
    s->sdata = s->sddata;
    s->maxrdatalen = tcp_MaxBufSize;
    s->maxsdatalen = tcp_MaxBufSize;

    s->ip_type = IPPROTO_TCP;
    s->ttl = in_DefaultTTL;
    s->mss = _mss;
    s->state = tcp_StateSYNSENT;
    s->timeout = set_timeout( tcp_LONGTIMEOUT );
    s->cwindow = 1;
    s->wwindow = 0;     /* slow start VJ algorithm */
    s->vj_sa = 4;      /* about 250 ms */
    lport = findfreeport( lport );  /* get a nonzero port val */
    s->myaddr = my_ip_addr;
    s->myport = lport;
    s->tunneladdr = 0;

    if ( ina - my_ip_addr <= multihomes ) return( 0 );
    if ( ! _arp_resolve(ina, (eth_address *) &s->hisethaddr[0], 0) )
        return( 0 );

    s->hisaddr = ina;
    s->hisport = port;
    s->seqnum = htonl( set_timeout( 1 ) ) & 0xffff0000;  /* make up a number */
    s->flags = tcp_FlagSYN;
    s->unhappy = true;
    s->rdataHandler = rdataHandler;
    s->usr_yield = system_yield;

    s->safetysig = SAFETYTCP;       /* insert into chain */
    s->next = tcp_allsocs;
    tcp_allsocs = s;

    s->rtt_delay = s->rtt_smooth = 18;  /* one second startup */
       // neither of above are ever used !! Nevil, 7 Jan 97
    s->rto = MINRTO;  // Nevil, 7 Jan 97
    tcp_send(s, "tcp_open()");
    s->rtt_time = set_timeout( 1 );
    return( 1 );
}

/*
 * Passive open: listen for a connection on a particular port
 */
int tcp_listen(tcp_Socket *s, word lport, longword ina, word port, dataprocref rdataHandler, word timeout)
{
    largecheck( s, sizeof( tcp_Socket ));
    tcp_unthread(s);                        /* just in case not totally closed */
    memset( s, 0, sizeof( tcp_Socket));

    s->rdata = s->rddata;
    s->sdata = s->sddata;
    s->maxrdatalen = tcp_MaxBufSize;
    s->maxsdatalen = tcp_MaxBufSize;

    s->ip_type = IPPROTO_TCP;
    s->ttl = in_DefaultTTL;
    s->mss = _mss;
    s->cwindow = 1;
    s->wwindow = 0;     /* slow start VJ algorithm */
    s->vj_sa = 4;      /* about 250 ms */

    s->state = tcp_StateLISTEN;
    if ( !timeout ) s->timeout = 0; /* forever... */
    else s->timeout = set_timeout( timeout );
    lport = findfreeport( lport );  /* get a nonzero port val */
    s->myport = lport;
    s->hisport = port;
    s->hisaddr = ina;
    s->tunneladdr = 0;
    s->seqnum = htonl( (word)(s));          /* make up a number */
    s->flags = 0;
    s->unhappy = false;
    s->rdataHandler = rdataHandler;
    s->usr_yield = system_yield;

    s->safetysig = SAFETYTCP;       /* insert into chain */
    s->next = tcp_allsocs;
    tcp_allsocs = s;

    return( 1 );
}


static void udp_close(udp_Socket *ds)
{
    udp_Socket *s, **sp;

    sp = &udp_allsocs;
    for (;;) {
        s = *sp;
        if ( s == ds ) {
            *sp = s->next;
            break;
        }
        if ( !s ) break;
        if ( ! s->err_msg ) s->err_msg = "UDP Close called";
        sp = &s->next;
    }
}

/*
 * Send a FIN on a particular port -- only works if it is open
 *   Must still allow receives
 */
static void tcp_close(tcp_Socket *s)
{
    if ( s->ip_type != IPPROTO_TCP )
        return;
    if ( s->state == tcp_StateESTAB ||
         s->state == tcp_StateESTCL ||
         s->state == tcp_StateSYNREC ) {

       if ( s->sdatalen ) {     /* must first flush all data */
            s->flags |= tcp_FlagPUSH | tcp_FlagACK;
            if ( s->state < tcp_StateESTCL ) {
                s->state = tcp_StateESTCL;
                tcp_sendsoon(s, 0);
            }
        } else { /* really closing */
            s->flags = tcp_FlagACK | tcp_FlagFIN;
            if (!s->err_msg)
                s->err_msg = "Connection closed normally";
            s->state = tcp_StateFINWT1;
            s->timeout = set_timeout( 4 ); /* should be a pretty lengthy time */
            tcp_send(s, "tcp_close(FINWT1)");
        }
        s->unhappy = true;
    } else if (s->state == tcp_StateCLOSWT ) {
        /* need to ack the fin and get on with it */
        s->state = tcp_StateLASTACK;
        s->flags |= tcp_FlagFIN;
        tcp_send(s, "tcp_close(CLOSWT)");
        s->unhappy = true;
    }
}

/*
 * Abort a tcp connection
 */
static void tcp_abort(tcp_Socket *s)
{
    if (!s->err_msg) s->err_msg = "TCP_ABORT";
    if ( s->state != tcp_StateLISTEN && s->state != tcp_StateCLOSED ) {
        s->flags = tcp_FlagRST  | tcp_FlagACK ;
        s->unhappy = true;
        tcp_send(s, "tcp_abort()");
    }
    s->unhappy = false;
    s->sdatalen = 0;
    s->sdataoffset = 0;
    s->ip_type = 0;
    s->state = tcp_StateCLOSED;
    if (s->rdataHandler) s->rdataHandler(s, 0, -1, 0, 0);
    tcp_unthread(s);
}

void sock_abort( sock_type *s )
{
    if ( s->udp.ip_type == IPPROTO_TCP )
        tcp_abort( &s->tcp );
    else if ( s->udp.ip_type == IPPROTO_UDP )
        udp_close( &s->udp );
    // do nothing when socket is closed
}

/*
 * tcp_sendsoon - schedule a transmission pretty soon
 */
void tcp_sendsoon(tcp_Socket *s, int now)
{
    longword temp;
    if (s->ip_type == IPPROTO_TCP ) {
        temp = set_ttimeout( 1 );
// !!   if ( temp == s->rtt_time && s->rto < 2 && s->recent == 0 ) {
        if ( temp == s->rtt_time && s->rto <= MINRTO && s->recent == 0 ) {
            s->karn_count = 0;
            tcp_send(s, "tcp_sendsoon()");
            s->recent = 1;
            return;
        }
      send_now |= now;  // Nevil, 13 Jan 97
        if ((s->unhappy || s->sdatalen > 0 || s->karn_count == 1)
          && (s->rtt_time < temp ))
            return;

// !!   s->rtt_time = set_ttimeout( 1 + (s->rto >> 4) );
        s->rtt_time = set_ttimeout(0);   // Nevil, 9 Jan 97
        s->karn_count = 1;
    }
}

/*
 * Retransmitter - called periodically to perform tcp retransmissions
 */
static longword retran_strat = 0L; /* timeout retran strategy */
static void tcp_Retransmitter(void)
{
    tcp_Socket *s;
#if defined(N_DEBUG_TX) || defined(N_DEBUG_RX)
   int kc;
   char *k_state[] = {"normal", "soon", "TIMEOUT", "???"};
#endif
    /* only do this once per RETRAN_STRAT_TIME ticks */
    if ( !send_now && !chk_timeout( retran_strat ))  // Nevil, 13 Jan 97
            // send_now != 0 means "don't wait to retransmit"
        return;
    retran_strat = set_ttimeout( RETRAN_STRAT_TIME );

    for ( s = tcp_allsocs; s; s = s->next ) {
        if ( s->sdatalen > 0 || s->unhappy || s->karn_count == 1 ) {
            /* retransmission strategy */
            if ( chk_timeout( s->rtt_time)) {

#ifdef DEBUG
   if(debug_on >1) printf("regular retran TO set unacked back to 0 from %u\n", s->unacked);
#endif DEBUG
                /* strategy handles closed windows   J.D. + E.E. */
               if (s->window == 0 && s->karn_count == 2)
                  s->window = 1;

#if defined(N_DEBUG_TX) || defined(N_DEBUG_RX)
                kc = s->karn_count;
#endif
                if ( s->karn_count == 0 ) {
                    /* if really did timeout */
                    s->karn_count = 2;

                    s->unacked = 0;
                    /* use the backed off rto - implied, no code necessary */
                    /* reduce the transmit window */
// !!               s->cwindow =  ((s->cwindow + 1) * 3) >> 2;
                    s->cwindow = (s->cwindow + 1)/2;  // Nevil, 6 Jan 97
                    if (s->cwindow == 0) s->cwindow = 1;
                    s->wwindow = 0;
                }
                if (s->sdatalen)
                    s->flags |= tcp_FlagPUSH | tcp_FlagACK;
#if defined(N_DEBUG_TX) || defined(N_DEBUG_RX)
   printf(".. Retransmitter(): kc=%s, cwindow=%u, window=%u, unacked=%u\n",
      k_state[kc], s->cwindow, s->window, s->unacked);
#endif
                tcp_send(s, "tcp_Retransmitter()");
            }
        }
        /* handle inactive tcp timeouts */
        if ( sock_inactive && s->inactive_to ) {
            if ( chk_timeout( s->inactive_to)) {
                /* this baby has timed out */
                s->err_msg = "Connection timed out - no activity";
                sock_close( (sock_type *) s );
            }
        }
        if ( s->timeout && chk_timeout( s->timeout)) {
            if ( s->state == tcp_StateTIMEWT ) {
                s->state = tcp_StateCLOSED;
                tcp_unthread(s);
                break;
            } else if (s->state != tcp_StateESTAB && s->state != tcp_StateESTCL ) {
                s->err_msg = "Timeout, aborting";
                tcp_abort(s);
                break;
            }
        }
    }

    /* do our various daemons */
    if ( wattcpd ) (*wattcpd)();
}


/*
 * Unthread a socket from the tcp socket list, if it's there
 */
static void tcp_unthread(tcp_Socket *ds)
{
    tcp_Socket *s, **sp;

    if (!ds->rdatalen || (ds->state > tcp_StateESTCL))
        ds->ip_type = 0;                /* fail io */
    ds->state = tcp_StateCLOSED;   /* tcp_tick needs this */
    sp = &tcp_allsocs;
    for (;;) {
        s = *sp;
        if ( s == ds ) {
            *sp = s->next;
            continue;           /* unthread multiple copies if necessary */
        }
        if ( !s ) break;
        sp = &s->next;
    }
}

// turn off compiler warning about incorrect assignment
// since I don't want to duplicate this code inside and
// outside the while() statement
//
#pragma warn -pia

/*
 * tcp_tick - called periodically by user application
 *          - returns 0 when s closes (TCP_PROT or IPPROTO_UDP while open)
 *          - called with socket parameter or NULL
 */
int tcp_tick( sock_type *s )
{
    in_Header *ip;
    int tcp_h_result;
    word packettype;

    /* finish off dead sockets */
    if ( s ) {
        if (( s->tcp.ip_type == IPPROTO_TCP ) &&
            ( s->tcp.state == tcp_StateCLOSED ) &&
            ( s->tcp.rdatalen == 0 )) {
                tcp_unthread( & s->tcp );
                s->tcp.ip_type = 0;
        }
    }

   send_now = 0;  // Nevil, 13 Jan 97
   tcp_h_result = 0;  // Nevil, 13 Jan 97

    while( ip = (in_Header *)_eth_arrived( &packettype ) )
    {
        switch ( packettype ) {
            case /*0x800*/ 0x008 :
                /* do IP */
                if( checksum(ip, in_GetHdrlenBytes(ip)) == 0xffff ) {
                    switch ( ip->proto ) {
                        case IPPROTO_TCP :
                            tcp_h_result = tcp_handler(ip);
                            break;
                        case IPPROTO_UDP :
                            udp_handler(ip);
                            break;
                        case IPPROTO_ICMP :
                            icmp_handler(ip);
                            break;
                    }
                } else  {
                    if (debug_on)
                        outs("IP Received BAD Checksum \n");
                }
                break;

            case /*0x806*/ 0x608 :
                /* do arp */
                _arp_handler( (arp_Header *)ip );
                break;
        }

        if( ip )
            _eth_free( ip );
    if (tcp_h_result) break;  // Nevil, 8 Jan 97
       // Let the user program have some processor cycles!
    }

    /* check for our outstanding packets */
    tcp_Retransmitter();

    return s ? s->udp.ip_type : 0;  /* Nevil, 31 Mar 93 */
}

// turn compiler warning about incorrect assignment
// back to whatever it was before I messed with it
//
#pragma warn .pia

void tcp_set_debug_state( int x )
{
    debug_on = x;
}

/* returns 1 if connection is established */
int tcp_established(tcp_Socket *s)
{
    return( s->state >= tcp_StateESTAB );
}


static udp_write(udp_Socket *s, byte *datap, int len)
{
    tcp_PseudoHeader ph;
    struct _pkt {
        in_Header  in;
        udp_Header udp;
        int        data;
/*      longword maxsegopt; */
    } *pkt;
    struct _tunnel {
        in_Header  in;
        struct _pkt pkt;
    } *tunnel;
    byte *dp;
    in_Header *inp;
    udp_Header *udpp;

    if( len < 0 )
        len = MAXINT;

    if( s->tunneladdr )
    {
        tunnel = (struct _tunnel *)_eth_formatpacket(&s->hisethaddr[0], /*0x800*/ 8);
        pkt = &tunnel->pkt;
        if( len > ETH_MAXUDP - sizeof(in_Header) )
            len = ETH_MAXUDP - sizeof(in_Header);
    }
    else
    {
        pkt = (struct _pkt *)_eth_formatpacket(&s->hisethaddr[0], /*0x800*/ 8);
        if( len > ETH_MAXUDP )
            len = ETH_MAXUDP;
    }
    dp = (byte *) &pkt->data;
    inp = &pkt->in;
    udpp = &pkt->udp;

    memset( inp, 0, sizeof( in_Header ));
    inp->length = htons( sizeof(in_Header) + UDP_LENGTH + len );

    /* udp header */
    udpp->srcPort = htons( s->myport );
    udpp->dstPort = htons( s->hisport );
    udpp->checksum = 0;
    udpp->length = htons( UDP_LENGTH + len );
    movmem(datap, dp, len );

    /* internet header */
    inp->ver = 4;
    inp->hdrlen = 5;
    inp->tos = 0;
    inp->identification = htons( ++ip_id );   /* was post inc */
//    inp->frag = 0;
    inp->ttl = s->ttl;
    inp->proto = IPPROTO_UDP;     /* udp */
    inp->checksum = 0;
    inp->source = htonl( s->myaddr );
    inp->destination = htonl( s->hisaddr );
    inp->checksum = ~checksum( inp, sizeof(in_Header));

    /* compute udp checksum if desired */
    if ( s->sock_mode & UDP_MODE_NOCHK )
        udpp->checksum = 0;
    else {
        ph.src = inp->source;   /* already INTELled */
        ph.dst = inp->destination;
        ph.mbz = 0;
        ph.protocol = IPPROTO_UDP;        /* udp */
        ph.length = udpp->length;       /* already INTELled */

        ph.checksum = checksum(&pkt->udp, ntohs(ph.length));
        udpp->checksum =  ~checksum(&ph, sizeof(ph));
    }

    if (_dbugxmit)
        (*_dbugxmit)((sock_type *)s,inp,(tcp_Header *)udpp, "udp_write()");

    if (s->tunneladdr)
    {
        inp = &tunnel->in;

        inp->ver = 4;
        inp->hdrlen = 5;
        inp->tos = 0;
        inp->identification = htons( ++ip_id );   /* was post inc */
        inp->frags = 0;
        inp->fo = 0;
        inp->ttl = in_DefaultTTL;
        inp->proto = IPPROTO_IP_IN_IP;      /* please unwrap me! */
        inp->source = htonl( s->myaddr );
        inp->destination = htonl( s->tunneladdr );
        inp->length = htons( sizeof(in_Header) + ntohs(tunnel->pkt.in.length) );
        inp->checksum = 0;
        inp->checksum = ~checksum( inp, sizeof(in_Header));
    }

    // if the send fails, tell the caller
    //
    if( _eth_send( ntohs( inp->length ) ) )
        len = 0;

    return ( len );
}

/*
 * udp_read - read data from buffer, does large buffering
 */
static int udp_read(udp_Socket *s, byte *datap, int maxlen)
{
    int x;
    int up_to_end;

    if (maxlen < 0) maxlen = MAXINT;
    if (( x = s->rdatalen ) > 0) {
        if ( x > maxlen ) x = maxlen;
        if ( x > 0 ) {
            if (datap) {
                up_to_end = s->maxrdatalen - s->rdataoffset;
                if (up_to_end < x)
                {
                    movmem( s->rdata + s->rdataoffset, datap, up_to_end );
                    movmem( s->rdata, datap+up_to_end, x-up_to_end );
                }
                else
                    movmem( s->rdata + s->rdataoffset, datap, x );
            }

            if ((s->rdatalen -= x ) > 0) {  // Still some data in buffer
                s->rdataoffset = wrap( s->rdataoffset + x, s->maxrdatalen );
            }
        }
    }
    return( x );
}

void _udp_cancel( in_Header *ip )
{
    int len;
    udp_Header *up;
    udp_Socket *s;

    /* match to a udp socket */
    len = in_GetHdrlenBytes(ip);
    up = (udp_Header *)((byte *)ip + len);      /* udp frame pointer */

    /* demux to active sockets */
    for ( s = udp_allsocs; s; s = s->next )
        if ( s->hisport != 0 &&
             ntohs( up->dstPort ) == s->hisport &&
             ntohs( up->srcPort ) == s->myport &&
             ntohl( ip->destination ) == s->hisaddr ) break;
    if ( !s ) {
        /* demux to passive sockets */
        for ( s = udp_allsocs; s; s = s->next )
            if ( s->hisport == 0 && ntohs( up->dstPort ) == s->myport ) break;
    }
    if (s) {
        s->rdatalen = 0;
        s->rdataoffset = 0;
        s->ip_type = 0;
    }
}

void *_tcp_lookup( longword hisip, word hisport, word myport )
{
    tcp_Socket *s;
    for ( s = tcp_allsocs; s; s = s->next ) {
        if ( ( myport == s->myport ) &&         /* always unique under WATTCP */
             ( hisport == s->hisport ) &&
             ( hisip == s->hisaddr ))
                return( s );
    }
    return( NULL );
}


void _tcp_cancel(in_Header *ip, int code, char *msg, longword dummyip )
{
    int len;
    tcp_Socket *s;
    tcp_Header *tp;

    len = in_GetHdrlenBytes(ip);        /* check work */

    tp = (tcp_Header *)((byte *)ip + len);      /* tcp frame pointer */

    /* demux to active sockets */
    for ( s = tcp_allsocs; s; s = s->next )
    {
        if ( ntohs( tp->srcPort) == s->myport &&
             ntohs( tp->dstPort ) == s->hisport &&
             ntohl( ip->destination ) == s->hisaddr )
        {
            switch (code) {
                /* halt it */
                case  1 :
                    if (( s->stress++ > s->rigid ) && ( s->rigid < 100 )) {
                        s->err_msg = (msg) ? msg : "ICMP closed connection";
                        s->rdatalen = s->sdatalen = 0;
                        s->rdataoffset = s->sdataoffset = 0;
                        s->unhappy = false;
                        tcp_abort( s );

                        // don't need to call s->rdataHandler()
                        // because tcp_abort() does that

                        break;
                    }

                    // otherwise fall through to next case

                /* slow it down */
                case  2 :
                    s->cwindow = 1;
                    s->wwindow = 1;
                    s->rto <<= 2;
                    if (s->rto < MINRTO) s->rto = MINRTO;
                    s->vj_sa <<= 2;
                    s->vj_sd <<= 2;
                    break;

                /* icmp redirect for host */
                case  5 :
                    /* save his NEW network address */
                    _arp_resolve(dummyip, &s->hisethaddr, 0);
                    break;

            }   // end
        }   // end if IP dest and both TCP ports match this socket
    }   // end for all sockets
}

static int tcp_read(tcp_Socket *s, byte *datap, int maxlen)
{
    int x;
    int up_to_end;

    if (maxlen < 0 ) maxlen = MAXINT;
    if ((x = s->rdatalen) > 0) {
      if (x > maxlen) x = maxlen;
      if (x > 0) {
         if (datap) {
             up_to_end = s->maxrdatalen - s->rdataoffset;
             if (up_to_end < x) {
                 movmem( s->rdata + s->rdataoffset, datap, up_to_end );
                 movmem( s->rdata, datap+up_to_end, x-up_to_end );
             }
             else
                 movmem( s->rdata + s->rdataoffset, datap, x );
         }
         if ((s->rdatalen -= x ) > 0) {  // Still some data in buffer
            s->rdataoffset = wrap( s->rdataoffset + x, s->maxrdatalen );
#if 0
            if (calcwindow(s) != s->oldwindow) {
#ifdef J_DEBUG_READ_WINDOW_UPDATE
               printf( "tcp_read() updating window soon\n" );
#endif
            tcp_sendsoon(s, 0);   /* Update the window */
               }
            }
         else { // Buffer now empty
#ifdef J_DEBUG_READ_WINDOW_UPDATE
            printf( " maybe soon\n" );
#endif
            if (calcwindow(s) != s->oldwindow) {
#ifdef J_DEBUG_READ_WINDOW_UPDATE
               printf( "tcp_read() updating window now\n" );
#endif
               tcp_send(s, "tcp_read()");  /* Update window el-pronto */
               }
            }

   /* Updating the window el-pronto tells the peer we can take
      more data right now, but it sends extra packets to do this.
      Always using sendsoon() - below - reduced the number of
      packets sent in a telnet session by about 40%, with no
      observable effect on performance.   Nevil, 20 April 93 */

   // Unfortunately, always using send_soon can leave us
   // with a large data transfer being paced by the sendsoon
   // timeouts at each end!  Better to live with a few more
   // packets during a telnet session.    Nevil, 9 Jan 97

   // need to avoid updating the window when our buffer is much
   // bigger than the window size can indicate (probably because
   // I haven't implemented the TCP window scaling option yet)
   // otherwise the receiver may think we want fast-retransmit
   //                                     Joel, 17 Jan 97

   // in general, should avoid updating the window when it
   // hasn't changed enough to be visible in the TCP window
   // field, even if scaling were implemented -- best way to
   // do that will be to keep a copy of last window we sent,
   // precalculate the window here, and compare them
   //                                     Joel, 17 Jan 97
#else
            }
         if (calcwindow(s) != s->oldwindow) {
              // s->oldwindow was set in tcp_send()
#ifdef J_DEBUG_READ_WINDOW_UPDATE
            printf( "tcp_read() updating window soon\n" );
#endif
            tcp_sendsoon(s, 0);   /* Update the window */
            }
#endif
         }
      }
   else if (s->state == tcp_StateCLOSWT)
      tcp_close(s);
#ifdef N_DEBUG_RX
   if (x != 0) printf(
      "<= tcp_read(): read %u, s->rdatalen=%u\n", x, s->rdatalen);
#endif
   return x;
   }

/*
 * Write data to a connection.
 * Returns number of bytes written, == 0 when connection is not in
 * established state.
 */
static tcp_write(tcp_Socket *s, byte *dp, int len)
{
    int x;
    int up_to_end;
    int startpos;

    /* no longer uses tcp_MaxBufSize */
    if (len < 0 ) len = MAXINT;

    if ( s->state != tcp_StateESTAB ) len = 0;
    if ( len > (x = s->maxsdatalen - s->sdatalen) ) len = x;
    if ( len > 0 ) {
        startpos = wrap( s->sdataoffset + s->sdatalen, s->maxsdatalen );
        up_to_end = s->maxsdatalen - startpos;
        if( up_to_end < len ) {
            movmem( dp, s->sdata + startpos, up_to_end );
            movmem( dp+up_to_end, s->sdata, len-up_to_end );
        }
        else
            movmem( dp, s->sdata + startpos, len );

        // note: data offset doesn't change when we add to buffer

        s->sdatalen += len;
#ifdef N_DEBUG_TX
   printf("=> tcp_write(): write %u, s->sdatalen=%u\n", len, s->sdatalen);
#endif
        s->unhappy = true;      /* redundant because we have outstanding data */

        if ( s->sock_mode & TCP_LOCAL )
            s->sock_mode &= ~TCP_LOCAL;
        else {
            if ( s->sock_mode & TCP_MODE_NONAGLE ) {
                tcp_send(s, "tcp_write(NN)");  /* Nevil: no tcp_send() ! Jan 93 */
            } else {
                /* transmit if first data or reached MTU */
                /* not true MTU, but better than nothing */
                if (s->sdatalen == len)  // First data in buffer
                    tcp_send(s, "tcp_write()");  // Nevil, 8 Jan 97
                else
                    tcp_sendsoon(s, 0 );
            }
        }
    }

    return ( len );
}

/*
 * Send pending data
 */
static void tcp_Flush(tcp_Socket *s)
{
    if ( s->sdatalen > 0 ) {
        s->flags |= tcp_FlagPUSH;
        tcp_send(s, "tcp_Flush()");
    }
}

/*
 * Handler for incoming packets.
 */
static void udp_handler(in_Header *ip)
{
    udp_Header *up;
    tcp_PseudoHeader ph;
    word len;
    byte *dp;
    longword temp;
    udp_Socket *s;

    temp = ntohl( ip->destination );

    // temp = ip number
    //     or 255.255.255.255
    //     or sin_mask.255.255

    if ( ((~temp & ~sin_mask) != 0) &&  /* not a broadcast packet*/
        ((( temp - my_ip_addr) > multihomes )   /* not my address */
        && my_ip_addr))                 /* and I know my address */
          return;


    len = in_GetHdrlenBytes(ip);
    up = (udp_Header *)((byte *)ip + len);      /* udp segment pointer */
    len = ntohs( up->length );

    /* demux to active sockets */
    for ( s = udp_allsocs; s; s = s->next ) {
        if ( s->safetysig != SAFETYUDP ) {
            if (debug_on) outs("chain error in udp\r\n");
        }
        if ( (s->hisport != 0) &&
             (ntohs( up->dstPort ) == s->myport) &&
             (ntohs( up->srcPort ) == s->hisport) &&
             ((ntohl( ip->destination ) & sin_mask)  == (s->myaddr & sin_mask)) &&
             (ntohl( ip->source ) == s->hisaddr )) break;
    }
    if (_dbugrecv)
       (*_dbugrecv)((sock_type *)s,ip,(tcp_Header *)up,  "udp_handler()");
    if ( !s ) {
        /* demux to passive sockets */
        for ( s = udp_allsocs; s; s = s->next )
            if ( ((s->hisaddr == 0) || (s->hisaddr == 0xffffffff))
              && ntohs( up->dstPort ) == s->myport ) {

#ifdef NEVER   /* Original code: _arp_resolve fails because we haven't
                    finished with this ip packet, it's still at the head
                    of the queue.  As a result we just loop sending
                    (and receiving) ARP packets.  Nevil,  Thu 2 Jul 92 */

                // do we record this information ???
                if ( s->hisaddr == 0 ) {
                    s->hisaddr = ntohl( ip->source );
                    s->hisport = ntohs( up->srcPort );
                    _arp_resolve(ntohl(ip->source), &s->hisethaddr[0], 0);
                    // take on value of expected destination unless it
                    // is broadcast
                    if ( (~ip->destination & ~sin_mask) != 0 )
                        s->myaddr = ntohl( ip->destination );
                }
#endif
                movmem(((eth_Header *)ip - 1)->source,
                   s->hisethaddr, sizeof(eth_address));
                s->hisaddr = ntohl( ip->source );
                s->hisport = ntohs( up->srcPort );

/*                  if ( (~ip->destination & ~sin_mask) != 0 )
                        s->myaddr = ntohl( ip->destination );
    Don't do this here - it was set by udp_open.  Nevil, Thu 2 Jul 92 */

                break;
            }
    }
    if ( !s ) {
        /* demux to broadcast sockets */
        for ( s = udp_allsocs; s; s = s->next )
            if ( (s->hisaddr == 0xffffffff) &&
                 (ntohs( up->dstPort ) == s->myport )) break;
    }

    if ( !s ) {
        if (debug_on > 2) outs("discarding...");
        return;
    }

    // these parameters are used for things other than just checksums
    ph.src = ip->source;    /* already INTELled */
    ph.dst = ip->destination;
    ph.mbz = 0;
    ph.protocol = IPPROTO_UDP;
    ph.length = up->length;
    if ( up->checksum ) {
        ph.checksum =  checksum(up, len);
        if (checksum(&ph, sizeof( tcp_PseudoHeader)) != 0xffff)
            return;
    }

    /* process user data */
    if ( (len -= UDP_LENGTH ) > 0) {
        dp = (byte *)( up );
        if (s->rdataHandler)
            s->rdataHandler( s, &dp[UDP_LENGTH], len, &ph, up);
        else {
            // just overwrite the beginning of buffer...
            //
            // should this preserve existing data?
            //
            if (len > s->maxrdatalen ) len = s->maxrdatalen;
            movmem( &dp[ UDP_LENGTH ], s->rdata, len );
            s->rdatalen = len;
            s->rdataoffset = 0;
        }
    }
}

// !! static void tcp_handler(in_Header *ip)
static int tcp_handler(in_Header *ip)
{
    tcp_Header *tp;
    tcp_PseudoHeader ph;
    int len;
    byte *dp;
    int diff;
    tcp_Socket *s;
    word flags;
    long diffticks, ldiff;      /* must be signed */
    long scheduleto;
   int his_window_changed;  // Nevil, 13 Jan 97

    if ( (longword)(ntohl( ip->destination ) - my_ip_addr) > multihomes )
        return 0;

    len = in_GetHdrlenBytes(ip);
    tp = (tcp_Header *)((byte *)ip + len);      /* tcp frame pointer */
    len = ntohs( ip->length ) - len;          /* len of tcp data */
    flags = ntohs( tp->flags );

    if (debug_on >= 1) {
            // display stuff on the far right of the second screen line
            //
            mono[310]++;
            colour[310]++;
            mono[312] = colour[312] = (flags & tcp_FlagSYN) ? 'S' : ' ';
            mono[314] = colour[314] = (flags & tcp_FlagACK) ? 'A' : ' ';
            mono[316] = colour[316] = (flags & tcp_FlagFIN) ? 'F' : ' ';
            mono[318] = colour[318] = (flags & tcp_FlagRST) ? 'R' : ' ';
        }

    /* demux to active sockets */
    for ( s = tcp_allsocs; s; s = s->next ) {
        if ( s->safetysig != SAFETYTCP ) {
            if (debug_on) outs("chain error in tcp\r\n");
        }
        if ( s->hisport != 0 &&
             ntohs( tp->dstPort ) == s->myport &&
             ntohs( tp->srcPort ) == s->hisport &&
             ntohl( ip->destination )   == s->myaddr     &&
             ntohl( ip->source ) == s->hisaddr ) break;
    }

    if ( !s && (flags & tcp_FlagSYN)) {
        /* demux to passive sockets, must be a new session */
        for ( s = tcp_allsocs; s; s = s->next )
            if ((s->hisport == 0) && (ntohs( tp->dstPort ) == s->myport )) {
                s->myaddr = ntohl( ip->destination );
                break;
            }
    }

    if (_dbugrecv) (*_dbugrecv)((sock_type *)s,ip,tp, "tcp_handler()");
    if ( !s ) {
        if (!(flags & tcp_FlagRST)) tcp_rst( ip, tp );
        return 0;
    }

    ph.src = ip->source;        /* already INTELled */
    ph.dst = ip->destination;
    ph.mbz = 0;
    ph.protocol = IPPROTO_TCP;
    ph.length = htons( len );
    ph.checksum =  checksum(tp, len);
    if ( checksum(&ph, sizeof(ph)) != 0xffff ) {
        if (debug_on) printf(
            "bad tcp checksum for packet at log/phys %.8X/%.8X len %X\n",
#ifdef __DPMI32__
            tp, pkt_flat2phys(tp), len);
#else
            tp, tp, len);
#endif
        tcp_sendsoon(s, 1);  // Send 'soon'
        return 0;
    }

/* reset code */
    if ( flags & tcp_FlagRST ) {
        if (debug_on) outs("\7\7\7\7\7\7\7connection reset\n");
        s->rdatalen = 0;
        s->rdataoffset = 0;
        s->err_msg = "Remote reset connection";
        s->state = tcp_StateCLOSED;
        tcp_abort(s);
        return 0;
    }

    if ( sock_inactive )
        s->inactive_to = set_timeout( sock_inactive );


    /* update our retransmission stuff */
    /* new algorithms */
    if (s->karn_count == 2) {
        s->karn_count = 0;
#ifdef DEBUG
        if (debug_on > 1 ) printf("finally got it safely zapped from %u to ????\n",s->unacked);
#endif /* DEBUG */
    } else {
        if ( s->vj_last ) {
            /* unnecessary to use unhappy || s->sdatalen ) */
            if ((diffticks = set_ttimeout(0) - s->vj_last) >= 0 ) {
                /* we ignore the overnight case */
                diffticks -= (longword)s->vj_sa >> 3;
                s->vj_sa += (int)diffticks;
                if (diffticks < 0) diffticks = -diffticks;
                diffticks -= s->vj_sd >> 2;
                s->vj_sd += (int)diffticks;
                if (s->vj_sa > MAXVJSA) s->vj_sa = MAXVJSA;
                if (s->vj_sd > MAXVJSD) s->vj_sd = MAXVJSD;
            }
            /* only recompute rtt hence rto after success */
//            s->rto = 1 + ((s->vj_sa >> 2) + (s->vj_sd)) >> 1 ;
            s->rto = MINRTO + ((s->vj_sa >> 2) + s->vj_sd) >> 1 ;
            // MINRTO = 2 ticks (111 ms).  Nevil, 3 Jan 97
            // Dramatically reduces nbr of timeouts on my Ethernet
            if (s->rto > MAXRTO) s->rto = MAXRTO;  // Nevil, 13 Jan 97
        }
      s->karn_count = 0;
      // Original code kept cwindow = 1 !!
      if (++s->wwindow >= s->cwindow) {  // Nevil, 9 Jan 97
         if (s->cwindow < MAXPKTSEND)
            ++s->cwindow;  // Don't let cwindow get too big.
            // If it does, the backoff takes too long to happen
         s->wwindow = 0;
         }

#if defined(N_DEBUG_TX) || defined(N_DEBUG_RX) || defined(DEBUG)
   if (debug_on > 1)
      printf("-- rto=%u  sa=%u  sd=%u   cwindow=%u  wwindow=%u  unacked=%u\n",
         s->rto, s->vj_sa, s->vj_sd, s->cwindow, s->wwindow, s->unacked);
#endif
    }

    /* all new */
    scheduleto = set_ttimeout(s->rto);
    if ( s->rtt_time < scheduleto ) s->rtt_time = scheduleto;

    switch ( s->state ) {

        case tcp_StateLISTEN:   /* accepting SYNs */
            /* save his ethernet address */
            if ( _pktipofs )
                movmem(&((((eth_Header *)ip) - 1)->source[0]), &s->hisethaddr[0], sizeof(eth_address));
            if ( flags & tcp_FlagSYN ) {

                if ( ip->tos > s->tos )
                    s->tos = ip->tos;
                else if ( ip->tos < s->tos ) {
                    /* RFC 793 says we should close connection */
                    /* we best not do that while SunOS ignores TOS */
                }

                s->acknum = ntohl( tp->seqnum ) + 1;
                s->hisport = ntohs( tp->srcPort );
                s->hisaddr = ntohl( ip->source );
                s->flags = tcp_FlagSYN | tcp_FlagACK;
                s->state = tcp_StateSYNREC;
                s->unhappy = true;
                tcp_send(s, "tcp(LISTEN)");  // Must respond immediately

                s->timeout = set_timeout( tcp_TIMEOUT );
            } else
                tcp_rst( ip , tp );  /* send a reset */

            return 0;

        case tcp_StateSYNSENT:  /* added ACK Section */
            if ( flags & tcp_FlagSYN ) {

                if ( ip->tos > s->tos )
                    s->tos = ip->tos;
                else if ( ip->tos < s->tos ) {
                    /* RFC 793 says we should close connection */
                    /* we best not do that while SunOS ignores TOS */
                }

                s->flags = tcp_FlagACK;
                s->timeout = set_timeout( tcp_TIMEOUT );

                /* FlagACK means connection established, else SYNREC */
                if ( flags & tcp_FlagACK) {
                    /* but is it for the correct session ? */
                if (tp->acknum == ntohl(s->seqnum + 1)) {
                s->state = tcp_StateESTAB;
                s->seqnum++;    /* good increment */
                s->acknum = ntohl( tp->seqnum  ) + 1;   /* 32 bits */
                tcp_ProcessData(s, tp, len);    /* someone may try it */
                s->unhappy = true;             /* rely on their attempts */
                tcp_send(s, "tcp(SYNSENT)");
            } else {
                /* wrong ack, force a RST and resend SYN soon*/
                s->flags = tcp_FlagRST;
                s->unhappy = true;
                tcp_send(s, "tcp(SYN:RST)");
                s->flags = tcp_FlagSYN;
                tcp_send(s, "tcp(SYN:SYN)");
                    }
                } else {
                    s->acknum++;
                    s->state = tcp_StateSYNREC;
                    return 0;
                }
            } else
                tcp_rst( ip, tp );
            break;

        case tcp_StateSYNREC:   /* recSYNSENT, sentACK, waiting  EST */
            if ( flags & tcp_FlagSYN ) {
                s->flags = tcp_FlagSYN | tcp_FlagACK;
                s->unhappy = true;
                tcp_send(s, "tcp(SYNREC)");
                s->timeout = set_timeout( tcp_TIMEOUT );
                return 0;
            }
            if ( (flags & tcp_FlagACK) && (ntohl( tp->acknum ) == (s->seqnum + 1))) {
                s->window = ntohs( tp->window );
                s->flags = tcp_FlagACK;
                s->state = tcp_StateESTAB;
                s->seqnum++;
                s->timeout = 0;     /* never timeout */
                s->unhappy = false;
                return 0;
            }

            break;
        case tcp_StateESTAB:
        case tcp_StateESTCL:
        case tcp_StateCLOSWT:

            /* handle lost SYN */
            if ((flags & tcp_FlagSYN) && (flags & tcp_FlagACK)) {
                tcp_send(s, "tcp(ESTAB:lost)");
                return 0;
            }

            if ( !(flags & tcp_FlagACK)) return 0;  /* must ack somthing */
            if ( flags & tcp_FlagSYN ) {
                tcp_rst( ip , tp );
                return 0;
            }
            s->timeout = 0L;  /* We do not timeout at this point */

            /* process ack value in packet - but only if it falls
             * within current window */

            ldiff = ntohl( tp->acknum ) - s->seqnum;
            diff = (int) ldiff;
            his_window_changed =  // Nevil, 13 Jan 97
               ntohs(tp->window) != s->window;

            if (ldiff > 0) {  // Nevil, 9 Jan 97
                if (ldiff <= s->maxunacked) {  // Ack of (sent but) unacked data
                    s->sdatalen -= diff;
                    s->maxunacked -= diff;
                    s->unacked -= diff;
                    if (s->unacked < 0) s->unacked = 0;
                    s->sdataoffset = wrap( s->sdataoffset + diff, s->maxsdatalen );
                    s->seqnum += ldiff;
                }
                else {  // Ack of unsent data
                    // This is an error!  Just ignore the ack
#ifdef DEBUG
                    if(debug_on >1) printf("tcphandler confused, but didn't set unacked back to 0 from %u\n",s->unacked);
#endif DEBUG
#ifdef N_DEBUG_TX
                    printf(">> Ack of unsent data <<\n");
#endif
                    // !!    s->unacked = 0;
                    // Don't force resend of whole window!  Nevil, 6 Jan 97
                }
            }
#if ALLOW_FAST_RETRANSMIT
            else if (ldiff == 0) {  // Zero-length ack
                // Is sender's window size the same?
                if (ntohs(tp->window) == s->window) {

                    // I would like to take this opportunity to document
                    // how fast retransmit is SUPPOSED to work, as opposed
                    // to the way this cruddy stack actually implements it
                    // 2/12/97 Joel Apisdorf
                    //
                    // (1) fast retransmit is defined as when you get 3
                    // duplicate acknowledgements ???in a row???, not 1!
                    //
                    // (2) (I think) it doesn't matter whether window size
                    // advertizement has changed!
                    //
                    // (3) you are only supposed to retransmit one segment
                    // starting where the ACK says, not the whole window
                    // as if you had timed out!
                    //
                    // (4) you are supposed to halve the ????congestion??? window,
                    // not set it all the way down to one packet, as if you
                    // had timed out!
                    //

                    // and another thing: receiver is supposed to keep track
                    // of segments it receives out-of-order, so when it finally
                    // gets the segment(s) that fill in the hole, it can ACK
                    // all the data (up to the next hole), so sender doesn't
                    // have to resend data that was received properly
                    //
                    // this also helps you implement the selective-acknowledge
                    // (SACK) TCP option
                    //
                    // this TCP implementation isn't so nice!
                    //
                    // 2/12/97 Joel Apisdorf
                    //

#ifdef J_DEBUG_FAST_RETRANSMIT
                    printf( "tcp_handler(ESTAB) doing fast retransmit\n" );
#endif
                    // No.  Should mean 'please resend the whole buffer'
                    s->unacked = 0;
                    s->cwindow = 1;  // Backoff send window size too
                    s->wwindow = 0;
                }
            }
#endif
            // else ldiff < 0.  Already acked, ignore the ack

#ifdef N_DEBUG_TX
            printf("-> s_unack=%u, s_datalen=%u,  ldiff=%ld\n",
               s->unacked, s->sdatalen, ldiff);
#endif

            s->flags = tcp_FlagACK;
            //len =
            tcp_ProcessData(s, tp, len);

            if (( flags & tcp_FlagFIN ) && (s->state != tcp_StateCLOSWT )
            && ( s->acknum == ntohl( tp->seqnum ))) {
                s->acknum ++;
                if ( ! s->err_msg ) s->err_msg = "Connection closed";
                s->state = tcp_StateCLOSWT;
                tcp_send(s, "tcp(ESTAB:CLOS)");
                s->state = tcp_StateLASTACK;
                s->flags |= tcp_FlagFIN;
                s->unhappy = true;
            }

    // if ( diff > 0 || ((len > 0) && (calcwindow(s) != s->oldwindow)) ) {

            if (his_window_changed ||    // Nevil, 13 Jan 97
                    diff > 0 || len > 0) {
#ifdef J_DEBUG_RX_WINDOW_UPDATE
                printf( "tcp_handler(ESTAB) updating window" );
#endif
                /* need to update window, but how urgent ??? */
#if 0 // Original method
                if (diff > 0 && len > s->mss/2) {
#ifdef J_DEBUG_RX_WINDOW_UPDATE
                    printf( " now\n" );
#endif
                    tcp_send(s, "tcp(ESTAB)");
                } else {
#ifdef J_DEBUG_RX_WINDOW_UPDATE
                    printf( " soon\n" );
#endif
                    tcp_sendsoon(s, 0);
                }
#endif
                tcp_sendsoon(s, 1);  // Nevil, 13 Jan 97
                   // Send 'now,' i.e. afer emptying the readyq
            }

            if ( s->state == tcp_StateESTCL )
                tcp_close( s );
#ifdef N_DEBUG_TX_RX_EXTRA
            printf("leaving ESTAB: seq=%lu, ack=%lu, flags=%x, state=%d\n",
                s->seqnum, s->acknum, s->flags, s->state);
#endif
            return 0;
        case tcp_StateFINWT1:
            /* They have not necessarily read all the data yet, we must
               still supply it as requested */

            ldiff = ntohl( tp->acknum ) - s->seqnum;
            diff = (int) ldiff;
#ifdef N_DEBUG_UPDOWN
printf("rcv FINWT1: tp->acknum=%lu, diff=%d, s->sdatalen=%d, len=%d, flags=%x\n",
   ntohl(tp->acknum), diff, s->sdatalen, len, flags);
#endif
            if ( ldiff >= 0 && diff <= s->maxunacked ) {
                s->sdatalen -= diff;
                s->maxunacked -= diff;
                s->unacked -= diff;
                if (s->sdatalen < 0) s->sdatalen = 0;
                s->sdataoffset = wrap( s->sdataoffset + diff, s->maxsdatalen );
                s->seqnum += ldiff;
                if (ldiff == 0 || s->unacked < 0) s->unacked = 0;
            }

            /* they may still be transmitting data, we must read it */
            tcp_ProcessData(s, tp, len);

            /* check if other tcp has acked all sent data and is ready
               to change states */

            if ( (flags & (tcp_FlagFIN|tcp_FlagACK) ) == (tcp_FlagFIN|tcp_FlagACK)) {
                /* trying to do simultaneous close */
#ifdef N_DEBUG_UPDOWN
printf("FINWT1: simultaneous close\n");
#endif
                if (( ntohl( tp->acknum ) >= s->seqnum + 1 )
                &&  ( ntohl( tp->seqnum) == s->acknum ))
                {
                    s->seqnum++;  /* He's acked our fin */
                    s->acknum++;
                    s->flags = tcp_FlagACK;
                    tcp_send(s, "tcp(FINWT1)");
                    s->unhappy = false;
                    s->timeout = set_timeout( 2*tcp_TIMEOUT );
                    s->state = tcp_StateTIMEWT;
                }
            } else if ( flags & tcp_FlagACK ) {
#ifdef N_DEBUG_UPDOWN
printf("FINWT1: acking our fin\n");
#endif
                /* other side is legitimately acking our fin */
                if (( ntohl( tp->acknum ) == s->seqnum + 1 ) &&
                    ( ntohl( tp->seqnum ) == s->acknum ) &&
                    (  s->sdatalen == 0 )) {
                        s->seqnum++;
// they are just acking our seq num, not sending more data for us to ack
//                      s->acknum++;  Do it in FINWT2.  Nevil, 22 Apr 93
                        s->state = tcp_StateFINWT2;
                        s->timeout = set_timeout( 3 );
                        s->unhappy = false; /* we don't send anything */
                }
            }
            break;

        case tcp_StateFINWT2:
#ifdef N_DEBUG_UPDOWN
printf("rcv FINWT2: s->sdatalen=%d, len=%d, flags=%x\n", s->sdatalen, len, flags);
#endif
            /* they may still be transmitting data, we must read it */
            tcp_ProcessData(s, tp, len);

            if ((flags & (tcp_FlagACK | tcp_FlagFIN)) ==
                  (tcp_FlagACK | tcp_FlagFIN)) {
                if (( ntohl( tp->acknum ) == s->seqnum) &&
                    ( ntohl( tp->seqnum ) == s->acknum )) {
                    s->acknum++;  /* Now we have his fin */
                    s->flags = tcp_FlagACK;
                    tcp_send(s, "tcp(FINWT2)");
                    s->unhappy = false; /* we don't send anything */
                    s->timeout = set_timeout( 2 );
                    s->state = tcp_StateCLOSED;
                    return 0;
                }
            }
            break;

        case tcp_StateCLOSING:
#ifdef N_DEBUG_UPDOWN
printf("rcv CLOSING: s->sdatalen=%d, len=%d, flags=%x\n", s->sdatalen, len, flags);
#endif
            if ((flags & (tcp_FlagACK | tcp_FlagFIN)) == tcp_FlagACK ) {
                if (( tp->acknum == ntohl(s->seqnum) ) &&
                    ( tp->seqnum == ntohl(s->acknum))) {
                    s->state = tcp_StateTIMEWT;
                    s->timeout = set_timeout( tcp_TIMEOUT );
                    s->unhappy = false;
                }
            }
            break;

        case tcp_StateLASTACK:
#ifdef N_DEBUG_UPDOWN
printf("rcv LASTACK: s->sdatalen=%d, len=%d, flags=%x\n",
   s->sdatalen,len, flags);
#endif
            if ( flags & tcp_FlagFIN ) {
                /* they lost our two packets, back up */
                s->flags = tcp_FlagACK | tcp_FlagFIN;
                tcp_send(s, "tcp(LASTACK)");
                s->unhappy = TRUE;  /* FALSE; */
                return 0;
            } else {
                if (( ntohl( tp->acknum ) == (s->seqnum + 1 )) &&
                    ( ntohl( tp->seqnum ) == s->acknum )) {
                        s->state = tcp_StateCLOSED;     /* no 2msl necessary */
                        s->unhappy = false;             /* we're done */
                        return 0;
                    }
            }
            break;

        case tcp_StateTIMEWT:
#ifdef N_DEBUG_UPDOWN
printf("rcv TIMEWT: s->sdatalen=%d, len=%d, flags=%x\n",
   s->sdatalen,len, flags);
#endif
/*            if ( flags & (tcp_FlagACK | tcp_FlagFIN) == tcp_FlagACK | tcp_FlagFIN) { !!! Nevil 22 Apr 93 */
            if ((flags & (tcp_FlagACK | tcp_FlagFIN)) == (tcp_FlagACK | tcp_FlagFIN)) {
                /* he needs an ack */
                s->flags = tcp_FlagACK;
                tcp_send(s, "tcp(TIMEWT)");
                s->unhappy = false;
                s->state = tcp_StateCLOSED;     /* support 2 msl in rst code */
            }
            break;
        }

#ifdef N_DEBUG_TX_RX_EXTRA
    printf("leaving tcp_handler: seq=%lu, ack=%lu, flags=%x, state=%d\n",
        s->seqnum, s->acknum, s->flags, s->state );
#endif
    if (s->unhappy) tcp_sendsoon(s, 0);
   return 0;
   }

/*
 * Process the data in an incoming packet.
 * Called from all states where incoming data can be received: established,
 * fin-wait-1, fin-wait-2
 *
 * returns the amount of new TCP data that we accepted from this packet
 */
static int tcp_ProcessData(tcp_Socket *s, tcp_Header *tp, int len)
{
    long ldiff;
    int diff, x;
    int startpos, up_to_end;
    word flags;
    byte *dp;

    word *options, numoptions, opt_temp;

#if defined(J_DEBUG_RX)
   printf("<- ProcessData: p->seq=%lu, p->ack=%lu, p->window=%u\n",
      ntohl(tp->seqnum), ntohl(tp->acknum), ntohs(tp->window) );
#endif
#if defined(N_DEBUG_RX)
   printf("<- ProcessData: diff=%d, s->rdatalen=%d, len=%d\n",
      diff, s->rdatalen, len-x);
#endif

    if ( s->stress > 0 ) s->stress--;

    s->window = ntohs( tp->window );

    flags = ntohs( tp->flags );
    ldiff = s->acknum - ntohl( tp->seqnum );

    if ( flags & tcp_FlagSYN ) ldiff--;  /* back up to 0 */
    diff = (int) ldiff;

    /* find the data portion */
    x = tcp_GetDataOffset(tp) << 2;     /* quadword to byte format */
    dp = (byte *)tp + x;
    /* process those options */
    if ( (numoptions = x - sizeof( tcp_Header )) != 0 ) {
        options = (word *)((byte *)(tp) + sizeof( tcp_Header));
        while ( numoptions-- > 0 ) {
            switch ( *options++ ) {
                case  0 : numoptions = 0;       /* end of options */
                          break;
                case  1 : break;                /* nop */

                          /* we are very liberal on MSS stuff */
                case  2 : if (*options == 2) {
                              opt_temp = ntohs( *(word*)(&options[1]));
                              if (opt_temp < s->mss )
                                  s->mss = opt_temp;
                          }
                          numoptions -= 2 + *options;
                          options += *options;
                          break;
            }
        }
    }
    /* done option processing */

    len -= x;           /* remove the header length */
    if (ldiff >= 0) {
        if (ldiff <= len) {  // New data received
            dp += diff;  /* Skip already received bytes */
            len -= diff;

            if (s->rdataHandler) {
                len = s->rdataHandler(s, dp, len, 0, tp);
                s->acknum += len;
            }
            else {
                /* no handler, just dump to buffer, should be indexed, handles goofs */
                /* limit receive size to our window */
                // !! if (s->rdatalen >= 0) {  // ?????  Original code
                if (len > (x = s->maxrdatalen - s->rdatalen))
                    len = x;
                if (len > 0) {
                    startpos = wrap( s->rdatalen + s->rdataoffset, s->maxrdatalen );
                    up_to_end = s->maxrdatalen - startpos;
                    if( up_to_end < len ) {
                        movmem(dp, s->rdata + startpos, up_to_end );
                        movmem(dp+up_to_end, s->rdata, len-up_to_end );
                    }
                    else
                        movmem(dp, s->rdata + startpos, len );

                    s->acknum += len;   /* Our new ack begins at end of data */
                    s->rdatalen += len;

                    // note: data offset doesn't change when we add to buffer
/*
                    s->karn_count = 3;
*/
                }
// !!           }
            }
        }
        else {  // ldiff > len  -> Duplicate packet, ignore it
#if defined(N_DEBUG_RX)
            printf(">> Duplicate packet <<\n");
#endif
            return 0;
        }
    }
    else {  // ldiff < 0  -> Missing packet, ask for resend
#if defined(N_DEBUG_RX)
        printf(">> Missing packet <<\n");
#endif
      tcp_sendsoon(s, 1);  // Send 'now.' Nevil, 13 Jan 97
    }

    // if we just sucked up some data
    // and this is a FIN packet...
    //              Joel Apisdorf 7/2/96
    //
    if( (len > 0) && (flags & tcp_FlagFIN) )
    {
        // fudge the sequence number in the packet
        // so our caller can assume no data was attached
        // when he decides whether he has received it all,
        // since we are the only code that really knows
        // how much was processed this time around
        //
        // this would not be necessary if we were willing
        // to send out an extra acknowledge of just-received
        // data (possibly with FIN bit also set), causing next
        // transmission (probably with FIN bit set) from other
        // side to have a sequence number our caller would like,
        // but that would cost an extra round-trip time
        //
        tp->seqnum = htonl( len + ntohl( tp->seqnum ) );
    }

    s->unhappy = s->sdatalen != 0;
    if (ldiff == 0 && s->unacked && chk_timeout( s->rtt_lasttran )) {
#ifdef DEBUG
        if(debug_on >1) printf("data process timeout so set unacked back to 0 from %u\n",s->unacked);
#endif DEBUG
        s->unacked = 0;
    }
#if defined(N_DEBUG_RX)
    printf("<= s->rdatalen=%d\n", s->rdatalen);
#endif

    s->timeout = set_timeout( tcp_TIMEOUT );
    return len;
}

/*
 * Format and send an outgoing segment
 */
static void tcp_send(tcp_Socket *s, char *from)
{
    tcp_PseudoHeader ph;
    struct _pkt {
        in_Header in;
        tcp_Header tcp;
        word maxsegopt[2];
    } *pkt;
    struct _tunnel {
        in_Header  in;
        struct _pkt pkt;
    } *tunnel;
    byte *dp;
    in_Header *inp;
    tcp_Header *tcpp;
    int senddatalen, sendtotlen, sendpktlen, startdata, sendtotdata;

    int ippkt;          /* 1..s->cwindow */
    int up_to_end;
    int abs_startpos;

    s->recent = 0;
    if( s->tunneladdr )
    {
        tunnel = (struct _tunnel *)_eth_formatpacket(&s->hisethaddr[0], /*0x800*/ 8);
        pkt = &tunnel->pkt;
    }
    else
    {
        pkt = (struct _pkt *)_eth_formatpacket(&s->hisethaddr[0], /*0x800*/ 8);
    }
    dp = (byte *) &pkt->maxsegopt;  /* dp constant for multi-packet sends */
    inp = &pkt->in;
    tcpp = &pkt->tcp;

    /* this is our total possible send size */
    if ( s->karn_count != 2 ) {
        /* BUG FIX : jason dent found this */
/*      sendtotdata = min( s->sdatalen - s->unacked, s->window );  */
/*      sendtotdata = max (min ( s->sdatalen, s->window ) - s->unacked, 0);  */
        sendtotdata = max(
           min((unsigned)s->sdatalen,s->window) - (unsigned)s->unacked,
           0);
/* Force unsigned compares - window can be up to 0xFFFF!  Nevil, Jan 93 */
        startdata = s->unacked;
    } else {
        sendtotdata = ((unsigned)s->sdatalen >= s->window) ?
           s->window : (unsigned)s->sdatalen;
/* Force unsigned compares - window can be up to 0xFFFF!  Nevil, Jan 93 */
/*      sendtotdata = (s->sdatalen >= s->window)? s->window : s->sdatalen;  */
        startdata = 0;
    }
/*
    if (sendtotdata < 0) sendtotdata = 0;
*/
    sendtotlen = 0;     /* running count of what we've sent */

    /* step through our packets */
// !!    for ( ippkt = 1; ippkt <= s->cwindow; ++ippkt ) {
   for (ippkt = 0; ; ) {  // Nevil, 6 Jan 97
      /* adjust size for each packet */
      if( s->tunneladdr )
         senddatalen = min( sendtotdata, (s->mss - sizeof(in_Header)) );
      else
         senddatalen = min( sendtotdata, s->mss );

    /*
      sendpktlen = senddatalen + sizeof( tcp_Header ) + sizeof( in_Header );
      inp->length = htons( sendpktlen );
    */
        /* tcp header */
      tcpp->srcPort = htons( s->myport );
      tcpp->dstPort = htons( s->hisport );
      tcpp->seqnum = htonl( s->seqnum + startdata ); /* unacked - no longer sendtotlen */
      tcpp->acknum = htonl( s->acknum );
      s->oldwindow = calcwindow( s );
      tcpp->window = htons( s->oldwindow );
#if defined(J_DEBUG_TX)
      printf("<- send:        p->ack=%lu, p->seq=%lu, p->window=%u\n",
         ntohl(tcpp->acknum), ntohl(tcpp->seqnum), ntohs(tcpp->window) );
#endif
      tcpp->flags = htons( s->flags | 0x5000 );
      tcpp->checksum = 0;
      tcpp->urgentPointer = 0;

        /* do options if this is our first packet */
      if ( s->flags & tcp_FlagSYN ) {
         sendpktlen = sizeof( tcp_Header ) + sizeof( in_Header ) + 4;
         tcpp->flags = htons( ntohs(tcpp->flags) + 0x1000 );
         pkt->maxsegopt[0] = 0x0402;
         pkt->maxsegopt[1] = htons( s->mss );
         dp += 4;
         }
      else {
         /* handle packets with data */
         if (senddatalen > 0) {
            sendpktlen = senddatalen + sizeof( tcp_Header ) + sizeof( in_Header );

            /* get data from appropriate place */
#ifdef J_DEBUG_TCP_SEND_STARTDATA
            printf( "tcp_send() ippkt=%d startdata=%d sendatalen=%d\n", ippkt, startdata, senddatalen );
#endif
#ifdef J_DEBUG_STARTDATA_PARANOIA
            if( startdata < 0 )
                printf( "warning: tcp_send() startdata %d is less than zero!\n", startdata );
#endif
            abs_startpos = wrap( s->sdataoffset + startdata, s->maxsdatalen );
            up_to_end = s->maxsdatalen - abs_startpos;
            if( up_to_end < senddatalen ) {
                movmem(s->sdata + abs_startpos, dp, up_to_end);
                movmem(s->sdata, dp+up_to_end, senddatalen-up_to_end);
            }
            else
                movmem(s->sdata + abs_startpos, dp, senddatalen);

            // note: data offset doesn't change because we are not
            // removing data from buffer, because we might need to
            // retransmit it later
            //

/*          dp[ senddatalen ] = 0; */
            }
         else {
            /* handle no-data, not-first-SYN packets */
                sendpktlen = sizeof( tcp_Header ) + sizeof( in_Header );
            }
         }

      /* internet header */
      memset( inp, 0, sizeof( in_Header ));
      inp->ver = 4;
      inp->hdrlen = 5;
      inp->tos = s->tos;
      inp->identification = htons( ++ip_id );   /* was post inc */
//      inp->frag = 0;
      inp->ttl = s->ttl;
      inp->proto = IPPROTO_TCP;
      inp->checksum = 0;
      inp->source = htonl( s->myaddr );
      inp->destination = htonl( s->hisaddr );
      inp->length = htons( sendpktlen );

      inp->checksum = ~checksum( inp, sizeof(in_Header));

      /* compute tcp checksum */
      ph.src = inp->source;   /* already INTELled */
      ph.dst = inp->destination;
      ph.mbz = 0;
      ph.protocol = IPPROTO_TCP;
      ph.length = htons( sendpktlen - sizeof(in_Header));
/*
      ph.checksum = checksum(&pkt->tcp, (sendpktlen - sizeof(in_Header) +1) & 0xfffe);
*/
      ph.checksum = checksum(&pkt->tcp, sendpktlen - sizeof(in_Header));

      tcpp->checksum = ~checksum(&ph, sizeof(ph));

      if (_dbugxmit) (*_dbugxmit)((sock_type *)s,inp,tcpp, "tcp_send()");
      if (debug_on >= 1) {
         // display stuff on the far right of the first screen line
         //
         mono[150]++;  colour[150]++;
         mono[152] = colour[152] = (s->flags & tcp_FlagSYN) ? 'S' : ' ';
         mono[154] = colour[154] = (s->flags & tcp_FlagACK) ? 'A' : ' ';
         mono[156] = colour[156] = (s->flags & tcp_FlagFIN) ? 'F' : ' ';
         mono[158] = colour[158] = (s->flags & tcp_FlagRST) ? 'R' : ' ';
         }

      if (s->tunneladdr) {
         inp = &tunnel->in;

         inp->ver = 4;
         inp->hdrlen = 5;
         inp->tos = 0;
         inp->identification = htons( ++ip_id );   /* was post inc */
         inp->frags = 0;
         inp->fo = 0;
         inp->ttl = in_DefaultTTL;
         inp->proto = IPPROTO_IP_IN_IP;      /* please unwrap me! */
         inp->source = htonl( s->myaddr );
         inp->destination = htonl( s->tunneladdr );
         inp->length = htons( sizeof(in_Header) + sendpktlen );
         inp->checksum = 0;
         inp->checksum = ~checksum( inp, sizeof(in_Header));
         }

      ++ippkt;  // Nevil, 6 Jan 97
      if (_eth_send(ntohs(inp->length))) { // Pkt send failed
         s->recent = 1;   // don't recurse to me now, try LATER
         tcp_sendsoon(s, 0);
         s->cwindow = 1;  // Reduce tx rate by setting window size to 1
         s->wwindow = 0;  // Nevil, 10 Jan 97
         return;
         }
      startdata += senddatalen;  // Packet sent OK

      // Update these here - that way it will be correct
      // if send fails the next time 'round this loop
      //
      s->unacked = startdata;
      if( s->maxunacked < s->unacked )
          s->maxunacked = s->unacked;

      /* do next ip pkt */
      sendtotlen += senddatalen;
      sendtotdata -= senddatalen;
      if (ippkt >= s->cwindow || ippkt >= MAXPKTSEND || sendtotdata <= 0)
         break;  // Nevil, 9 Jan 97
      }

#ifdef DEBUG
   if (debug_on)
      printf(" Sent %u/%u bytes in %u/%u packets  "
         "with (%u) unacked SYN %lu  from %s\n",
         sendtotlen,s->window, ippkt,s->cwindow,
         s->unacked, s->seqnum, from);
#endif DEBUG
    s->vj_last = 0;
    if ( s->karn_count == 2 ) {
        if (s->rto) {
           s->rto = (s->rto * 3) / 2;
          }
        else s->rto = 4;
        if (s->rto > MAXRTO) s->rto = MAXRTO;  // Nevil, 13 Jan 97
    } else {
        /* vj_last nonzero if we expect an immediate response */
        if (s->unhappy || s->sdatalen)
            s->vj_last = set_ttimeout( 0 );
        s->karn_count = 0;
    }
#ifdef N_DEBUG_TX
printf("-> send: %u/%u bytes, %u/%u pkts, s_unack=%u, s->rto=%u;  %s\n",
   sendtotlen,s->window, ippkt,s->cwindow,  s->unacked,  s->rto, from);
#endif

    s->rtt_time = set_ttimeout(s->rto);
    if (sendtotlen > 0) s->rtt_lasttran = s->rtt_time + s->rto;
}

/*
 * Format and send a reset tcp packet
 */
static void tcp_rst( in_Header *his_ip, tcp_Header *oldtcpp)
{
    tcp_PseudoHeader ph;
    struct _pkt {
        in_Header in;
        tcp_Header tcp;
        word maxsegopt[2];
    } *pkt;

    static longword nextrst = 0L;
    word oldflags;
    in_Header *inp;
    tcp_Header *tcpp;
    eth_Header *eth;
    int sendtotlen;     /* length of packet */
    int temp;
    longword templong;

    /* see RFC 793 page 65 for details */

    if ( !chk_timeout( nextrst )) return;
    nextrst = set_ttimeout( 1 );


    oldflags = ntohs( oldtcpp->flags );
    if (oldflags & tcp_FlagRST ) return;
#ifdef NEVER
    if ( (oldflags & (tcp_FlagACK | tcp_FlagFIN)) == (tcp_FlagACK | tcp_FlagFIN) ){
        templong = oldtcpp->seqnum;
        oldtcpp->seqnum = oldtcpp->acknum;
        oldtcpp->acknum = templong;
        oldflags = tcp_FlagACK;
    } else if ((oldflags & (tcp_FlagSYN | tcp_FlagACK)) ==  tcp_FlagSYN ) {
        oldtcpp->acknum = htonl( ntohl( oldtcpp->seqnum ) + 1 );
        oldtcpp->seqnum = 0;
        oldflags = tcp_FlagACK | tcp_FlagRST;
    } else if ( oldflags & tcp_FlagACK ) {
        oldtcpp->seqnum = oldtcpp->acknum;
        oldtcpp->acknum = 0;
    } else {
        oldtcpp->acknum = htonl( ntohl(oldtcpp->seqnum) + 1);
        oldtcpp->seqnum = 0;
    }
    if ( oldflags & ( tcp_FlagFIN | tcp_FlagSYN ) == 0 )
        oldflags ^= tcp_FlagACK | tcp_FlagRST;

    if ( oldflags & tcp_FlagACK ) {
        oldtcpp->seqnum = oldtcpp->acknum;

#else
    /* better strategy - Dean Roth */
    if ( oldflags & tcp_FlagACK ) {
        oldtcpp->seqnum = oldtcpp->acknum;
        oldtcpp->acknum = 0;
        oldflags = tcp_FlagRST;
    } else if ((oldflags & (tcp_FlagSYN | tcp_FlagACK)) ==  tcp_FlagSYN ) {
        oldtcpp->acknum = htonl( ntohl( oldtcpp->seqnum ) + 1 );
        oldtcpp->seqnum = 0;
        oldflags = tcp_FlagACK | tcp_FlagRST;
    } else {
        temp = ntohs( his_ip->length) - in_GetHdrlenBytes( his_ip );
        oldtcpp->acknum = htonl( ntohl( oldtcpp->seqnum ) + temp );
        oldtcpp->seqnum = 0;
        oldflags = tcp_FlagRST;
    }
#endif

    /* convoluted mechanism - reads his ethernet address or garbage */
    eth = _eth_hardware( (byte *)his_ip );

    pkt = (struct _pkt *)_eth_formatpacket( eth, 8);
    inp = &pkt->in;
    tcpp = &pkt->tcp;

    sendtotlen = sizeof( tcp_Header ) + sizeof( in_Header );
    memset( inp, 0, sizeof( in_Header ));
    inp->length = htons( sendtotlen );

    /* tcp header */
    tcpp->srcPort = oldtcpp->dstPort;
    tcpp->dstPort = oldtcpp->srcPort;
    tcpp->seqnum = oldtcpp->seqnum;
    tcpp->acknum = oldtcpp->acknum;
    tcpp->window = 0;
/*    tcpp->flags = htons( oldflags ); */
    /* BUG FIX : jason dent found this thanks to SCO */
    tcpp->flags = htons( (oldflags & 0x0fff ) | 0x5000 );
    tcpp->checksum = 0;
    tcpp->urgentPointer = 0;

    /* internet header */
    inp->ver = 4;
    inp->hdrlen = 5;
    inp->tos = his_ip->tos;
    inp->identification = htons( ++ip_id );
//    inp->frag = 0;
    inp->ttl = in_DefaultTTL;
    inp->proto = IPPROTO_TCP;
    inp->checksum = 0;
    inp->source = his_ip->destination;
    inp->destination = his_ip->source;

    inp->checksum = ~checksum( inp, sizeof(in_Header))/* 0*/;

    /* compute tcp checksum */
    ph.src = inp->source;       /* already INTELled */
    ph.dst = inp->destination;
    ph.mbz = 0;
    ph.protocol = 6;
    ph.length = htons( sendtotlen - sizeof(in_Header));

    ph.checksum = checksum(&pkt->tcp, sizeof(tcp_Header));
    tcpp->checksum =  ~checksum(&ph, sizeof(ph));

    if (_dbugxmit) (*_dbugxmit)(NULL,inp,tcpp, "tcp_rst()");

    _eth_send( ntohs( inp->length ));
}




/**********************************************************************
 * socket functions
 **********************************************************************/

/* socket based stuff */

/*
 * sock_read - read a socket with maximum n bytes
 *           - busywaits until buffer is full but calls s->usr_yield
 *           - returns count also when connection gets closed
 */
sock_read(sock_type *s, byte *dp, int len )
{
    int templen, count, proto_is_tcp;

    proto_is_tcp = (s->udp.ip_type == IPPROTO_TCP);
    if (!proto_is_tcp && s->udp.ip_type!=IPPROTO_UDP)
        return( 0 );                // do nothing when socket is closed

    count = 0;
    do {
        if (proto_is_tcp)
            templen = tcp_read( &(s->tcp), dp, len);
        else
            templen = udp_read( &(s->udp), dp, len );
        if (s->tcp.usr_yield) (s->tcp.usr_yield)();
        if (templen < 1 ) {
            if (!tcp_tick( s )) return( count );
        } else {
            count += templen;
            dp += templen;
            len -= templen;
        }
    } while ( len );
    return( count );
}

/*
 * read a socket with maximum n bytes
 *
 * does not busywait until buffer is full
 */
sock_fastread(sock_type *s, byte *dp, int len )
{
    if ( s->udp.ip_type == IPPROTO_UDP )
        return udp_read( &s->udp, dp, len );

    if ( s->udp.ip_type == IPPROTO_TCP )
        return tcp_read( &s->tcp, dp, len);

    // do nothing when socket is closed
    return 0;
}

/*
 * sock_write - writes data and returns length written
 *            - does not perform flush
 *            - repeatedly calls s->usr_yield
 */
sock_write(sock_type *s, byte *dp, int len)
{
    int offset, oldlen, oldmode, proto_is_tcp;

    oldlen = len;
    offset = 0;

    proto_is_tcp = (s->udp.ip_type == IPPROTO_TCP);
    if ( proto_is_tcp )
        oldmode = s->tcp.flags & tcp_FlagPUSH;
    else if (s->udp.ip_type != IPPROTO_UDP)
        return( 0 );                // do nothing when socket is closed
    while ( len  > 0) {
        if (proto_is_tcp) {
            s->tcp.flags |= oldmode;
            offset += tcp_write( &s->tcp, &dp[ offset ], len );
        } else
            offset += udp_write( &s->udp, &dp[ offset ], len );
        len = oldlen - offset;
        if (s->udp.usr_yield)
            (s->udp.usr_yield)();
        if (!tcp_tick(s))
            return( 0 );
    }
    return( oldlen );
}

sock_fastwrite(sock_type *s, byte *dp, int len)
{
    if ( s->udp.ip_type == IPPROTO_UDP )
        return udp_write( &s->udp, dp, len );

    if ( s->udp.ip_type == IPPROTO_TCP )
        return tcp_write( &s->tcp, dp, len);

    // do nothing when socket is closed
    return 0;
}

/* change whether default or user-allocated
 * buffer will be used for received data
 *
 * note that fields are used rather than tcp fields,
 * to highlight the fact that they are common to both
 */
int sock_set_rx_buf( udp_Socket *s, byte *dp, int len )
{
    int up_to_end;

    // length must be positive
    //
    if( len < 0 )
        return 0;

    // socket must really be open
    //
    if( s->ip_type != IPPROTO_UDP
    &&  s->ip_type != IPPROTO_TCP )
        return 0;

    // fix pointer and length if they want default buffering
    //
    if( len == 0 || dp == NULL )
    {
        dp = s->rddata;

        if (s->ip_type == IPPROTO_TCP)
            len = tcp_MaxBufSize;
        else
            len = udp_MaxBufSize;
    }

    // if new buffer isn't big enough to hold existing
    // data, we must stop now, since it has already
    // been acknowledged
    //
    if( s->rdatalen > len )
        return 0;

    // if there was any old data and pointer changed,
    // copy it to the new spot, at offset zero
    //
    if( s->rdatalen && dp!=s->rdata )
    {
        up_to_end = s->maxrdatalen - s->rdataoffset;
        if( up_to_end < s->rdatalen )
        {
            memcpy( dp, s->rdata + s->rdataoffset, up_to_end );
            memcpy( dp+up_to_end, s->rdata, s->rdatalen - up_to_end);
        }
        else
            memcpy( dp, s->rdata + s->rdataoffset, s->rdatalen );

        s->rdataoffset = 0;
    }

    // now we can change the pointer and length
    //
    s->maxrdatalen = len;
    s->rdata = dp;

    // successful!
    //
    return s->maxrdatalen;
}

/* change whether default or user-allocated
 * buffer will be used for transmitted data
 *
 * only works for TCP sockets, since only they
 * have buffers for output data
 */
int sock_set_tx_buf( tcp_Socket *s, byte *dp, int len )
{
    int up_to_end;

    // length must be positive
    //
    if( len < 0 )
        return 0;

    // socket must really be open and must be TCP
    //
    if( s->ip_type != IPPROTO_TCP )
        return 0;

    // fix pointer and length if they want default buffering
    //
    if( len == 0 || dp == NULL )
    {
        dp = s->sddata;
        len = tcp_MaxBufSize;
    }

    // if new buffer isn't big enough to hold existing
    // data, we must stop now, since it has already
    // been acknowledged
    //
    if( s->sdatalen > len )
        return 0;

    // if there was any old data and pointer changed,
    // copy it to the new spot, at offset zero
    //
    if( s->sdatalen && dp!=s->sdata )
    {
        up_to_end = s->maxsdatalen - s->sdataoffset;
        if( up_to_end < s->sdatalen )
        {
            memcpy( dp, s->sdata + s->sdataoffset, up_to_end );
            memcpy( dp+up_to_end, s->sdata, s->sdatalen - up_to_end);
        }
        else
            memcpy( dp, s->sdata + s->sdataoffset, s->sdatalen );

        s->sdataoffset = 0;
    }

    // now we can change the pointer and length
    //
    s->maxsdatalen = len;
    s->sdata = dp;

    // successful!
    //
    return s->maxsdatalen;
}

void sock_noflush( sock_type *s )
{
    if ( s->tcp.ip_type == IPPROTO_TCP ) {
        s->tcp.flags &= ~tcp_FlagPUSH;
        s->tcp.sock_mode |= TCP_LOCAL ;
    }
}

void sock_flush( sock_type *s )
{
    if ( s->tcp.ip_type == IPPROTO_TCP ) {
        s->tcp.sock_mode &= ~TCP_LOCAL;
        tcp_Flush( &s->tcp );
    }
}

/*
 * sock_flushnext - cause next transmission to have a flush
 */
void sock_flushnext( sock_type *s)
{
    if (s->tcp.ip_type == IPPROTO_TCP ) {
        s->tcp.flags |= tcp_FlagPUSH;
        s->tcp.sock_mode &= ~TCP_LOCAL;
    }
}

/*
 * sock_putc - put a character
 *           - no expansion but flushes on '\n'
 *           - returns character
 */
byte sock_putc( sock_type *s, byte c )
{
    if (( c == '\n') || ( c == '\r'))
        sock_flushnext( s );
    sock_write( s, &c, 1 );
    return( c );
}

word sock_getc( sock_type *s )
{
    char ch;
    return( sock_read( s, &ch, 1 ) < 1 ? EOF : ch );
}

/*
 * sock_puts - does not append carriage return in binary mode
 *           - returns length
 */
sock_puts( sock_type *s, byte *dp )
{
    int len, oldmode;

    len = strlen( dp );

    if (s->tcp.sock_mode & TCP_MODE_ASCII ) {
        if (s->tcp.ip_type == IPPROTO_TCP )
            s->tcp.sock_mode |= TCP_LOCAL;
        sock_noflush( s );
        if (len) sock_write( s, dp, len );
        sock_flushnext( s );
        sock_write( s, "\r\n", 2 );
    } else {
        sock_flushnext( s );
        sock_write( s, dp, len );
    }
    return( len );
}

/*
 * sock_update - update the socket window size to the other guy
 */
static void sock_update( tcp_Socket *s )
{
    if (s->ip_type == IPPROTO_TCP) {
        if ( !s->rdatalen )
            tcp_send(s, "sock_update()");  /* Update the window */
        else
            tcp_sendsoon(s, 0);
    }
}

/*
 * sock_gets - read a string from any socket
 *           - return length of returned string
 *           - removes end of line terminator(s)
 *
 *           - Quentin Smart fixed some problems
 *
 * ignores whether socket is TCP or UDP         why?
 * ignores whether socket mode is ASCII         why?
 */
int sock_gets( sock_type *s, byte *dp, int n )
{
    int len, templen, up_to_end;
    char *src_p, *temp, *temp2;

    if ( s->udp.ip_type != IPPROTO_UDP
    &&   s->udp.ip_type != IPPROTO_TCP )
        // do nothing when socket is closed
        return( 0 );

    src_p = s->udp.rdata + s->udp.rdataoffset;

    if ( s->udp.rdatalen == 0 ) return( 0 );

    // eat trailing \n or \0 from previous line
    if ( *src_p == 0 || *src_p == '\n' ) {
        s->udp.rdatalen--;
        if ( !s->udp.rdatalen ) return( 0 );
        s->udp.rdataoffset = wrap( s->udp.rdataoffset + 1, s->udp.maxrdatalen );
    }

    if ( --n > s->udp.rdatalen ) n = s->udp.rdatalen;

    // Q.Smart found and fixed a bug here

    // copy everything
    up_to_end = s->udp.maxrdatalen - s->udp.rdataoffset;
    if( up_to_end < n )
    {
        memcpy( dp, s->udp.rdata + s->udp.rdataoffset, up_to_end );
        memcpy( dp+up_to_end, s->udp.rdata, n-up_to_end );
    }
    else
        memcpy( dp, s->udp.rdata + s->udp.rdataoffset, n );
    dp[ n ] = 0;                // terminate new string
    temp = memchr( dp, '\n', n);
    temp2= memchr( dp, '\r', n);

    if (temp)  *temp = 0;
    if (temp2) *temp2= 0;

    // fail if there were no carriage returns
    if ( !temp2 ) {
        *dp = 0;
        return( 0 );
    }

    // instead of calling strlen(dp) calculate length
    //
    // for speed?  what a joke!  if we wanted to do that,
    // we should keep a count of bytes in buffer that this
    // function has already dealt with, instead of the 2
    // calls to movmem() that are in here!
    //
    // of course, that would keep the whole buffer tied
    // up until the caller had read all data out of it,
    // which might never occur before buffer starvation
    // if TCP RX continues (only matters to TCP, since UDP
    // RX overwrites the whole buffer anyways -- has that
    // bitten you by the way?)
    //
    // better yet, we should give the caller a pointer
    // into our buffer instead of calling memcpy() at all
    //
    // Joel 6/26/96 fixed bug: wasn't subtracting dp
    // when both \n and \r had been found
    //
    len = temp ? (min(temp, temp2) - dp) : (temp2 - dp);

    // expect \r\n or \r\0 or \n or \r
    // so get first of the two

    n = len + 1;                // account for first \r

    // we check next char if it exists, and skip it if 0, \r, or \n
    //
    if ((s->udp.rdatalen > n)
    && !s->udp.rdata[ wrap(s->udp.rdataoffset + n, s->udp.maxrdatalen) ] )
        n++;
    s->udp.rdataoffset = wrap( s->udp.rdataoffset + n, s->udp.maxrdatalen );
    s->udp.rdatalen -= n;
    if (s->udp.rdatalen < 0) s->udp.rdatalen = 0;

    sock_update( &s->tcp );   /* new window */
    return( len );
}


/*
 * sock_dataready - returns number of bytes waiting to be read
 *                - if in ASCII mode, return 0 until a line is present
 *                  or the buffer is full
 */
word sock_dataready( sock_type *s )
{
    int len;
    int up_to_end;
    char *p;

    if (!(len = s->tcp.rdatalen)) return( 0 );

    if ( s->tcp.sock_mode & TCP_MODE_ASCII ) {
        p = s->tcp.rdata + s->tcp.rdataoffset;
        if ( *p == '\n' ) {
            s->tcp.rdataoffset = wrap( s->tcp.rdataoffset + 1, s->tcp.maxrdatalen );
            p = s->tcp.rdata + s->tcp.rdataoffset;
            s->tcp.rdatalen = --len;
            if ( ! len ) return( 0 );
        }
        /* check for terminating \r */
        up_to_end = s->tcp.maxrdatalen - s->tcp.rdataoffset;
        if ( up_to_end < len )
        {
            if ( memchr( p, '\r', up_to_end))
                return( len );
            if ( memchr( s->tcp.rdata, '\r', len-up_to_end))
                return( len );
        }
        else
            if ( memchr( p, '\r', len))
                return( len );
        return( 0 );
    } else
        return( len );
}

// this used to also return TRUE when TCP
// state was ESTCL or CLOSWT
//
// I couldn't find any callers, and I wanted a function
// that only returned TRUE when ready to TX, so I changed it
//
// someday maybe create separate functions that tell you
// whether you can TX and RX
//
//              Joel Apisdorf 7/3/96
//
sock_established( sock_type *s )
{
    switch ( s->tcp.ip_type ) {
        case IPPROTO_UDP :
                return( 1 );
        case IPPROTO_TCP :
                return( s->tcp.state == tcp_StateESTAB );
        default :
                return( 0 );
    }
}

void sock_close( sock_type *s )
{
    switch (s->udp.ip_type) {
        case IPPROTO_UDP :
                udp_close( &s->udp );
                break;
        case IPPROTO_TCP :
                tcp_close( &s->tcp );
                tcp_tick( s );
                break;
    }
}

void sock_sturdy( sock_type *s, int level )
{
    s->tcp.rigid = level;
    if ( s->tcp.rigid < s->tcp.stress ) sock_abort( s );
}

/* called by macro sock_wait_established()
 */
_ip_delay0( s, timeoutseconds, fn, statusptr )
    sock_type *s;
    int timeoutseconds;
    waitprocref fn;
    int *statusptr;
{
    int status;
    ip_timer_init( &s->udp, timeoutseconds );
    do {
        if ( s->tcp.ip_type == IPPROTO_TCP ) {
            if ( tcp_established( &s->tcp )) {
                status = 0;
                break;
            }
        }
        kbhit();        /* permit ^c */
        if ( !tcp_tick( s )) {
             if (!s->tcp.err_msg) s->tcp.err_msg = "Host refused connection";
             status = -1;       /* get an early reset */
             break;
        }
        if ( ip_timer_expired( &s->udp )) {
            s->tcp.err_msg = "Open timed out";
            sock_close( s );
            status = -1;
            break;
        }

        if ( fn )
        {
            status = fn(s);
            if ( status )
                break;
        }

        if ( s->tcp.usr_yield )
            (*s->tcp.usr_yield)();

        if ( s->tcp.ip_type == IPPROTO_UDP ) {
            status = 0;
            break;
        }
    } while ( 1 );
    if (statusptr) *statusptr = status;
    return( status );
}

/* called by macro sock_wait_input()
 */
_ip_delay1( s, timeoutseconds, fn, statusptr)
    sock_type *s;
    int timeoutseconds;
    waitprocref fn;
    int *statusptr;
{
    int status;
    ip_timer_init( &s->udp, timeoutseconds );

    sock_flush( s );    /* new enhancement */

    do {
        if ( sock_dataready( s )) {
            status = 0;
            break;
        }
        kbhit();        /* permit ^c */

        if ( !tcp_tick( s )) {
            status = 1;
            break;
        }
        if ( ip_timer_expired( &s->udp )) {
            s->tcp.err_msg = "Connection timed out";
            sock_close( s );
            status = -1;
            break;
        }
        if (fn) {
            status = fn(s);
            if( status )
                break;
        }
        if ( s->tcp.usr_yield )
            (*s->tcp.usr_yield)();
    } while ( 1 );
    if (statusptr) *statusptr = status;
    return( status );
}

/* called by macro sock_wait_closed();
 */
_ip_delay2( s, timeoutseconds, fn, statusptr)
    sock_type *s;
    int timeoutseconds;
    waitprocref fn;
    int *statusptr;
{
    int status;
    ip_timer_init( &s->udp, timeoutseconds );

    if (s->tcp.ip_type != IPPROTO_TCP ) {
        if ( statusptr ) * statusptr = 1;
        return( 1 );
    }

    do {
        /* in this situation we KNOW user not planning to read rdata */
        s->tcp.rdatalen = 0;
        kbhit();        /* permit ^c */
        if ( !tcp_tick( s )) {
            status = 1;
            break;
        }
        if ( ip_timer_expired( &s->udp )) {
            s->tcp.err_msg = "Connection timed out";
            sock_abort( s );
            status = -1;
            break;
        }
        if (fn) {
            status = fn(s);
            if( status )
                break;
        }
        if ( s->tcp.usr_yield ) (*s->tcp.usr_yield)();

    } while ( 1 );
    if (statusptr) *statusptr = status;
    return( status );
}


char *rip( char *s )  // Replace \n and \r by \0 in a string
{
    char *temp;

    temp = strchr( s, '\n' );
    if( temp )
        *temp = 0;

    temp = strchr( s, '\r' );
    if( temp )
        *temp = 0;

    return( s );
}

// turn off compiler warning about unused parameters
// since I know they won't be used and I must have
// them because calling code cannot be changed
//
#pragma warn -par

/* Default version, never used */
void au_send_mon(int length, unsigned char *buffer)
{
    return;
}

// turn compiler warning about unused parameters
// back to whatever it was before I messed with it
//
#pragma warn .par

